////////////////////////////////////////////////////////////////////
// Net.h: Class that describes a generic Net. 
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "Net.h"

vector<vector<double>*>* AstarTable;

// double cost_to_target(routingNode* rnode,routingNode* sink_rnode){

//   if(rnode->type==INPUT_PIN || rnode->type==OUTPUT_PIN)
//     return 0;

//   // special case for IOB wires that have no LRList
//   if((rnode->LRList).size()==0)
//     return (double)((sink_rnode->LRList).size() + 2);

//   // special case for sink at root
//   if((sink_rnode->LRList).size()==0)
//     return (double)((rnode->LRList).size() + 2);

//   list<char>::iterator sink_itr = (sink_rnode->LRList).begin();
//   list<char>::iterator rnode_itr = (rnode->LRList).begin();

//   int common_ctr;
//   for(common_ctr=0; *sink_itr==*rnode_itr && rnode_itr!=(rnode->LRList).end();
//       rnode_itr++,sink_itr++,common_ctr++);

//   if(common_ctr==(sink_rnode->LRList).size())  // alright, the wire is directly connected to sink clb
//     return 2.0;
//   else{  // wire is not directly connected to sink clb ==> common_ctr < size of sink's LRList 
//     if(common_ctr==(rnode->LRList).size())
//       return (double)( (sink_rnode->LRList.size()) - (rnode->LRList.size()) + 2 );
//     else{  // common_ctr < size of rnode's LRList and size of sink_rnode's LRList 
//       if(*rnode_itr == *sink_itr)
// 	ERROR("shouldn't be in this branch...");
//       else
// 	return (double)((rnode->LRList).size()-common_ctr-1  +  (sink_rnode->LRList).size()-common_ctr + 2);
//     }
//   }

// }


// double cost_to_target(routingNode* rnode,routingNode* sink_rnode){

//   if(rnode->type==INPUT_PIN || rnode->type==OUTPUT_PIN) return 0.;
//   list<char>::iterator sink_itr = (sink_rnode->LRList).begin();
//   list<char>::iterator rnode_itr = (rnode->LRList).begin();
//   int common_ctr;
//   for(common_ctr=0; *sink_itr==*rnode_itr && rnode_itr!=(rnode->LRList).end();rnode_itr++,sink_itr++,common_ctr++);
//   return (double)((rnode->lrSize-common_ctr  +  sink_rnode->lrSize-common_ctr  +  1));

// }


double cost_to_target(routingNode* rnode,routingNode* sink_rnode){
  if(rnode==0 || sink_rnode==0)
    ERROR("\nNULL node in cost to target calculation function...\n");

  int rnode_cluster = rnode->clusterNumber;

  /* if we have an input pin or sink node, return 0 */
  if(rnode_cluster<0.0)
    return 0.0;

  double estimate = AstarTable->at(rnode_cluster)->at(sink_rnode->sinkID);
  if(estimate < 0.0)
    ERROR("currently, the estimate to a sink cannot be -ve...");

  return estimate;

}

// Give this net a name
void Net::set_name(char* n){
  if(!n)
    ERROR("net name cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Set this net's source terminal
void Net::set_source(Terminal* s){  
  if(!s)
    ERROR("net source cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  source=s;
}

// Initial creation of vector of sinks
void Net::create_sinkVector(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(numSinks==0){
//      ERROR("num sinks is 0 here...");
    cout<<endl<<endl<<name<<" HAS NO SINKS!!";
    return;
  }
  sinkVector=new vector<Terminal*>(numSinks);

  list<Terminal*>::iterator sink_itr=sinkList.begin();
  for(int i=0;sink_itr!=sinkList.end();sink_itr++,i++){
    sinkVector->at(i)=(*sink_itr);
  }


}

// Sort vector of sinks in increasing order of separation from source...A*
void Net::sort_sinkVector(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  double i_dist;double j_dist;
  Terminal* temp_terminal;

  if(sinkVector){
 
//    for(int i=0;i<sinkVector->size();i++){
//       i_dist=cost_to_target(source->get_rNode(),sinkVector->at(i)->get_rNode());
//       for(int j=i+1;j<sinkVector->size();j++){
// 	j_dist=cost_to_target(source->get_rNode(),sinkVector->at(j)->get_rNode());
// 	if(i_dist>j_dist){
// 	  //gotta swap the positions here...
// 	  temp_terminal=sinkVector->at(i);
// 	  sinkVector->at(i)=sinkVector->at(j);
// 	  sinkVector->at(j)=temp_terminal;
// 	}
//       }
//     }

    for(int i=0;i<sinkVector->size();i++){
      i_dist = (source->rNode->allSinksTable).at(sinkVector->at(i)->rNode->sinkID);
      if(i_dist < 1.0)
	ERROR("sink distance should not be less than 1.0 here...");
      for(int j=i+1;j<sinkVector->size();j++){
	j_dist = (source->rNode->allSinksTable).at(sinkVector->at(j)->rNode->sinkID);
	if(j_dist < 1.0)
	  ERROR("sink distance should not be less than 1.0 here...");
	if(j_dist < i_dist){
	  temp_terminal=sinkVector->at(i);
	  sinkVector->at(i)=sinkVector->at(j);
	  sinkVector->at(j)=temp_terminal;
	}
      }

    }

  }

}

// Add a terminal to this net's sink list
void Net::addto_sinkList(Terminal* s){
  if(!s)
    ERROR("sink cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  sinkList.push_back(s);
  numSinks++;
}

// Get name of this net
char* Net::get_name(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Rip up the existing routing tree, and decrement sharing
void Net::rip_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"RIPPING ROUTE TREE OF "<<name;
#endif

  rTree.postorderTraverse(rTree.get_root(),2);
  rTree.set_root(0);
}

// Reset the routing nodes in the tree to not counted in sharing calculations
void Net::reset_counted(){
  rTree.postorderTraverse(rTree.get_root(),5);
}

// Rip up the branch of the current route tree that contains sink_rnode
void Net::rip_rTreeBranch(routingNode* sink_rnode){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(sink_rnode==0)
    ERROR("sink routing node cannot be NULL...");

  treeNode* sink_tree_node=rTree.findNode(rTree.get_root(),sink_rnode);
  if(sink_tree_node==0)
    ERROR("could not find sink in route tree...");

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"LOOKING AT BRANCH OF NET "<<name<<" THAT SINKS AT "<<sink_rnode->name;
#endif

  list<treeNode*>::iterator ch_itr;
  treeNode* tree_node=sink_tree_node; treeNode* temp_node;
  while(1){
    if(tree_node->get_numChildren()>1 || tree_node->get_rNode()->get_type()==SOURCE){
      tree_node->eraseChild(&ch_itr);
      /* this is the root of the subtree being deleted...break out */
      break;
    }

    temp_node=tree_node->get_parent();
    if(temp_node->get_numChildren()>1 || temp_node->get_rNode()->get_type()==SOURCE)
      ch_itr=temp_node->get_childItr(tree_node);

    /* change sharing and wire cost to refelct the deletion of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost-=1;
    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
      wireCost-=1;

//     wireCost-=1;

    /* decrement occupancy and delete tree node */
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()-1);

#if defined(PLACER_DEBUG)
    cout<<endl<<"\tdecrement and delete "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy();
#endif

    delete tree_node;

    tree_node=temp_node;
  }

}

// Print the current route tree of this net
void Net::print_rTree(FILE* fp){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  rTree.print_info(fp);
}

void Net::print_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  rTree.print_info();
}

// Whether this net's cost contributions have already been considered during
// an annealing move
void Net::set_isCounted(bool c){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  isCounted=c;
}

// Get whether this net's cost contributions have already been considered during
// an annealing move
bool Net::get_isCounted(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return isCounted;
}

// Reconnect the sub-tree rooted at join_node to this net's current route tree.
// Increment occupancies of the routing nodes that constitute the subtree.
int Net::reconnect(treeNode* join_node){
  treeNode* child;

  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(join_node==0)
    ERROR("NULL join node here...");

  /* locate the tree node in rTree that corresponds to join_node */
  treeNode* tree_node=rTree.getNode(join_node->get_rNode());
  if(tree_node==0){
//     ERROR("could not find tree node in route tree...");
    return -1;
  }

  /* change sharing and wire cost to avoid double counting*/
  if(join_node->get_rNode()->get_occupancy()>join_node->get_rNode()->get_capacity())
    sharingCost-=1;

//   wireCost-=1;

  if(join_node->get_rNode()->get_type()==BUS_OR_WIRE)
    wireCost-=1;

  /* decrement the occupancy of join_node...since we have double counted */
  join_node->get_rNode()->set_occupancy(join_node->get_rNode()->get_occupancy() - 1);
  /* now increment the occupancies of routing nodes in the sub-tree */
  incrementOccupancies(join_node,0);

  /* reconnect the sub-tree rooted at join_node */
  join_node->create_childListItr();
  for(child=join_node->getnext_childList();child!=0;child=join_node->getnext_childList()){
    tree_node->addto_childList(child);
    child->set_parent(tree_node);
  }
  join_node->remove_childListItr();

  /* finally delete join_node */
  delete join_node;
  return 1;

}

// Increment occupancies of the routing nodes that constitute sub-tree rooted at r
void Net::incrementOccupancies(treeNode* r,int how){
  treeNode* tree_node=r;
  treeNode* child;

  if(how==0){
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy() + 1);
    /* change sharing and wire cost to reflect the addition of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost+=1;

//     wireCost+=1;

    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
      wireCost+=1;
  }
  else{
    if(tree_node->get_rNode()->get_type()!=SOURCE && tree_node->get_rNode()->get_type()!=SINK)
      tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy() + 1);
  }

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    incrementOccupancies(child,how);
  tree_node->remove_childListItr();

}

// Calculate the sharing and wire cost of this net's current route
void Net::calcCosts(int* sh_cost,int* wr_cost){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(*sh_cost!=0 || *wr_cost!=0)
    ERROR("uninitialized cost parameters here...");

  cout<<endl<<endl<<"------- Calculating costs for "<<name<<" -------";

  wsCostCalc(rTree.get_root(),sh_cost,wr_cost);
}

// Recursively calculate wire and sharing cost
void Net::wsCostCalc(treeNode* r,int* sh_cost,int* wr_cost){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    wsCostCalc(child,sh_cost,wr_cost);
  tree_node->remove_childListItr();

  *wr_cost+=1;
  cout<<endl<<"\tincremented wire cost due to "<<tree_node->get_rNode()->name<<
    " to "<<*wr_cost;

  routingNode* rnode=tree_node->get_rNode();
//   if(rnode->get_occupancy()>rnode->get_capacity()){
//     *sh_cost+=1;
//     cout<<" and sharing cost to "<<*sh_cost;
//   }

  if(rnode->get_occupancy()>rnode->get_capacity() && rnode->get_counted()==false){
//     *sh_cost+=(rnode->get_occupancy() - rnode->get_capacity());
    *sh_cost+=1;
    cout<<" and sharing cost to "<<*sh_cost;
    rnode->set_counted(true);
  }

}

// Get the total number of nodes in the current route tree
int Net::get_rNodeCount(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  int node_count=0;
  countNodes(rTree.get_root(),&node_count);
  return node_count;
}

// Count number of nodes in the current route tree
void Net::countNodes(treeNode* r,int* count_ptr){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    countNodes(child,count_ptr);
  tree_node->remove_childListItr();

//   *count_ptr+=1;
  if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
    *count_ptr+=1;

}

// Make the source routing node the root of  rTree
void Net::init_rTreeRoot(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  treeNode* root_node=new treeNode;
  root_node->set_rNode(source->get_rNode());
  rTree.set_root(root_node);
}

// Get this net's source terminal
Terminal* Net::get_source(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return source;
}

// Get num sinks of this net
int Net::get_numSinks(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return numSinks;
}

// Print details about this net
void Net::print_Info(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"===================================";
  cout<<endl<<"              PRINT NET              ";
  cout<<endl<<"===================================";

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR NET "<<name;
  cout<<endl<<"SOURCE LOGIC BLOCK IS "<<source->get_lBlock()->get_name();

  cout<<endl<<"SINK LOGIC BLOCKS ARE -->";
  list<Terminal*>::iterator term_itr = sinkList.begin();
  for(;term_itr!=sinkList.end();term_itr++)
    cout<<"  "<<(*term_itr)->get_lBlock()->get_name();

  cout<<endl;

}

void Net::print_Info(FILE* fp){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  fprintf(fp,"\n\n===================================");
  fprintf(fp,"\n              PRINT NET              ");
  fprintf(fp,"\n===================================");

  fprintf(fp,"\n\nBEGIN PRINT INFO FOR NET %s",name);
  fprintf(fp,"\nSOURCE LOGIC BLOCK IS %s",source->get_lBlock()->get_name());

  fprintf(fp,"\nSINK LOGIC BLOCKS ARE -->");
  list<Terminal*>::iterator term_itr = sinkList.begin();
  for(;term_itr!=sinkList.end();term_itr++)
    fprintf(fp,"  %s",(*term_itr)->get_lBlock()->get_name());

  fprintf(fp,"\n");

}

// Get a pointer to this net's route tree
routeTree* Net::get_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  return &rTree;
}

// Get the root of the subtree that sinks at sink_rnode
treeNode* routeTree::get_subTree(routingNode* sink_rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(sink_rnode==0)
    ERROR("sink routing node cannot be NULL...");

  /* first locate the tree node that points to sink_rnode */
  treeNode* sink_tree_node=findNode(root,sink_rnode);
  if(sink_tree_node==0)
    ERROR("could not find sink routing node in this tree...");

  /* now iteratively traverse the parent of tree nodes until the root of the subtree */
  /* that sinks at sink_rnode is reached. */
  treeNode* tree_node;
  for(tree_node=sink_tree_node;tree_node->get_numChildren()<=1 &&
      tree_node->get_rNode()->get_type()!=SOURCE;tree_node=tree_node->get_parent()){ 
    ;
  }

  return tree_node;

}

// Make this route tree a copy of the tree rooted at tree_root.
// Special case, because we want only the tree branch between tree_root and
// the tree node that points to sink_rnode
void routeTree::populateWith(treeNode* tree_root,routingNode* sink_rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(tree_root==0 || sink_rnode==0)
    ERROR("route tree's root cannot be NULL...");

  treeNode* tree_node=findNode(tree_root,sink_rnode);
  treeNode* tree_node_copy;treeNode* prev_tree_node=0;

  for(;tree_node->get_numChildren()<=1 && tree_node->get_rNode()->get_type()!=SOURCE;
      tree_node=tree_node->get_parent()){
    /* create a new tree node that will be a copy of tree node */
    tree_node_copy=new treeNode;
    tree_node_copy->set_rNode(tree_node->get_rNode());

    if(prev_tree_node){
      prev_tree_node->set_parent(tree_node_copy);
      tree_node_copy->addto_childList(prev_tree_node);
    }

    prev_tree_node=tree_node_copy;
  }

  tree_node_copy=new treeNode;
  tree_node_copy->set_rNode(tree_node->get_rNode());
  prev_tree_node->set_parent(tree_node_copy);
  tree_node_copy->addto_childList(prev_tree_node);

  root=tree_node_copy;

}



// Make this route tree a copy of the tree rooted at tree_root.
void routeTree::populateWith(treeNode* tree_root){

//   cout<<endl<<endl<<"IN POPULATE WITH...";

  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(tree_root==0)
    ERROR("route tree's root cannot be NULL...");

  root=makeCopy(tree_root);

//   cout<<endl<<"EXITING POPULATE WITH...";

}

// Find tree node that points to routing node rnode
treeNode* routeTree::getNode(routingNode* rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(rnode==0)
    ERROR("trying to look for NULL routing node...");

  treeNode* ret_node=findNode(root,rnode);
  if(ret_node==0){

#if defined(PLACER_DEBUG)
    cout<<endl<<endl<<"COULD NOT FIND "<<rnode->name;
    print_info();
#endif

  }

  return ret_node;
}

// Recursively build a copy of this route tree
treeNode* routeTree::makeCopy(treeNode* r){
  treeNode* child;treeNode* child_copy;
  treeNode* tree_node=r;treeNode* tree_node_copy;

  /* create a new tree node that will be a copy of tree node */
  tree_node_copy=new treeNode;
  tree_node_copy->set_rNode(tree_node->get_rNode());

//   cout<<endl<<"\tcreated "<<tree_node_copy->get_rNode()->get_name();

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    /* recursively create a copy of the subtree rooted at this node */
    child_copy=makeCopy(child);
    tree_node_copy->addto_childList(child_copy);
    child_copy->set_parent(tree_node_copy);
  }
  tree_node->remove_childListItr();

  /* return the newly created copy of tree node */
  return tree_node_copy;

}

// Clean the routing nodes in this tree. This is a post-BFS reinit operation.
void routeTree::clean_rNodes(treeNode* r){
  treeNode* tree_node=r; treeNode* child;
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");

  tree_node->get_rNode()->isExplored=false;
  tree_node->get_rNode()->inrTree=false;
  tree_node->get_rNode()->predecessor=0;
  tree_node->get_rNode()->pathCost=HUGE_NUMBER;
  tree_node->get_rNode()->isRouted=false;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    clean_rNodes(child);
  tree_node->remove_childListItr();

}

// Set the root of this tree
void routeTree::set_root(treeNode* rt){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
//   if(rt==0)
//     ERROR("route tree's root cannot be NULL...");
  root=rt;
}

// Get the root of this tree
treeNode* routeTree::get_root(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  return root;
}

// Print this tree and info relating to it 
void routeTree::print_info(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(root==0)
    ERROR("trying to print NULL tree here...");
  cout<<endl<<"PRINT ROUTE TREE -->"<<endl;
  preorderTraverse(root,0);
  cout<<endl<<endl;
}

void routeTree::print_info(FILE* fp){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(root==0)
    ERROR("trying to print NULL tree here...");
  fprintf(fp,"\nPRINT ROUTE TREE -->\n");
  preorderTraverse(root,0,fp);
  fprintf(fp,"\n\n");
}

// Clean up this routing tree
void routeTree::clear(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  postorderTraverse(root,1);
  root=0;
}

// Increment the occupancies of the routing nodes that constitute this tree
void routeTree::incrementOccupancies(int how){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(how!=0 && how!=1)
    ERROR("spurious parameter value here...");

  if(how==0)
    postorderTraverse(root,3);
  else
    postorderTraverse(root,4);
}

// Traverse this tree preorder. The actions to be taken at each node are specified by do_what
void routeTree::preorderTraverse(treeNode* r,int do_what){
  treeNode* child;
  treeNode* tree_node=r;

  if(do_what==0)
    /* print this tree node */
    cout<<endl<<tree_node->get_rNode()->get_name();

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    preorderTraverse(child,do_what);
  tree_node->remove_childListItr();

}

void routeTree::preorderTraverse(treeNode* r,int do_what,FILE* fp){
  treeNode* child;
  treeNode* tree_node=r;

  if(do_what==0)
    /* print this tree node */
    fprintf(fp,"\n%s",tree_node->get_rNode()->get_name());

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    preorderTraverse(child,do_what,fp);
  tree_node->remove_childListItr();

}

// Traverse this tree postorder. The action to be taken at each node are specified by do_what
void routeTree::postorderTraverse(treeNode* r,int do_what){
  treeNode* child;
  treeNode* tree_node=r;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    postorderTraverse(child,do_what);
  tree_node->remove_childListItr();

  if(do_what==0)
    /* print this tree node */
    cout<<endl<<tree_node->get_rNode()->get_name()<<"  "<<tree_node->get_pathCost();
  else if(do_what==1)
    /* delete this tree node */
    delete tree_node;
  else if(do_what==2){
    /* change sharing and wire cost to refelct the deletion of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost-=1;
    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
      wireCost-=1;

//     wireCost-=1;

    /* decrement occupancy */
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()-1);

#if defined(PLACER_DEBUG)
    cout<<endl<<"\tdecrement and delete "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy();
#endif

    /* delete this tree node */
    delete tree_node;
  }
  else if(do_what==3){
    /* increment occupancy of all routing nodes*/
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
    /* change sharing and wire cost to reflect the addition of this node */
    if(tree_node->get_rNode()->get_occupancy()>tree_node->get_rNode()->get_capacity())
      sharingCost+=1;

//     wireCost+=1;

    if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
      wireCost+=1;
  }
  else if(do_what==4){
    /* increment occupancy of non source/sink routing nodes*/
    if(tree_node->get_rNode()->get_type()!=SOURCE && tree_node->get_rNode()->get_type()!=SINK)
      tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
  }
  else if(do_what==5)
    tree_node->get_rNode()->set_counted(false);


}

// Find rnode in this tree and return the relevant tree node
treeNode* routeTree::findNode(treeNode* r,routingNode* rnode){
  treeNode* tree_node=r;treeNode* ret_node=0;
  treeNode* child;

//   cout<<endl<<"this is tree node "<<tree_node->get_rNode()->get_name();

  if(tree_node->get_rNode()==rnode){
    /* found rnode...return this node */
    return tree_node;
  }

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    ret_node=findNode(child,rnode);
    if(ret_node){
      tree_node->remove_childListItr();
      return ret_node;
    }
  }
  tree_node->remove_childListItr();

  /* rnode does not exist in any node in the sub-tree rooted at tree_node */
  return 0;

}

// Return tree node that holds the routing node pointed to by rnode
treeNode* Net::find_treeNode(routingNode* rnode){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(rnode==0)
    ERROR("cant find NULL routing node in tree...");
  treeNode* tree_node=rTree.findNode(rTree.get_root(),rnode);
  if(tree_node==0)
    ERROR("uh,oh! could not find routing node in tree...");

//   cout<<endl<<"FOUND tree node corresponding to "<<tree_node->get_rNode()->get_name();

  return tree_node;
}

