////////////////////////////////////////////////////////////////////
// Terminal.cc : Definition of Terminal member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "Terminal.h"

// Give this terminal a name
void Terminal::set_name(char* n){
  if(!n)
    ERROR("terminal name cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Set logic block associated with this terminal
void Terminal::set_lBlock(logicBlock* l){
  if(!l)
    ERROR("logic block cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  lBlock=l;
  
}

// Map this terminal to a routing node
void Terminal::set_rNode(routingNode* r){
  if(!r)
    ERROR("routing node cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  rNode=r;
}

// Net that this terminal drives or is driven by
void Terminal::set_net(Net* n){
  if(!n)
    ERROR("net cannot be NULL...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  net=n;
}

// Set the type of this terminal
void Terminal::set_terminalType(int t){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  terminalType=t;
}

// Whether this is a sink terminal of net
void Terminal::set_isSink(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  isSink=u;
}

// Whether this is the source terminal of net
void Terminal::set_isSource(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  isSource=u;
}

// Get name of this terminal
char* Terminal::get_name(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Get logic block associated with this terminal
logicBlock* Terminal::get_lBlock(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return lBlock;
}

// Get net that this terminal drives or is driven by
Net* Terminal::get_net(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return net;
}

// Get net that this terminal drives or is driven by
routingNode* Terminal::get_rNode(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return rNode;
}

// Get terminal type
int Terminal::get_terminalType(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return terminalType;
}

// Whether this is a sink terminal
bool Terminal::get_isSink(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return isSink;
}

// Whether this is a source terminal
bool Terminal::get_isSource(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");
  return isSource;
}

// Print details about this terminal
void Terminal::print_Info(){
  if(magicNum!=16111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"\tBEGIN PRINT INFO FOR TERMINAL "<<name<<" OF TYPE "<<terminalType;

  if(isSink){
    cout<<endl<<"\tSINK TERMINAL OF NET "<<net->get_name();
    cout<<endl<<"\tINPUT TERMINAL OF "<<lBlock->get_name();
  }
  if(isSource){
    cout<<endl<<"\tSOURCE TERMINAL OF NET "<<net->get_name();
    cout<<endl<<"\tOUTPUT TERMINAL OF "<<lBlock->get_name();
  }

}


