////////////////////////////////////////////////////////////////////
// blifClb.h: Class representing a blif logic block.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition of a clb as found in a blif netlist. For our
purposes, each clb is a cluster of one or more 4-LUT/DFF pairs. 
****************************************************************/ 
#ifndef BLIF_CLB_H
#define BLIF_CLB_H

#include "error.h"
#include "blifNet.h"

class blifNet;

class blifClb{
 public:
  char name [MAX_STR_LEN];  //name of clb
  list<blifNet*> inputList;  //list of clb inputs
  list<blifNet*> outputList;  //list of clb outputs
  list<blifNet*>::iterator* input_itr;
  list<blifNet*>::iterator* output_itr;
  bool isClb;  //whether this is a clb
  bool isOutputBlock;  //whether this is an output block
  bool isInputBlock;  //whether this is an input block
  bool isLatch;  //whether this is a blif latch
  int numInputNets;  //num nets that fanin to this clb
  int numOutputNets;  //num nets that fanin to this clb
  int magicNum;

  blifClb(){
    isClb=isInputBlock=isOutputBlock=isLatch=false;
    numInputNets=numOutputNets=0;
    input_itr=0;output_itr=0;
    magicNum=12111;
  }

  ~blifClb(){
    if(magicNum!=12111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //set name of clb
  void addto_inputList(blifNet* p_net);  //add a net to inputList
  void addto_outputList(blifNet* p_net);  //add a net to outputList
  void set_isClb(bool u);  //set type to clb
  void set_isOutputBlock(bool u);  //set type to output block
  void set_isInputBlock(bool u);  //set type to input block
  void create_inputListItr();  //create a new iterator for inputList
  void remove_inputListItr();  //remove an already existing inputList iterator 
  void create_outputListItr();  //create a new iterator for outputList
  void remove_outputListItr();  //remove an already existing outputList iterator 

  char* get_name();  //get name of clb
  blifNet* getnext_inputList();  //get next element from inputList
  blifNet* getnext_outputList();  //get next element from outputList
  bool get_isClb();  //check whether this is a clb
  bool get_isOutputBlock();  //check whether this is an output block
  bool get_isInputBlock();  //check whether this is an input block
  int get_numInputNets();  //get num clb input nets
  int get_numOutputNets();  //get num clb output nets

  void print_ClbInfo();  //print info about this clb
};

#endif

