/**********************************************************************
blifLexer.l: Lex-file for parsing input files.
Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
University of Washington.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

The GNU General Public License can be found at 
http://www.gnu.org/licenses/gpl.txt
***********************************************************************/

%{
#include <strings.h>
#include <iostream.h>
#include "hsraMain.tab.h"
%}

%%
"("                     {}
")"                     {}
","                     {}
(-?[0-9]+\.[0-9]+)([eE][-+]?[0-9]+)?  {yylval.double_num=atof(yytext); return DOUBLENUM_TOK; }
".model"                            {yylval.word=strdup(yytext); return MODEL_TOK; }
".inputs"                           {yylval.word=strdup(yytext); return INPUTS_TOK; }
".outputs"                          {yylval.word=strdup(yytext); return OUTPUTS_TOK; }
".wire_load_slope"                  {yylval.word=strdup(yytext); return SLOPE_TOK; }
".latch"                            {yylval.word=strdup(yytext); return LATCH_TOK; }
".names"                            {yylval.word=strdup(yytext); return NAMES_TOK; }
".end"                              {yylval.word=strdup(yytext); return END_TOK; }
"NUM_CLUSTERS"                      {yylval.word=strdup(yytext); return NUM_CLUSTERS_TOK; }
"NUM_SINKS"                         {yylval.word=strdup(yytext); return NUM_SINKS_TOK; }
^[0123456789-]+                     {yylval.word=strdup(yytext); return LOGIC_TOK; }
"["[a-zA-Z0-9_:<>\.]+"]"            { /* printf("found...%s\n",yytext); */
                                      yylval.word=strdup(yytext); return NET_TOK; }
[-?a-zA-Z0-9_:<>\.]+                  { /* printf("found...%s\n",yytext); */
                                      yylval.word=strdup(yytext); return NET_TOK; }
[\t\n ]+                            { /* printf("found whitespace...%s\n",yytext); */}
"\\"                                { ; }
%%

void yyerror(char* s){
  fprintf(stderr,"%s. Erroneous token is %s\n",s,yytext);
}

