////////////////////////////////////////////////////////////////////
// blifNet.cc: Definition of blifNet member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "blifNet.h"

// Set name of net
void blifNet::set_name(char* n){
  if(!n || strlen(n)>MAX_STR_LEN)
    ERROR("error in naming net...");
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Add a clb to sinkList
void blifNet::addto_sinkList(blifClb* p_clb){
  if(!p_clb)
    ERROR("cannot add NULL sink...");
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  sinkList.push_back(p_clb);
  numSinks++;
}

// Set source of net
void blifNet::set_source(blifClb* p_clb){
  if(!p_clb)
    ERROR("source cannot be NULL...");
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  source=p_clb;
}

// Get name of net
char* blifNet::get_name(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Get source of net
blifClb* blifNet::get_source(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  return source;
}

// Get num of sink clbs
int blifNet::get_numSinks(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  return sinkList.size();
}

// Print info about this net. This includes name of net, source clb and sink
// clbs.
void blifNet::print_NetInfo(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR NET "<<name;

  //print source
  cout<<endl<<"SOURCE: "<<source->get_name();

  //print sinks
  cout<<endl<<"SINK LIST -->";
  list<blifClb*>::iterator sink_list_itr = sinkList.begin();
  for(;sink_list_itr!=sinkList.end();sink_list_itr++)
    cout<<endl<<"\t"<<(*sink_list_itr)->get_name();

}

// Create a new iterator for sinkList
void blifNet::create_sinkListItr(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  if(sinkListItr)
    ERROR("sink list iterator already exists...");
  sinkListItr=new list<blifClb*>::iterator;
  *sinkListItr = sinkList.begin();
}

// Remove an already existing sinkList iterator 
void blifNet::remove_sinkListItr(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");
  if(sinkListItr==0)
    ERROR("trying to delete NULL sink list iterator...");
  delete sinkListItr;
  sinkListItr=0;
}

// Get next element from sinkList
blifClb* blifNet::getnext_sinkList(){
  if(magicNum!=13111)
    ERROR("trying to access a deleted node...");

  if(*sinkListItr == sinkList.end())
    return 0;
  else{
    blifClb* ret_ptr = **sinkListItr;
    (*sinkListItr)++;
    return ret_ptr;
  }

}

