////////////////////////////////////////////////////////////////////
// blifNet.h: Class that describes a net in a blif netlist.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition of a net as found in an blif netlist. Each net
has a clb source and possibly multiple clb sinks. Note that input
and output blocks are considered clbs here.
****************************************************************/ 
#ifndef BLIF_NET_H
#define BLIF_NET_H

#include "error.h"
#include "blifClb.h"

class blifClb;

class blifNet{
  char name [MAX_STR_LEN];  //name of net
  list<blifClb*> sinkList;  //list of sink clbs
  list<blifClb*>::iterator* sinkListItr;  //iterator for sinkList
  blifClb* source;  //source clb
  int numSinks;  //num sink clbs
  int magicNum;
 public:
  blifNet(){
    source=0;
    numSinks=0;
    magicNum=13111;
    sinkListItr=0;
  }

  ~blifNet(){
    if(magicNum!=13111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //set name of net
  void addto_sinkList(blifClb* p_clb);  //add a clb to sinkList
  void set_source(blifClb* p_clb);  //set source of net

  void create_sinkListItr();
  blifClb* getnext_sinkList();
  void remove_sinkListItr();

  char* get_name();  //get name of net
  blifClb* get_source();  //get source of net
  int get_numSinks();  //get num of sink clbs

  void print_NetInfo();  //print info about this net
};

#endif


