////////////////////////////////////////////////////////////////////
// blifNetlist.cc: Definitions of blifNetlist member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "blifNetlist.h"

// Set name of blif netlist to n.
void blifNetlist::set_name(char* n){
  if(!n || strlen(n)>MAX_STR_LEN)
    ERROR("error in naming netlist...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Add p_clb to list of clbs.
// This function assumes that p_clb has been updated
// to reflect whether it is a logic, input-only or
// output-only block. 
void blifNetlist::addto_blifClbList(blifClb* p_clb){
  if(!p_clb)
    ERROR("clb cannot be NULL...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  blifClbList.push_back(p_clb);

  if(p_clb->get_isClb())
    numClbs++;
  else if(p_clb->get_isInputBlock())
    numInputs++;
  else if(p_clb->get_isOutputBlock())
    numOutputs++;
}

// Add p_net to list of nets and increment num_nets.
void blifNetlist::addto_blifNetList(blifNet* p_net){
  if(!p_net)
    ERROR("net cannot be NULL...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  blifNetList.push_back(p_net);
  numNets++;
}

// Get total num clbs in this netlist.
int blifNetlist::get_numClbs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numClbs;
}

// Get total num input-only IO blocks in this netlist.
int blifNetlist::get_numInputs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get total num output-only IO blocks in this netlist.
int blifNetlist::get_numOutputs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Get total num nets in this netlist.
int blifNetlist::get_numNets(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numNets;
}

// Get the name of this netlist
char* blifNetlist::get_name(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Print details of each clb in ClbList.
void blifNetlist::print_blifClbList(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT CLB LIST";
  cout<<endl<<"***************************************************************";

  list<blifClb*>::iterator clb_itr = blifClbList.begin();
  for(;clb_itr!=blifClbList.end();clb_itr++)
    (*clb_itr)->print_ClbInfo();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT CLB LIST";
  cout<<endl<<"***************************************************************";

}

// Print details of each net in NetList.
void blifNetlist::print_blifNetList(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT NET LIST";
  cout<<endl<<"***************************************************************";

  list<blifNet*>::iterator net_itr = blifNetList.begin();
  for(;net_itr!=blifNetList.end();net_itr++)
    (*net_itr)->print_NetInfo();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT NET LIST";
  cout<<endl<<"***************************************************************";

}

// Print the name of this blif netlist.
void blifNetlist::print_name(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  cout<<endl<<endl<<"NAME OF NETLIST IS "<<name;
}

// Print the details of clb pointed to by p_clb.
void blifNetlist::print_blifClb(blifClb* p_clb){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!p_clb)
    ERROR("clb cannot be NULL...");
  p_clb->print_ClbInfo();
}

// Print the details of net pointed to by p_net.
void blifNetlist::print_blifNet(blifNet* p_net){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!p_net)
    ERROR("net cannot be NULL...");
  p_net->print_NetInfo();
}

// Look for clb with clb_name in ClbList. If found, return the clb.
blifClb* blifNetlist::found_blifClb(char* clb_name){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!clb_name)
    ERROR("clb name cannot be NULL...");
  list<blifClb*>::iterator clb_itr = blifClbList.begin();
  for(;clb_itr!=blifClbList.end();clb_itr++){
    if(strcmp((*clb_itr)->get_name(),clb_name)==0)
      return *clb_itr;
  }
  return 0;
}

// Look for net with net_name in NetList. If found, return the net.
blifNet* blifNetlist::found_blifNet(char* net_name){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!net_name)
    ERROR("net name cannot be NULL...");
  list<blifNet*>::iterator net_itr = blifNetList.begin();
  for(;net_itr!=blifNetList.end();net_itr++){
    if(strcmp((*net_itr)->get_name(),net_name)==0)
      return *net_itr;
  }
  return 0;
}

// Convert this blif netlist to a generic netlist.
// This function will have to be defined and called by a user routine.
// The objective here is to convert the user's netlist data-structs
// to the generic format that Independence will work with.
void blifNetlist::convert2Generic(Netlist& n){
  Net* out_net;
  Net* in_net;
  blifNet* blif_net;blifNet* latch_net;blifNet* latch_input_net;
  blifClb* sink_clb;
  Terminal* output_terminal;
  Terminal* input_terminal;
  logicBlock* lb;logicBlock* sink_lb;
  int y;

  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  // give the generic netlist a name
  n.set_name(name);

  // for each clb, create a new logic block in generic netlist
  // ignore latches, because HSRA does not place these
  list<blifClb*>::iterator clb_itr = blifClbList.begin();
  for(;clb_itr!=blifClbList.end();clb_itr++){
    if((*clb_itr)->isLatch==true){
      ;
    }
    else{
      lb = new logicBlock;
      // give the logic block a name
      lb->set_name((*clb_itr)->get_name());
      // determine whether we have an IO or logic block
      if((*clb_itr)->get_isClb()){
	lb->set_isLogic(true);
	lb->set_functionType(BLIF_CLB_TYPE);
      }
      if((*clb_itr)->get_isOutputBlock()){
	lb->set_isOutput(true);
	lb->set_functionType(BLIF_IO_TYPE);
      }
      if((*clb_itr)->get_isInputBlock()){
	lb->set_isInput(true);
	lb->set_functionType(BLIF_IO_TYPE);
      }
      // add lb to list of logicBlocks in n
      n.addto_lbList(lb);
    }

  }

  // now, create nets and terminals
  clb_itr = blifClbList.begin();
  for(;clb_itr!=blifClbList.end();clb_itr++){
    // find logic block in generic netlist corresponding 
    // to this blifClb.
    lb = n.found_logicBlock((*clb_itr)->get_name());
    // if the clb is a latch, do nothing
    if(lb==0)
      continue;

    // loop thru the input nets of the blif clb
    (*clb_itr)->create_inputListItr();
    blif_net=(*clb_itr)->getnext_inputList();
    for(y=0;blif_net!=0;blif_net=(*clb_itr)->getnext_inputList(),y++){
      // if blif_net is sourced at a latch,
      // get the latch's input net
      if(blif_net->get_source()->isLatch==true){
	blif_net->get_source()->create_inputListItr();
	latch_input_net=blif_net->get_source()->getnext_inputList();
	blif_net->get_source()->remove_inputListItr();
	in_net = n.found_Net(latch_input_net->get_name());
	if(!in_net){
	  in_net=createNet(latch_input_net);
	  n.addto_netList(in_net);
	}
      }
      else{
	in_net = n.found_Net(blif_net->get_name());
	if(!in_net){
	  in_net = createNet(blif_net);
	  n.addto_netList(in_net);
	}
      }

      // create a new input terminal
      input_terminal = createTerminal(lb,in_net,BLIF_INPUT_TERMINAL_TYPE,false,y);
      lb->addto_inputList(input_terminal);
      in_net->addto_sinkList(input_terminal);

    }
    (*clb_itr)->remove_inputListItr();

    // loop thru the output nets of the blif clb
    (*clb_itr)->create_outputListItr();
    blif_net=(*clb_itr)->getnext_outputList();
    for(y=0;blif_net!=0;blif_net=(*clb_itr)->getnext_outputList(),y++){
      // find output net
      out_net = n.found_Net(blif_net->get_name());
      if(!out_net){
	out_net = createNet(blif_net);
	n.addto_netList(out_net);
      }
      // create a new output terminal
      output_terminal = createTerminal(lb,out_net,BLIF_OUTPUT_TERMINAL_TYPE,true,y);
      lb->addto_outputList(output_terminal);
      out_net->set_source(output_terminal);
    }
    (*clb_itr)->remove_outputListItr();

  }

  /* now create vector of logic blocks */
  n.create_lbVector();

  /* for each net, create vector of sinks, and populate */
  n.create_netSinkVectors();

}

// this function creates a new Net object based on the blifNet pointed
// to by n. 
Net* blifNetlist::createNet(blifNet* n){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(n==0)
    ERROR("NULL blif net here...");
  Net* new_net = new Net;
  new_net->set_name(n->get_name());
  return new_net;
}

// create a new Terminal object. this object represents a terminal of the logicBlock
// lb which is a source / sink of tnet. term_type is a user-defined terminal type,
// is_source indicates whether this terminal is a source or sink of tnet, and 
// term_num is used to derive the name of this terminal.
Terminal* blifNetlist::createTerminal(logicBlock* lb,Net* tnet,int term_type,
                                           bool is_source,int term_num){
  char term_name [MAX_STR_LEN];
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(lb==0)
    ERROR("NULL logic block...");
  if(tnet==0)
    ERROR("NULL net here...");
  if(term_type!=BLIF_INPUT_TERMINAL_TYPE && term_type!=BLIF_OUTPUT_TERMINAL_TYPE)
    ERROR("faulty terminal types...");

  Terminal* new_terminal = new Terminal;

  if(term_type==BLIF_INPUT_TERMINAL_TYPE)
    sprintf(term_name,"%s_INPUT%d",lb->get_name(),term_num);
  else
    sprintf(term_name,"%s_OUTPUT%d",lb->get_name(),term_num);

  new_terminal->set_name(term_name);
  new_terminal->set_lBlock(lb);
  new_terminal->set_net(tnet);
  new_terminal->set_terminalType(term_type);
  if(is_source)
    new_terminal->set_isSource(true);
  else
    new_terminal->set_isSink(true);
  return new_terminal;

}


