////////////////////////////////////////////////////////////////////
// blifNetlist.h: Class that describes a blif netlist.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition for a blif netlist. Note that this class 
definition is specific to blif netlists. For every new benchmark
set, a separate class definition will have to be created and
implemented.
****************************************************************/ 
#ifndef BLIF_NETLIST_H
#define BLIF_NETLIST_H

#include "error.h"
#include "blifClb.h"
#include "blifNet.h"

/* include generic netlist here so that member functions of blifNetlist */
/* can populate the generic netlist */
#include "Netlist.h"

class blifNetlist{
  char name [MAX_STR_LEN];  //name of blif netlist
  list<blifClb*> blifClbList;  //list of clbs in blif netlist
  list<blifNet*> blifNetList;  //list of nets in blif netlist
  int numClbs;  //num clbs in blif netlist
  int numInputs;  //num input-only IO blocks
  int numOutputs;  //num output-only IO blocks
  int numNets;  //num nets in blif netlist
  int magicNum;
 public:
  blifNetlist(){
    numClbs=0;
    numNets=0;
    numInputs=numOutputs=0;
    magicNum=11111;
  }

  ~blifNetlist(){
    if(magicNum!=11111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //set the name of blif netlist
  void addto_blifClbList(blifClb* p_clb);  //add a clb to blifClbList
  void addto_blifNetList(blifNet* p_net);  //add a net to blifNetList

  int get_numClbs();  //get num clbs in blif netlist
  int get_numNets();  //get num nets in blif netlist
  char* get_name();  //get the name of blif netlist
  int get_numInputs();  //total num input-only IO blocks
  int get_numOutputs();  //total num output-only IO blocks

  void print_blifClbList();  //print the entire list of clbs
  void print_blifNetList();  //print the entire list of nets
  void print_name();  //print the name of blif netlist
  void print_blifClb(blifClb* p_clb);  //print the details of clb pointed to by p_clb
  void print_blifNet(blifNet* p_net);    //print the details of net pointed to by p_net

  blifClb* found_blifClb(char* clb_name);  //look for clb in ClbList. if found, return clb.
  blifNet* found_blifNet(char* net_name);  //look for net in NetList. if found, return net.

  void convert2Generic(Netlist& n);  //convert this blif netlist to a generic netlist
  Net* createNet(blifNet* n);  //create a new Net object based on the blifNet pointed
                               //to by n.
  Terminal* createTerminal(logicBlock* lb,Net* tnet,int term_type,bool is_source,
                           int term_num);  //create a new Terminal object


};

#endif


