////////////////////////////////////////////////////////////////////
// hsraArch.cc : Definition of the member functions of classes that
//               describe the HSRA architecture.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "hsraArch.h"

hsraBlock::hsraBlock(int p, int t){
  position=p;type=t;
  magicNum=21111;
  sourceNode=sinkNode=0;
  ipin_itr=opin_itr=0;
  currBlock=0;
}

hsraBlock::~hsraBlock(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;
}

int hsraBlock::get_position(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return position;
}

int hsraBlock::get_type(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return type;
}

void hsraBlock::set_sourceNode(routingNode* r){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("source routing node cannot be NULL...");
  sourceNode=r;
}

void hsraBlock::set_sinkNode(routingNode* r){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("sink routing node cannot be NULL...");
  sinkNode=r;
}

void hsraBlock::set_currBlock(logicBlock* l){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  currBlock=l;
}

void hsraBlock::addto_inputPinList(routingNode* r){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("input routing node cannot be NULL...");
  inputPinList.push_back(r);
}

void hsraBlock::addto_outputPinList(routingNode* r){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("output routing node cannot be NULL...");
  outputPinList.push_back(r);
}

void hsraBlock::create_inputPinListItr(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(ipin_itr)
    ERROR("an iterator already exists...");
  ipin_itr = new list<routingNode*>::iterator;
  *ipin_itr = inputPinList.begin();
}

void hsraBlock::create_outputPinListItr(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(opin_itr)
    ERROR("an iterator already exists...");
  opin_itr = new list<routingNode*>::iterator;
  *opin_itr = outputPinList.begin();
}

void hsraBlock::remove_inputPinListItr(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!ipin_itr)
    ERROR("there is no iterator to remove...");
  delete ipin_itr;
  ipin_itr = 0;
}

void hsraBlock::remove_outputPinListItr(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!opin_itr)
    ERROR("there is no iterator to remove...");
  delete opin_itr;
  opin_itr = 0;
}

routingNode* hsraBlock::getnext_inputPinList(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!ipin_itr)
    ERROR("NULL iterator here...");

  if(*ipin_itr == inputPinList.end())
    return 0;
  else{
    routingNode* ret_ptr = **ipin_itr;
    (*ipin_itr)++;
    return ret_ptr;
  }

}

routingNode* hsraBlock::getnext_outputPinList(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  if(!opin_itr)
    ERROR("NULL iterator here...");

  if(*opin_itr == outputPinList.end())
    return 0;
  else{
    routingNode* ret_ptr = **opin_itr;
    (*opin_itr)++;
    return ret_ptr;
  }

}

logicBlock* hsraBlock::get_currBlock(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return currBlock;
}

routingNode* hsraBlock::get_sourceNode(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return sourceNode;
}

routingNode* hsraBlock::get_sinkNode(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return sinkNode;
}

void hsraBlock::print_info(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");

  if(type==HSRA_CLB)
    cout<<endl<<"HSRA CLB AT POSITION "<<position;
  else
    cout<<endl<<"HSRA IOB AT POSITION "<<position;

  if(currBlock)
    cout<<endl<<"LOGIC BLOCK : "<<currBlock->get_name();
  else
    cout<<endl<<"LOGIC BLOCK : ----";

  cout<<endl<<"SOURCE NODE : "<<sourceNode->get_name()<<", SINK NODE : "<<sinkNode->get_name();

  list<routingNode*>::iterator input_itr = inputPinList.begin();
  cout<<endl<<"INPUT ROUTING NODES  :";
  for(;input_itr!=inputPinList.end();input_itr++)
    cout<<"  "<<(*input_itr)->get_name();

  list<routingNode*>::iterator output_itr = outputPinList.begin();
  cout<<endl<<"OUTPUT ROUTING NODES  :";
  for(;output_itr!=outputPinList.end();output_itr++)
    cout<<"  "<<(*output_itr)->get_name();

}

/******************************************************************/

switchBox::switchBox(int num_switches,int switch_type){
  numSwitches=num_switches;
  switchType=switch_type;
  LArray=new vector<routingNode*>(numSwitches);
  RArray=new vector<routingNode*>(numSwitches);
  for(int i=0;i<numSwitches;i++)
    LArray->at(i)=RArray->at(i)=0;
  magicNum=102938;
}

int switchBox::get_numSwitches(){
  if(magicNum!=102938)
    ERROR("trying to access a deleted node...");
  return numSwitches;
}

int switchBox::get_switchType(){
  if(magicNum!=102938)
    ERROR("trying to access a deleted node...");
  return switchType;
}

routingNode* switchBox::get_LRNodeAt(int index){
  if(magicNum!=102938)
    ERROR("trying to access a deleted node...");
  if(index<0 || index>=LArray->size())
    ERROR("array out of bounds...");
  return LArray->at(index);
}

routingNode* switchBox::get_RRNodeAt(int index){
  if(magicNum!=102938)
    ERROR("trying to access a deleted node...");
  if(index<0 || index>=RArray->size())
    ERROR("array out of bounds...");
  return RArray->at(index);
}

void switchBox::addto_LArray(routingNode* r,int index){
  if(magicNum!=102938)
    ERROR("trying to access a deleted node...");
  if(index<0 || index>=LArray->size())
    ERROR("array out of bounds...");
  if(!r)
    ERROR("adding NULL routing node...");
  LArray->at(index)=r;
}

void switchBox::addto_RArray(routingNode* r,int index){
  if(magicNum!=102938)
    ERROR("trying to access a deleted node...");
  if(index<0 || index>=RArray->size())
    ERROR("array out of bounds...");
  if(!r)
    ERROR("adding NULL routing node...");
  RArray->at(index)=r;
}

/******************************************************************/

levelStats::levelStats(int chan_width,int switch_type,int num_switch_boxes){
  channelWidth=chan_width;
  switchType=switch_type;
  numSwitchBoxes=num_switch_boxes;
  switchArray=new vector<switchBox*>(numSwitchBoxes);
  for(int i=0;i<numSwitchBoxes;i++)
    switchArray->at(i)=new switchBox(channelWidth,switchType);
  magicNum=45678;
}

int levelStats::get_channelWidth(){
  if(magicNum!=45678)
    ERROR("trying to access a deleted node...");
  return channelWidth;
}

int levelStats::get_switchType(){
  if(magicNum!=45678)
    ERROR("trying to access a deleted node...");
  return switchType;
}

int levelStats::get_numSwitchBoxes(){
  if(magicNum!=45678)
    ERROR("trying to access a deleted node...");
  return numSwitchBoxes;
}

switchBox* levelStats::get_switchBoxAt(int index){
  if(magicNum!=45678)
    ERROR("trying to access a deleted node...");
  return switchArray->at(index);
}

void levelStats::addto_switchArray(switchBox* sw,int index){
  if(magicNum!=45678)
    ERROR("trying to access a deleted node...");
  if(sw==0 || index<0 || index>=switchArray->size())
    ERROR("dubious array parameters here...");
  switchArray->at(index)=sw;
}

void levelStats::print_info(FILE* fp){
  if(magicNum!=45678)
    ERROR("trying to access a deleted node...");

  cout<<endl<<"Channel width = "<<channelWidth;
  fprintf(fp,"\nChannel width = %d",channelWidth);
  if(switchType==COMPRESSING){
    cout<<", switch type is COMPRESSING.";
    fprintf(fp,", switch type is COMPRESSING.");
  }
  else{
    cout<<", switch type is NON_COMPRESSING.";
    fprintf(fp,", switch type is NON_COMPRESSING.");
  }
  printf("\nContinuous approx of interconnect requirements = %32.24f",continuousApprox);
  fprintf(fp,"\nContinuous approx of interconnect requirements = %32.24f",continuousApprox);
  cout<<endl<<"Num switchboxes at this level = "<<numSwitchBoxes;
  fprintf(fp,"\nNum switchboxes at this level = %d",numSwitchBoxes);
}

/******************************************************************/

long int hsraFPGA::countSharing(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  int sharing=0;long int totalSharing=0;
  list<routingNode*>::iterator rnode_itr = rnodeList.begin();
  for(;rnode_itr!=rnodeList.end();rnode_itr++){
    sharing=(*rnode_itr)->get_occupancy() - (*rnode_itr)->get_capacity();
    if(sharing<0)
      sharing=0;
    totalSharing+=sharing;
    if(opt==1)
      cout<<endl<<(*rnode_itr)->get_name()<<"  "<<sharing;
    else if(opt==2){
      if(sharing>0)
	cout<<endl<<(*rnode_itr)->get_name()<<"  "<<(*rnode_itr)->get_occupancy();
    }
  }

  return totalSharing;

}

long int hsraFPGA::countSharing(FILE* fp){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  int sharing=0;long int totalSharing=0;
  list<routingNode*>::iterator rnode_itr = rnodeList.begin();
  for(;rnode_itr!=rnodeList.end();rnode_itr++){
    sharing=(*rnode_itr)->get_occupancy() - (*rnode_itr)->get_capacity();
    if(sharing<0)
      sharing=0;
    totalSharing+=sharing;
    if(sharing>0)
      fprintf(fp,"\n%s  %d",(*rnode_itr)->get_name(),(*rnode_itr)->get_occupancy());
  }

  return totalSharing;

}

void hsraFPGA::updateHistoryCosts(int iter,bool during_placement){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  list<routingNode*>::iterator rnodelist_itr = rnodeList.begin();
  int occ=0;int cap=0;double hist=0.;double new_hist=0.;
  for(;rnodelist_itr!=rnodeList.end();rnodelist_itr++){
    occ=(*rnodelist_itr)->get_occupancy();
    cap=(*rnodelist_itr)->get_capacity();
    hist=(*rnodelist_itr)->get_historyCost(iter);

    if(iter==1)
      new_hist=1.0;
    else{
      if(during_placement){
        if(occ>cap)
          new_hist=flomax(1.0,0.9*hist + 0.5);
        else
          new_hist=flomax(1.0,0.9*hist);
      }
      else{
        if(occ>cap){
	  new_hist=hist + flomax(0.,(double)(occ - cap));
//            new_hist+=1.0;
	}
        else
          new_hist=hist;
      }
    }

    (*rnodelist_itr)->historyCost=new_hist;

  }

}

void hsraFPGA::print_info(FILE* fp){
  int i;
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  cout<<endl<<endl<<"********************* BEGIN PRINT FPGA ARCHITECTURE ***********************";
  cout<<endl<<"Num CLBs = "<<numCLBs<<", and num IOBs = "<<numIOBs;
  cout<<endl<<"Base chan width = "<<baseChanWidth<<", and growth rate = "<<growthRate;
  cout<<endl<<"Num interconnect levels = "<<numLevels;
  cout<<endl<<"Routing Graph:  V = "<<rnodeList.size()<<",  E = "<<redgeList.size();

  fprintf(fp,"\n\n********************* BEGIN PRINT FPGA ARCHITECTURE ***********************");
  fprintf(fp,"\nNum CLBs = %d, and num IOBs = %d",numCLBs,numIOBs);
  fprintf(fp,"\nBase chan width = %d, and growth rate = %f",baseChanWidth,growthRate);
  fprintf(fp,"\nNum interconnect levels = %d",numLevels);
  fprintf(fp,"\nRouting Graph:  V = %d,  E = %d",rnodeList.size(),redgeList.size());

//   for(i=0;i<numLevels;i++){
//     cout<<endl<<"Level "<<i<<": ";
//     fprintf(fp,"\nLevel %d: ",i);
//     levelArray->at(i)->print_info(fp);
//   }


   list<routingNode*>::iterator r_iter = rnodeList.begin();
   for(;r_iter!=rnodeList.end();r_iter++)
     (*r_iter)->print_info();

}

hsraFPGA::~hsraFPGA(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;
}

hsraFPGA::hsraFPGA(int num_clbs,int base_width,int num_levels,double growth_rate,int lsz){

  numCLBs=num_clbs;
  baseChanWidth=base_width;
  numLevels=num_levels;
  growthRate=growth_rate;
  magicNum=23111;
  lsize=lsz;
  numSinkNodes=0;

  // create logic blocks
  clbArray=new vector<hsraBlock*>(numCLBs);
  for(int i=0;i<numCLBs;i++)
    clbArray->at(i)=new hsraBlock(i,HSRA_CLB);

//    cout<<endl<<"Created logic blocks...";

  // provide level information
  levelArray=new vector<levelStats*>(numLevels);
  double next_chan_width;
  int curr_chan_width=baseChanWidth;
  for(int k=0;k<numLevels;k++){
    next_chan_width=((double)baseChanWidth)*pow((pow(2.0,growthRate)),k+1);
    if( (next_chan_width - ((double)curr_chan_width)) < 0.000003 ){
      levelArray->at(k)=new levelStats(curr_chan_width,COMPRESSING,int(pow(2.0,numLevels-k-1)));
      levelArray->at(k)->continuousApprox=next_chan_width;
    }
    else{
      levelArray->at(k)=new levelStats(curr_chan_width,NON_COMPRESSING,int(pow(2.0,numLevels-k-1)));
      curr_chan_width=2*curr_chan_width;
      levelArray->at(k)->continuousApprox=next_chan_width;
    }
  }

//    cout<<endl<<"Created level stats...";

  // in HSRA, there is one IO location for every track that connects
  // to the root switchbox from the outside world.  
  numIOBs=curr_chan_width;

  // create IO blocks
  iobArray=new vector<hsraBlock*>(numIOBs);
  for(int j=0;j<numIOBs;j++)
    iobArray->at(j)=new hsraBlock(j,HSRA_IOB);

//    cout<<endl<<"Created IO blocks...";

  // create the array that will hold the routing nodes that connect IOBs
  // to the interconnect structure
  ioWireArray=new vector<routingNode*>(numIOBs);

//    cout<<endl<<"Created IO wire array...";

  // create routing nodes
  create_routingNodes();

//    cout<<endl<<"Created routing nodes...";

  // connect routing nodes with edges
  create_routingEdges();

  // finally, populate the LRLists of CLB input_pins and sinks
  populate_CLB_LRLists();

//    cout<<endl<<"Created routing edges...";

//    list<routingNode*>::iterator r_itr = rnodeList.begin();
//    for(;r_itr!=rnodeList.end();r_itr++){
//      cout<<endl;
//      (*r_itr)->print_LRList();
//    }
 
}

void hsraFPGA::populate_CLB_LRLists(){
  hsraBlock* clb;
  routingNode* output_rnode;routingNode* wire_rnode;
  routingNode* input_rnode;
  routingEdge* fanout_edge;
  list<char>::iterator str_itr;

  for(int i=0;i<clbArray->size();i++){
    clb=clbArray->at(i);

    // get an output pin
    clb->create_outputPinListItr();
    output_rnode=clb->getnext_outputPinList();
    clb->remove_outputPinListItr();

    // get a wire that the output pin connects to
    output_rnode->create_fanoutListItr();
    fanout_edge=output_rnode->getnext_fanoutList();
    for(;fanout_edge!=0;fanout_edge=output_rnode->getnext_fanoutList()){
      if(fanout_edge->destNode->get_type()!=BUS_OR_WIRE)
	ERROR("dubious fanout node here...");
      wire_rnode=fanout_edge->destNode;
      break;
    }
    output_rnode->remove_fanoutListItr();

    // now copy the wire's LRList to this CLB's routing nodes
    str_itr=(wire_rnode->LRList).begin();
    for(;str_itr!=(wire_rnode->LRList).end();str_itr++){
      // copy LR info to each input pin
      clb->create_inputPinListItr();
      input_rnode=clb->getnext_inputPinList();
      for(;input_rnode!=0;input_rnode=clb->getnext_inputPinList()){
	(input_rnode->LRList).push_back(*str_itr);
	input_rnode->lrSize = input_rnode->LRList.size();
      }
      clb->remove_inputPinListItr();


      // copy LR info to each output pin
      clb->create_outputPinListItr();
      output_rnode=clb->getnext_outputPinList();
      for(;output_rnode!=0;output_rnode=clb->getnext_outputPinList()){
	(output_rnode->LRList).push_back(*str_itr);
	output_rnode->lrSize = output_rnode->LRList.size();
      }
      clb->remove_outputPinListItr();

      //copy LR info to sink
      (clb->get_sinkNode()->LRList).push_back(*str_itr);

      //copy LR info to source
      (clb->get_sourceNode()->LRList).push_back(*str_itr);

    }
    clb->get_sinkNode()->lrSize = clb->get_sinkNode()->LRList.size();
    clb->get_sourceNode()->lrSize = clb->get_sourceNode()->LRList.size();
  }

}

void hsraFPGA::create_routingNodes(){
  int i,j,k;
  char rname [MAX_STR_LEN];
  routingNode* rnode;
  switchBox* sbox;
  int sink_id=0;

  if(magicNum!=23111)
    ERROR("trying to access a deleted node...");

  // first create routing nodes for CLBs
  for(i=0;i<numCLBs;i++){
    sprintf(rname,"CLB%d_SOURCE",i);
    rnode=new routingNode(rname,SOURCE,0.,0.,0.,1,1.0);
    clbArray->at(i)->set_sourceNode(rnode);
    rnodeList.push_back(rnode);
    if(i>=lsize)
      rnode->canUse=false;

    sprintf(rname,"CLB%d_OUTPUT",i);
    rnode=new routingNode(rname,OUTPUT_PIN,0.,0.,0.,1,1.0);
    clbArray->at(i)->addto_outputPinList(rnode);
    rnodeList.push_back(rnode);
    if(i>=lsize)
      rnode->canUse=false;

    sprintf(rname,"CLB%d_SINK",i);
    rnode=new routingNode(rname,SINK,0.,0.,0.,4,1.0);
    ++numSinkNodes; rnode->sinkID=sink_id; sink_id++;  //for Astar table indexing purposes
    clbArray->at(i)->set_sinkNode(rnode);
    rnodeList.push_back(rnode);
    if(i>=lsize)
      rnode->canUse=false;

    for(j=0;j<4;j++){
      sprintf(rname,"CLB%d_INPUT%d",i,j);
      rnode=new routingNode(rname,INPUT_PIN,0.,0.,0.,1,1.0);
      clbArray->at(i)->addto_inputPinList(rnode);
      rnodeList.push_back(rnode);

//       cout<<endl<<"Added input-pin "<<rnode->name<<"...";

      if(i>=lsize)
	rnode->canUse=false;
    }
  }

//   list<routingNode*>::iterator r_iter = rnodeList.begin();
//   for(;r_iter!=rnodeList.end();r_iter++)
//     (*r_iter)->print_info();
//   exit(1);

  // now create routing nodes for IOBs
  sprintf(rname,"IOB_SOURCE");
  iobSource=new routingNode(rname,SOURCE,0.,0.,0.,numIOBs,1.0);
  rnodeList.push_back(iobSource);
  iobSource->LRList.push_front('X');iobSource->lrSize = iobSource->LRList.size();

  sprintf(rname,"IOB_SINK");
  iobSink=new routingNode(rname,SINK,0.,0.,0.,numIOBs,1.0);
  ++numSinkNodes; iobSink->sinkID=sink_id; sink_id++;  //for Astar table indexing purposes
  rnodeList.push_back(iobSink);
  iobSink->LRList.push_front('X');iobSink->lrSize = iobSink->LRList.size();

  for(i=0;i<numIOBs;i++){
    iobArray->at(i)->set_sourceNode(iobSource);

    sprintf(rname,"IOB%d_OUTPUT",i);
    rnode=new routingNode(rname,OUTPUT_PIN,0.,0.,0.,1,1.0);
    iobArray->at(i)->addto_outputPinList(rnode);
    rnodeList.push_back(rnode);

    iobArray->at(i)->set_sinkNode(iobSink);

    sprintf(rname,"IOB%d_INPUT",i);
    rnode=new routingNode(rname,INPUT_PIN,0.,0.,0.,1,1.0);
    iobArray->at(i)->addto_inputPinList(rnode);
    rnodeList.push_back(rnode);
  }

  // create interconnect routing nodes
  int next_lsize;
  int curr_lsize=lsize;
  if(curr_lsize%2==1)
    next_lsize=curr_lsize/2+1;
  else
    next_lsize=curr_lsize/2;

  for(i=0;i<numLevels;i++){

    for(j=0;j<levelArray->at(i)->get_numSwitchBoxes();j++){
      sbox=levelArray->at(i)->get_switchBoxAt(j);
      for(k=0;k<sbox->get_numSwitches();k++){
	sprintf(rname,"LEVEL%d_SBOX%d_LWIRE%d",i,j,k);
	rnode=new routingNode(rname,BUS_OR_WIRE,0.,0.,0.,1,1.0);
	sbox->addto_LArray(rnode,k);
	rnodeList.push_back(rnode);
	populate_LRList(rnode,i,j,true);
	if(j>=next_lsize)
	  rnode->canUse=false;
//  	cout<<endl;
//  	rnode->print_LRList();

	sprintf(rname,"LEVEL%d_SBOX%d_RWIRE%d",i,j,k);
	rnode=new routingNode(rname,BUS_OR_WIRE,0.,0.,0.,1,1.0);
	sbox->addto_RArray(rnode,k);
	rnodeList.push_back(rnode);
	populate_LRList(rnode,i,j,false);
	if(j>=next_lsize)
	  rnode->canUse=false;
	if(j==(next_lsize-1) && curr_lsize%2==1)
	  rnode->canUse=false;

//  	cout<<endl;
//  	rnode->print_LRList();

      }
    }

    curr_lsize=next_lsize;
    if(curr_lsize%2==1)
      next_lsize=curr_lsize/2+1;
    else
      next_lsize=curr_lsize/2;

  }

  // finally, create IO routing nodes
  for(i=0;i<numIOBs;i++){
    sprintf(rname,"IOB%d_WIRE",i);
    rnode=new routingNode(rname,BUS_OR_WIRE,0.,0.,0.,1,1.0);
    ioWireArray->at(i)=rnode;
    (rnode->LRList).push_front('X');rnode->lrSize = rnode->LRList.size();
    rnodeList.push_back(rnode);
  }

  // create a map of routing nodes
  list<routingNode*>::iterator r_iter = rnodeList.begin();
  string sname;
  for(;r_iter!=rnodeList.end();r_iter++){
    sname = (*r_iter)->name;
    rnodeMap[sname] = *r_iter;
  }

  // create vectors of routing nodes
  int num_rnodes = rnodeList.size();
  rnodeVec = new vector<routingNode*> (num_rnodes); 
  r_iter = rnodeList.begin();int wire_ctr=0;int sink_ctr=0;
  for(int i=0;r_iter!=rnodeList.end();r_iter++,i++){
    rnodeVec->at(i) = *r_iter;
    if((*r_iter)->type==BUS_OR_WIRE)
      wire_ctr++;
    else if((*r_iter)->type==SINK)
      sink_ctr++;
  }

  sink_rnodeVec = new vector<routingNode*> (sink_ctr);
  wire_rnodeVec = new vector<routingNode*> (wire_ctr);
  r_iter = rnodeList.begin();int w=0;int s=0;
  for(;r_iter!=rnodeList.end();r_iter++){
    if((*r_iter)->type==BUS_OR_WIRE){
      wire_rnodeVec->at(w) = *r_iter;
      w++;
    }
    else if((*r_iter)->type==SINK){
      sink_rnodeVec->at(s) = *r_iter;
      s++;
    }
  }

}

void hsraFPGA::populate_LRList(routingNode* rnode,int level_num,int sbox_num,bool left){

  int curr_sbox_num=sbox_num;
  if(left)
    (rnode->LRList).push_front('L');
  else
    (rnode->LRList).push_front('R');

  for(int i=level_num+1;i<numLevels;i++){
    if((curr_sbox_num%2)==1){
      // we have an odd sbox num
      (rnode->LRList).push_front('R');
    }
    else{
      // we have an even sbox num
      (rnode->LRList).push_front('L');
    }
    curr_sbox_num/=2;
  }

  rnode->lrSize = rnode->LRList.size();

}

void hsraFPGA::create_routingEdge(routingNode* orig_node,routingNode* dest_node){
  if(magicNum!=23111)
    ERROR("trying to access a deleted node here...");

  if(orig_node->canUse==false || dest_node->canUse==false)
    return;

  routingEdge* redge=new routingEdge(orig_node,dest_node,0.,0.,0.,0.);
  orig_node->addto_fanoutList(redge);
  dest_node->addto_faninList(redge);
  redgeList.push_back(redge);
}

void hsraFPGA::create_routingEdges(){
  routingNode* source_rnode;
  routingNode* sink_rnode;
  routingNode* input_rnode;
  routingNode* output_rnode;
  routingNode* wire_rnode;
  routingEdge* redge;
  hsraBlock* clb;hsraBlock* iob;
  int i,j,y,n,k,sbox_num,wirenum;
  switchBox* sbox;switchBox* next_sbox;
  levelStats* level;levelStats* next_level;
  routingNode* lrnode;routingNode* rrnode;
  routingNode* lrnode0;routingNode* lrnode1;  
  routingNode* rrnode0;routingNode* rrnode1;
  routingNode* wire_rnode0;routingNode* wire_rnode1;

  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  // first, establish edges between CLB routing nodes
  for(i=0;i<numCLBs;i++){

    clb=clbArray->at(i);

    // connect source and output pins
    source_rnode=clb->get_sourceNode();
    clb->create_outputPinListItr();
    output_rnode=clb->getnext_outputPinList();
    for(;output_rnode!=0;output_rnode=clb->getnext_outputPinList())
      create_routingEdge(source_rnode,output_rnode);
    clb->remove_outputPinListItr();

    // connect sink and input pins
    sink_rnode=clb->get_sinkNode();
    clb->create_inputPinListItr();
    input_rnode=clb->getnext_inputPinList();
    for(;input_rnode!=0;input_rnode=clb->getnext_inputPinList())
      create_routingEdge(input_rnode,sink_rnode);
    clb->remove_inputPinListItr();

    // connect CLB input pins to base channel tracks.
    // this is the n-choose-k strategy as per DeHon's
    // FPGA '96 Entropy and Counting paper.
    // here we assume that the base channel width is
    // strictly greater than the number of CLB inputs.
    k=(clb->inputPinList).size();n=baseChanWidth;
    clb->create_inputPinListItr();
    input_rnode=clb->getnext_inputPinList();
    for(j=1;input_rnode!=0;input_rnode=clb->getnext_inputPinList(),j++){
      sbox_num=i/2;
      sbox=levelArray->at(0)->get_switchBoxAt(sbox_num);
      for(y=(n-j);y>=(k-j);y--){
	if((i%2)==0)
	  wire_rnode=sbox->get_LRNodeAt(y);
	else
	  wire_rnode=sbox->get_RRNodeAt(y);
	create_routingEdge(wire_rnode,input_rnode);
      }
    }
    clb->remove_inputPinListItr();

    // connect CLB output pins to base channel tracks.
    // full connectivity assumed.
    clb->create_outputPinListItr();
    output_rnode=clb->getnext_outputPinList();
    for(;output_rnode!=0;output_rnode=clb->getnext_outputPinList()){
      sbox_num=i/2;
      sbox=levelArray->at(0)->get_switchBoxAt(sbox_num);
      for(y=0;y<levelArray->at(0)->channelWidth;y++){
	if((i%2)==0)
	  wire_rnode=sbox->get_LRNodeAt(y);
	else
	  wire_rnode=sbox->get_RRNodeAt(y);
	create_routingEdge(output_rnode,wire_rnode);
      }
    }
    clb->remove_outputPinListItr();

    // connect CLB input pins to base channel tracks.
    // full connectivity assumed.
//      clb->create_inputPinListItr();
//      input_rnode=clb->getnext_inputPinList();
//      for(;input_rnode!=0;input_rnode=clb->getnext_inputPinList()){
//        sbox_num=i/2;
//        sbox=levelArray->at(0)->get_switchBoxAt(sbox_num);
//        for(y=0;y<levelArray->at(0)->channelWidth;y++){
//  	if((i%2)==0)
//  	  wire_rnode=sbox->get_LRNodeAt(y);
//  	else
//  	  wire_rnode=sbox->get_RRNodeAt(y);
//  	create_routingEdge(wire_rnode,input_rnode);
//        }
//      }
//      clb->remove_inputPinListItr();

  }

  // now connect IOB routing nodes
  for(i=0;i<numIOBs;i++){

    iob=iobArray->at(i);

    // connect source to output pin
    source_rnode=iob->get_sourceNode();
    iob->create_outputPinListItr();
    output_rnode=iob->getnext_outputPinList();
    for(;output_rnode!=0;output_rnode=iob->getnext_outputPinList())
      create_routingEdge(source_rnode,output_rnode);
    iob->remove_outputPinListItr();

    // connect sink to input pin
    sink_rnode=iob->get_sinkNode();
    iob->create_inputPinListItr();
    input_rnode=iob->getnext_inputPinList();
    for(;input_rnode!=0;input_rnode=iob->getnext_inputPinList())
      create_routingEdge(input_rnode,sink_rnode);
    iob->remove_inputPinListItr();

    // connect output pin to associated IO wire
    iob->create_outputPinListItr();
    output_rnode=iob->getnext_outputPinList();
    for(;output_rnode!=0;output_rnode=iob->getnext_outputPinList()){
      wire_rnode=ioWireArray->at(i);
      create_routingEdge(output_rnode,wire_rnode);
    }
    iob->remove_outputPinListItr();

    // connect input pin to associated IO wire
    iob->create_inputPinListItr();
    input_rnode=iob->getnext_inputPinList();
    for(;input_rnode!=0;input_rnode=iob->getnext_inputPinList()){
      wire_rnode=ioWireArray->at(i);
      create_routingEdge(wire_rnode,input_rnode);
    }
    iob->remove_inputPinListItr();

  }

  // now connect up the interconnect structure.
  for(i=0;i<numLevels;i++){
    level=levelArray->at(i);
    for(j=0;j<level->get_numSwitchBoxes();j++){
      sbox=level->get_switchBoxAt(j);

      for(k=0;k<sbox->get_numSwitches();k++){
	// first, connect the L and R terminals of each switch
	// with bidirectional edges
	lrnode=sbox->get_LRNodeAt(k);
	rrnode=sbox->get_RRNodeAt(k);

	create_routingEdge(lrnode,rrnode);
	create_routingEdge(rrnode,lrnode);

	// now, establish edges between routing nodes in
	// consecutive levels. if we have a COMPRESSING
	// switchbox, then only 1 bidirectional edge will
	// be created between corresponding nodes. In case
	// the switchbox is NON_COMPRESSING, then each
	// routing node in the lower level will connect
	// to 2 routing nodes in the higher level.
	if(i<=(numLevels-2)){
	  next_level=levelArray->at(i+1);
	  next_sbox=next_level->get_switchBoxAt(j/2);
	  if(level->switchType==NON_COMPRESSING){
	    if((j%2)==0){
	      // even position switchbox
	      // connect to nodes in LArray of next_sbox
	      lrnode0=next_sbox->get_LRNodeAt(2*k);
	      lrnode1=next_sbox->get_LRNodeAt(2*k+1);

	      create_routingEdge(lrnode,lrnode0);
	      create_routingEdge(lrnode0,lrnode);

	      create_routingEdge(lrnode,lrnode1);
	      create_routingEdge(lrnode1,lrnode);

	      create_routingEdge(rrnode,lrnode0);
	      create_routingEdge(lrnode0,rrnode);

	      create_routingEdge(rrnode,lrnode1);
	      create_routingEdge(lrnode1,rrnode);
	    }
	    else{
	      // odd position switchbox
	      // connect to nodes in RArray of next_sbox
	      rrnode0=next_sbox->get_RRNodeAt(2*k);
	      rrnode1=next_sbox->get_RRNodeAt(2*k+1);

	      create_routingEdge(lrnode,rrnode0);
	      create_routingEdge(rrnode0,lrnode);

	      create_routingEdge(lrnode,rrnode1);
	      create_routingEdge(rrnode1,lrnode);

	      create_routingEdge(rrnode,rrnode0);
	      create_routingEdge(rrnode0,rrnode);

	      create_routingEdge(rrnode,rrnode1);
	      create_routingEdge(rrnode1,rrnode);
	    }
	  }
	  else{
	    if((j%2)==0){
	      // even position switchbox
	      // connect to nodes in LArray of next_sbox
	      lrnode0=next_sbox->get_LRNodeAt(k);

	      create_routingEdge(lrnode,lrnode0);
	      create_routingEdge(lrnode0,lrnode);

	      create_routingEdge(rrnode,lrnode0);
	      create_routingEdge(lrnode0,rrnode);
	    }
	    else{
	      // odd position switchbox
	      // connect to nodes in RArray of next_sbox
	      rrnode0=next_sbox->get_RRNodeAt(k);

	      create_routingEdge(lrnode,rrnode0);
	      create_routingEdge(rrnode0,lrnode);

	      create_routingEdge(rrnode,rrnode0);
	      create_routingEdge(rrnode0,rrnode);
	    }
	  }

	}
	else{
	  // we are the root switchbox here.
	  // the root switchbox connects the IO wires
	  // to the inteconnect wires at the top of
	  // the interconnect hierarchy.
	  if(level->switchType==NON_COMPRESSING){
	    wire_rnode0=ioWireArray->at(2*k);
	    wire_rnode1=ioWireArray->at(2*k+1);

	    create_routingEdge(lrnode,wire_rnode0);
	    create_routingEdge(wire_rnode0,lrnode);

	    create_routingEdge(lrnode,wire_rnode1);
	    create_routingEdge(wire_rnode1,lrnode);

	    create_routingEdge(rrnode,wire_rnode0);
	    create_routingEdge(wire_rnode0,rrnode);

	    create_routingEdge(rrnode,wire_rnode1);
	    create_routingEdge(wire_rnode1,rrnode);
	  }
	  else{
	    wire_rnode0=ioWireArray->at(k);

	    create_routingEdge(lrnode,wire_rnode0);
	    create_routingEdge(wire_rnode0,lrnode);

	    create_routingEdge(rrnode,wire_rnode0);
	    create_routingEdge(wire_rnode0,rrnode);
	  }

	}

      }
    }
  }

}

int hsraFPGA::get_baseChanWidth(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return baseChanWidth;
}

int hsraFPGA::get_numLevels(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numLevels;
}

double hsraFPGA::get_growthRate(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return growthRate;
}

hsraBlock* hsraFPGA::get_clbAt(int pos){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(pos<0 || pos>=clbArray->size())
    ERROR("array overflow here...");
  return clbArray->at(pos);

}

hsraBlock* hsraFPGA::get_iobAt(int pos){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(pos<0 || pos>=iobArray->size())
    ERROR("array overflow here...");
  return iobArray->at(pos);

}

int hsraFPGA::get_numCLBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numCLBs;
}

int hsraFPGA::get_numIOBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numIOBs;
}

void hsraFPGA::addto_rnodeList(routingNode* r){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("trying to add NULL rnode...");
  rnodeList.push_back(r);
}

void hsraFPGA::create_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(rnode_itr)
    ERROR("an iterator already exists...");
  rnode_itr=new list<routingNode*>::iterator;
  *rnode_itr=rnodeList.begin();
}

void hsraFPGA::remove_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("there is no iterator to remove...");
  delete rnode_itr;
  rnode_itr=0;
}

routingNode* hsraFPGA::getnext_rnodeList(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("trying to access NULL iterator here...");

  if(*rnode_itr == rnodeList.end())
    return 0;
  else{
    routingNode* ret_ptr = **rnode_itr;
    (*rnode_itr)++;
    return ret_ptr;
  }

}

routingNode* hsraFPGA::found_routingNode(char* rname){
  if(magicNum!=23111)
    ERROR("trying to access a deleted node...");
  if(!rname)
    ERROR("node name cannot be NULL...");
  list<routingNode*>::iterator node_itr = rnodeList.begin();
  for(;node_itr!=rnodeList.end();node_itr++){
    if(strcmp((*node_itr)->get_name(),rname)==0)
      return *node_itr;
  }
  return 0;
}

