////////////////////////////////////////////////////////////////////
// logicBlock.cc: Definitions of logicBlock member functions.
// 
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "logicBlock.h"

// Create a new iterator for inputList
void logicBlock::create_inputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(interm_itr)
    ERROR("input terminal iterator already exists...");
  interm_itr=new list<Terminal*>::iterator;
  *interm_itr = inputList.begin();
}

// Create a new iterator for outputList
void logicBlock::create_outputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(outterm_itr)
    ERROR("output terminal iterator already exists...");
  outterm_itr=new list<Terminal*>::iterator;
  *outterm_itr = outputList.begin();
}

// Remove an already existing inputList iterator 
void logicBlock::remove_inputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(interm_itr==0)
    ERROR("trying to delete NULL input terminal iterator...");
  delete interm_itr;
  interm_itr=0;
}

// Remove an already existing outputList iterator 
void logicBlock::remove_outputListItr(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  if(outterm_itr==0)
    ERROR("trying to delete NULL output terminal iterator...");
  delete outterm_itr;
  outterm_itr=0;
}

// Get next element from inputList
Terminal* logicBlock::getnext_inputList(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  if(*interm_itr == inputList.end())
    return 0;
  else{
    Terminal* ret_ptr = **interm_itr;
    (*interm_itr)++;
    return ret_ptr;
  }

}

// Get next element from outputList
Terminal* logicBlock::getnext_outputList(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  if(*outterm_itr == outputList.end())
    return 0;
  else{
    Terminal* ret_ptr = **outterm_itr;
    (*outterm_itr)++;
    return ret_ptr;
  }

}

// Set the funtionality type of this block
void logicBlock::set_functionType(int f){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  functionType = f;
}

// Give this block a name
void logicBlock::set_name(char* n){
  if(!n)
    ERROR("logic block name cannot be NULL...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Set this if block actually implements logic
void logicBlock::set_isLogic(bool u){
  if(u!=true && u!=false)
    ERROR("non-bool value...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  isLogic = u;
}

// Set this if block is an input-only IO block
void logicBlock::set_isInput(bool u){
  if(u!=true && u!=false)
    ERROR("non-bool value...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  isInput = u;
}

// Set this if block is an output-only IO block
void logicBlock::set_isOutput(bool u){
  if(u!=true && u!=false)
    ERROR("non-bool value...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  isOutput = u;
}

// Set this if block is a bidirectional IO block
void logicBlock::set_isBidirectional(bool u){
  if(u!=true && u!=false)
    ERROR("non-bool value...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  isBidirectional = u;
}

// Set this if this is a registered / latched block
void logicBlock::set_isRegister(bool u){
  if(u!=true && u!=false)
    ERROR("non-bool value...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  isRegister = u;
}

// Add a terminal to the input terminal list
void logicBlock::addto_inputList(Terminal* t){
  if(!t)
    ERROR("trying to add NULL terminal...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  inputList.push_back(t);
  numInputs++;
}

// Add a terminal to the output terminal list
void logicBlock::addto_outputList(Terminal* t){
  if(!t)
    ERROR("trying to add NULL terminal...");
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  outputList.push_back(t);
  numOutputs++;
}

// Get name of this block
char* logicBlock::get_name(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Get functionality type of this block
int logicBlock::get_functionType(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return functionType;
}

// Check if this block actually implements logic
bool logicBlock::get_isLogic(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return isLogic;
}

// Check if this block is an input-only IO block
bool logicBlock::get_isInput(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return isInput;
}

// Check if this block is an output-only IO block
bool logicBlock::get_isOutput(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return isOutput;
}

// Check if this block is a bidirectional IO block
bool logicBlock::get_isBidirectional(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return isBidirectional;
}

// Check if this block is a registered / latched block
bool logicBlock::get_isRegister(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return isRegister;
}

// Get num input terminals
int logicBlock::get_numInputs(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get num output terminals
int logicBlock::get_numOutputs(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Print details about this logic block
void logicBlock::print_Info(){
  if(magicNum!=15111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR CLB "<<name;

  //print IO or logic
  if(isOutput)
    cout<<endl<<"OUTPUT BLOCK";
  if(isInput)
    cout<<endl<<"INPUT BLOCK";
  if(isLogic)
    cout<<endl<<"LOGIC BLOCK";
  if(isBidirectional)
    cout<<endl<<"BI-DIRECTIONAL BLOCK";

  //print functionality type
  if(isLogic)
    cout<<endl<<"LOGIC BLOCK TYPE IS "<<functionType;

  //print whether registered
  if(isRegister)
    cout<<endl<<"REGISTERED BLOCK";

  //print input list
  cout<<endl<<"INPUT TERMINALS -->";
  list<Terminal*>::iterator input_itr = inputList.begin();
  for(;input_itr!=inputList.end();input_itr++)
    (*input_itr)->print_Info();

  //print output list
  cout<<endl<<"OUTPUT TERMINALS -->";
  list<Terminal*>::iterator output_itr = outputList.begin();
  for(;output_itr!=outputList.end();output_itr++)
    (*output_itr)->print_Info();

}


