////////////////////////////////////////////////////////////////////
// Net.cc: Definition of Net member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "Net.h"

void Net::create_sinkVector(){
  if(sinkList.size()<1)
    ERROR("trying to create a vector of 0 sinks...");
  sinkVector=new vector<routingNode*> (sinkList.size());

  list<routingNode*>::iterator sink_itr = sinkList.begin();
  for(int i=0;sink_itr!=sinkList.end();sink_itr++,i++)
    sinkVector->at(i)=*sink_itr;
}

// Give this net a name
void Net::set_name(char* n){
  if(!n)
    ERROR("net name cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Set this net's source terminal
void Net::set_source(routingNode* s){  
  if(!s)
    ERROR("net source cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  source=s;
}

// Add a terminal to this net's sink list
void Net::addto_sinkList(routingNode* s){
  if(!s)
    ERROR("sink cannot be NULL...");
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  sinkList.push_back(s);
  numSinks++;
}

// Get name of this net
char* Net::get_name(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Rip up the existing routing tree, and decrement sharing
void Net::rip_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"RIPPING ROUTE TREE OF "<<name;
#endif

  rTree.postorderTraverse(rTree.get_root(),2);
  rTree.set_root(0);
}

// Print the current route tree of this net
void Net::print_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  rTree.print_info();
}

// Calculate the sharing and wire cost of this net's current route
void Net::calcCosts(int* wr_cost){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(*wr_cost!=0)
    ERROR("uninitialized cost parameters here...");
  wsCostCalc(rTree.get_root(),wr_cost);
}

// Recursively calculate wire and sharing cost
void Net::wsCostCalc(treeNode* r,int* wr_cost){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    wsCostCalc(child,wr_cost);
  tree_node->remove_childListItr();

  *wr_cost+=1;
  cout<<endl<<"\tincremented wire cost due to "<<tree_node->get_rNode()->name<<
    " to "<<*wr_cost;

}

// Get the total number of nodes in the current route tree
int Net::get_wireCount(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  int wire_count=0;
  countWires(rTree.get_root(),&wire_count);
  return wire_count;
}

// Count number of nodes in the current route tree
void Net::countWires(treeNode* r,int* count_ptr){
  treeNode* tree_node=r;
  treeNode* child;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    countWires(child,count_ptr);
  tree_node->remove_childListItr();

  if(tree_node->get_rNode()->get_type()==BUS_OR_WIRE)
    *count_ptr+=1;

}

// Make the source routing node the root of  rTree
void Net::init_rTreeRoot(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  treeNode* root_node=new treeNode;
  root_node->set_rNode(source);
  rTree.set_root(root_node);
}

// Get this net's source terminal
routingNode* Net::get_source(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return source;
}

// Get num sinks of this net
int Net::get_numSinks(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  return numSinks;
}

// Print details about this net
void Net::print_Info(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"===================================";
  cout<<endl<<"              PRINT NET              ";
  cout<<endl<<"===================================";

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR NET "<<name;
  cout<<endl<<"SOURCE NODE IS "<<source->get_name();

  cout<<endl<<"SINK LOGIC BLOCKS ARE -->";
  list<routingNode*>::iterator sink_itr = sinkList.begin();
  for(;sink_itr!=sinkList.end();sink_itr++)
    cout<<"  "<<(*sink_itr)->get_name();

  cout<<endl;

}

// Get a pointer to this net's route tree
routeTree* Net::get_rTree(){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");

  return &rTree;
}

// Make this route tree a copy of the tree rooted at tree_root.
void routeTree::populateWith(treeNode* tree_root){

//   cout<<endl<<endl<<"IN POPULATE WITH...";

  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(tree_root==0)
    ERROR("route tree's root cannot be NULL...");

  root=makeCopy(tree_root);

//   cout<<endl<<"EXITING POPULATE WITH...";

}

// Find tree node that points to routing node rnode
treeNode* routeTree::getNode(routingNode* rnode){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(rnode==0)
    ERROR("trying to look for NULL routing node...");

  treeNode* ret_node=findNode(root,rnode);
  if(ret_node==0){

#if defined(PLACER_DEBUG)
    cout<<endl<<endl<<"COULD NOT FIND "<<rnode->name;
    print_info();
#endif

  }

  return ret_node;
}

// Recursively build a copy of this route tree
treeNode* routeTree::makeCopy(treeNode* r){
  treeNode* child;treeNode* child_copy;
  treeNode* tree_node=r;treeNode* tree_node_copy;

  /* create a new tree node that will be a copy of tree node */
  tree_node_copy=new treeNode;
  tree_node_copy->set_rNode(tree_node->get_rNode());

//   cout<<endl<<"\tcreated "<<tree_node_copy->get_rNode()->get_name();

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    /* recursively create a copy of the subtree rooted at this node */
    child_copy=makeCopy(child);
    tree_node_copy->addto_childList(child_copy);
    child_copy->set_parent(tree_node_copy);
  }
  tree_node->remove_childListItr();

  /* return the newly created copy of tree node */
  return tree_node_copy;

}

// Clean the routing nodes in this tree. This is a post-BFS reinit operation.
void routeTree::clean_rNodes(treeNode* r){
  treeNode* tree_node=r; treeNode* child;
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");

  tree_node->get_rNode()->isExplored=false;
  tree_node->get_rNode()->inrTree=false;
  tree_node->get_rNode()->predecessor=0;
  tree_node->get_rNode()->pathCost=HUGE_NUMBER;
  tree_node->get_rNode()->isRouted=false;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    clean_rNodes(child);
  tree_node->remove_childListItr();

}

// Set the root of this tree
void routeTree::set_root(treeNode* rt){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
//   if(rt==0)
//     ERROR("route tree's root cannot be NULL...");
  root=rt;
}

// Get the root of this tree
treeNode* routeTree::get_root(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  return root;
}

// Print this tree and info relating to it 
void routeTree::print_info(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(root==0)
    ERROR("trying to print NULL tree here...");
  cout<<endl<<"PRINT ROUTE TREE -->"<<endl;
  preorderTraverse(root,0);
  cout<<endl<<endl;
}

// Clean up this routing tree
void routeTree::clear(){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  postorderTraverse(root,1);
  root=0;
}

// Increment the occupancies of the routing nodes that constitute this tree
void routeTree::incrementOccupancies(int how){
  if(magicNum!=32111)
    ERROR("trying to access already deleted node...");
  if(how!=0 && how!=1)
    ERROR("spurious parameter value here...");

  if(how==0)
    postorderTraverse(root,3);
  else
    postorderTraverse(root,4);
}

// Traverse this tree preorder. The actions to be taken at each node are specified by do_what
void routeTree::preorderTraverse(treeNode* r,int do_what){
  treeNode* child;
  treeNode* tree_node=r;

  if(do_what==0)
    /* print this tree node */
    cout<<endl<<tree_node->get_rNode()->get_name();

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    preorderTraverse(child,do_what);
  tree_node->remove_childListItr();

}

// Traverse this tree postorder. The action to be taken at each node are specified by do_what
void routeTree::postorderTraverse(treeNode* r,int do_what){
  treeNode* child;
  treeNode* tree_node=r;

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList())
    postorderTraverse(child,do_what);
  tree_node->remove_childListItr();

  if(do_what==0)
    /* print this tree node */
    cout<<endl<<tree_node->get_rNode()->get_name()<<"  "<<tree_node->get_pathCost();
  else if(do_what==1)
    /* delete this tree node */
    delete tree_node;
  else if(do_what==2){
    /* decrement occupancy */
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()-1);

#if defined(PLACER_DEBUG)
    cout<<endl<<"\tdecrement and delete "<<tree_node->get_rNode()->get_name()
        <<" occ = "<<tree_node->get_rNode()->get_occupancy();
#endif

    /* delete this tree node */
    delete tree_node;
  }
  else if(do_what==3){
    /* increment occupancy of all routing nodes*/
    tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
  }
  else if(do_what==4){
    /* increment occupancy of non source/sink routing nodes*/
    if(tree_node->get_rNode()->get_type()!=SOURCE && tree_node->get_rNode()->get_type()!=SINK)
      tree_node->get_rNode()->set_occupancy(tree_node->get_rNode()->get_occupancy()+1);
  }
  else if(do_what==5){
//     tree_node->get_rNode()->set_counted(false);
  }


}

// Find rnode in this tree and return the relevant tree node
treeNode* routeTree::findNode(treeNode* r,routingNode* rnode){
  treeNode* tree_node=r;treeNode* ret_node=0;
  treeNode* child;

//   cout<<endl<<"this is tree node "<<tree_node->get_rNode()->get_name();

  if(tree_node->get_rNode()==rnode){
    /* found rnode...return this node */
    return tree_node;
  }

  tree_node->create_childListItr();
  for(child=tree_node->getnext_childList();child!=0;child=tree_node->getnext_childList()){
    ret_node=findNode(child,rnode);
    if(ret_node){
      tree_node->remove_childListItr();
      return ret_node;
    }
  }
  tree_node->remove_childListItr();

  /* rnode does not exist in any node in the sub-tree rooted at tree_node */
  return 0;

}

// Return tree node that holds the routing node pointed to by rnode
treeNode* Net::find_treeNode(routingNode* rnode){
  if(magicNum!=17111)
    ERROR("trying to access a deleted node...");
  if(rnode==0)
    ERROR("cant find NULL routing node in tree...");
  treeNode* tree_node=rTree.findNode(rTree.get_root(),rnode);
  if(tree_node==0)
    ERROR("uh,oh! could not find routing node in tree...");

//   cout<<endl<<"FOUND tree node corresponding to "<<tree_node->get_rNode()->get_name();

  return tree_node;
}

