////////////////////////////////////////////////////////////////////
// Net.h : Class that describes a net in a generic netlist.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#ifndef NET_H
#define NET_H
#include "error.h"
#include "routingNode.h"

class treeNode;
class Net;
class routingNode;

class routeTree{
 public:
  treeNode* root;  //the root node of this route tree
  int magicNum;

  routeTree(){
    root=0;magicNum=32111;
  }
  routeTree(treeNode* rt){
    root=rt;magicNum=32111;
  }
  ~routeTree(){
    if(magicNum!=32111)
      ERROR("trying to double delete tree here...");
    else
      magicNum=0;
  }

  void set_root(treeNode* rt);  //set the root of this tree
  treeNode* get_root();  //get the root of this tree
  void print_info();  //print this tree and info relating to it 
  void preorderTraverse(treeNode* r,int do_what);  //traverse this tree preorder
  void postorderTraverse(treeNode* r,int do_what);  //traverse this tree postorder
  void clear();  //clean up this routing tree
  treeNode* findNode(treeNode* r,routingNode* rnode);  //find rnode in this tree and return relevant tree node

  //make this route tree a copy of the tree rooted at tree_root
  void populateWith(treeNode* tree_root);
  treeNode* makeCopy(treeNode* r);

  //increment the occupancies of the routing nodes that constitute this tree
  void incrementOccupancies(int how);

  //find tree node that points to routing node rnode
  treeNode* getNode(routingNode* rnode);

  // Clean the routing nodes in this tree. This is a post-BFS reinit operation.
  void clean_rNodes(treeNode* r);

};

class Net{
 public:
  char name [MAX_STR_LEN];
  routingNode* source;
  list<routingNode*> sinkList;
  vector<routingNode*>* sinkVector;
  int numSinks;
  routeTree rTree;
  int magicNum;

  Net(){
    magicNum=17111;
    source=0;
    numSinks=0;
    sinkVector=0;
  }

  ~Net(){
    if(magicNum!=17111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //give this net a name
  void set_source(routingNode* s);  //set this net's source terminal
  void addto_sinkList(routingNode* s);  //add a terminal to this net's sink list
  void init_rTreeRoot();  //make the source routing node the root of rTree
  void rip_rTree();  //rip up the existing routing tree, and decrement sharing

  //get a pointer to this net's route tree
  routeTree* get_rTree();

  char* get_name();  //get name of this net
  routingNode* get_source();  //get this net's source terminal
  int get_numSinks();  //get num sinks of this net
  treeNode* find_treeNode(routingNode* rnode);  //return tree node that holds the routing node rnode
  void print_Info();  //print details about this net
  void print_rTree();  //print the current route tree of this net
  int get_wireCount();  //get the total number of nodes in the current route tree

  //count number of nodes in the current route tree
  void countWires(treeNode* r,int* count_ptr);
  
  //calculate the wire cost of this net's current route
  void calcCosts(int* wr_cost); 
  
  //recursively calculate wire and sharing cost
  void wsCostCalc(treeNode* r,int* wr_cost);

  void create_sinkVector();

};

#endif


