////////////////////////////////////////////////////////////////////
// kMeans.h : Global + function declarations for kMeans.y.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#ifndef MCNC_PARSER_H
#define MCNC_PARSER_H

#include "routingNode.h"
#include "routingEdge.h"
#include "routingGraph.h"
#include "Net.h"
#include "cluster.h"

/* curr_tok holds the current token type that is being parsed */
int curr_tok=0;

/* stringList is a list of strings (generally net names, 'open' and digits) */
list<char*> stringList;
list<char*>::iterator stringListItr;

/* The routing node that is currently being read in from rr_graph.echo */
routingNode* G_currNode=0;

/* Declare a routingGraph object */
routingGraph rrgraph;

int G_FinalNumClusters = 0;

/* define an entry of the priority queue */
class pqNode{
 public:
  routingNode* rNode;  //the routing node that this priority queue entry represents
  double pathCost;  //the cost of the path from the source to rNode
  routingNode* predNode;  //the fanin routing node that was used to explore rNode
  int magicNum;

  pqNode(){
    rNode=0;
    pathCost=HUGE_NUMBER;
    predNode=0;
    magicNum=33111;
  }
  pqNode(routingNode* r,double p,routingNode* pnode){
    if(r==0 || p<0.)
      ERROR("error in creating a new PQ node...");
    rNode=r;
    pathCost=p;
    predNode=pnode;
    magicNum=33111;
  }
  ~pqNode(){
    if(magicNum!=33111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  bool operator< (const pqNode& x) const { 
    return pathCost > x.pathCost;
  }

};

/* Declare a cost table */
/* vector<vector<costTableEntry*> *>* costTable; */
vector<vector<float> *>* costTable;

/* Declare vector of clusters */
vector<Cluster*>* clusterVec;

/***
* FUNCTION DECLARATIONS
***/
void yyerror(char*);
int yylex();
void Dijkstra(priority_queue<pqNode>*,bool,Cluster*,routingNode*);
void createCostTable(long int);
void populateCostTable();
void printCostTable();
int determineNumClusters();
void createClusterVector(int,long int);
void pickStartCentroids(int);
void printClusterInfo();
double calculateDistanceBetween(vector<float*>*,Cluster*);
bool terminateClustering();
void performClustering(FILE*);
void populateEstimateTables();
void writeOutFiles(char*,float,float);
void writeOutSourceNode(routingNode*,FILE*);
void writeOutSinkNode(routingNode*,FILE*);
void writeOutInputPinNode(routingNode*,FILE*);
void writeOutOutputPinNode(routingNode*,FILE*);
void writeOutWireNode(routingNode*,FILE*);
routingNode* getNewWireConnectedToSource();
void visualClusterDump();

routingNode* find_rnode(char*);
void process_sourceNode(int,int,int);
void process_sinkNode(int,int,int);
void process_ipinNode(int,int,int);
void process_opinNode(int,int,int);
void process_chanNode(int,int,int,int,int);
void process_Edges(int);
void process_switchTypes();
double flomin(double,double);

#endif

