////////////////////////////////////////////////////////////////////
// routingEdge.h : Class that describes an edge in the routing graph.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#ifndef ROUTING_EDGE_H
#define ROUTING_EDGE_H

#include "error.h"
#include "routingNode.h"

class routingNode;

class routingEdge{
 public:
  int type;  //type of routing edge based on rr_graph.echo and arch.echo
  routingNode* origNode;  //the routing node that originates this edge
  routingNode* destNode;  //the routing node that sinks this edge
  int magicNum;

  routingEdge(){
    type=NO_TYPE;
    origNode=destNode=0;
    magicNum=20111;
  }
  routingEdge(int t,routingNode* orig,routingNode* dest){
    if(orig==0 || dest==0)
      ERROR("source / destination routing node cannot be NULL...");
    type=t;
    origNode=orig;destNode=dest;
    magicNum=20111;
  }
  ~routingEdge(){
    if(magicNum!=20111)
      ERROR("trying to double delete node here...");
    else
      magicNum=0;
  }

  void set_type(int t);  //set type of routing edge i.e. buffered or unbuffered
  void set_origNode(routingNode* n);  //set the routing node that originates this edge
  void set_destNode(routingNode* n);  //set the routing node that sinks this edge

  int get_type();  //get type of this routing edge
  routingNode* get_origNode();  //get the routing node that originates this edge
  routingNode* get_destNode();  //get the routing node that sinks this edge

  void print_info();  //print this routing edge's info 

};

#endif


