////////////////////////////////////////////////////////////////////
// routingNode.cc : Definition of routingNode member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "routingNode.h"

/* return the larger of two floating point numbers */
double flomax(double x,double y){
  double diff=x-y;
  if(diff<-1.0e-15)
    return y;
  else
    return x;
}

// Give this routing node a name
void routingNode::set_name(char* n){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(n==0)
    ERROR("name of routing node cannot be NULL...");
  strcpy(name,n);
}

// Set type of routing node
void routingNode::set_type(int t){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(t<29 || t>33)
    ERROR("mysterious routing node type...");
  type=t;
}

// Set this node's wavefront predecessor
void routingNode::set_predecessor(routingNode* pr){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  predecessor=pr;
}

// Set resistance of wire routing node
void routingNode::set_R(double r){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(r<0.)
    ERROR("resistance cannot be -ve...");
  R=r;
}

// Set input cap of wire routing node
void routingNode::set_Cin(double c){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(c<0.)
    ERROR("capacitance cannot be -ve...");
  Cin=c;
}

// Set output cap of wire routing node
void routingNode::set_Cout(double c){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(c<0.)
    ERROR("capacitance cannot be -ve...");
  Cout=c;
}

// Set base cost of this routing node
void routingNode::set_baseCost(double b){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(b<0.)
    ERROR("base cost cannot be set -ve...");
  baseCost=b;
}

// Set the occupancy of this routing node
void routingNode::set_occupancy(int occ){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(occ<0)
    ERROR("occupancy of routing node cannot be < 0...");
  if(type==SOURCE || type==SINK){
    if(occ>capacity){
      cout<<endl<<endl<<"VIOLATED TERMINAL IS "<<name<<" AT OCC = "<<occ;
      ERROR("occupancy of source/sink routing node cannot exceed capacity...");
    }
  }
  occupancy=occ;
}

// Set the capacity of this routing node
void routingNode::set_capacity(int c){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(c<0)
    ERROR("capacity of routing node cannot be < 0...");
  capacity=c;
}

// Whether this node has been explored by a wavefront
void routingNode::set_isExplored(bool e){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(e!=true && e!=false)
    ERROR("non-bool value here...");
  isExplored=e;
}

// Set the current path cost of this routing node
void routingNode::set_pathCost(double p){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(p<0.)
    ERROR("path cost cannot be -ve...");
  pathCost=p;
}

// Add a routing edge to the fanin list   
void routingNode::addto_faninList(routingEdge* e){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(e==0)
    ERROR("adding NULL routing edge here...");
  faninList.push_back(e);
}

// Add a routing edge to the fanout list   
void routingNode::addto_fanoutList(routingEdge* e){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(e==0)
    ERROR("adding NULL routing edge here...");
  fanoutList.push_back(e);
}

// Create a new iterator for faninList
void routingNode::create_faninListItr(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(fanin_itr)
    ERROR("fanin iterator already exists...");
  fanin_itr=new list<routingEdge*>::iterator;
  *fanin_itr = faninList.begin();
}

// Remove an already existing faninList iterator 
void routingNode::remove_faninListItr(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(fanin_itr==0)
    ERROR("trying to delete NULL fanin iterator...");
  delete fanin_itr;
  fanin_itr=0;
}

// Create a new iterator for fanoutList
void routingNode::create_fanoutListItr(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(fanout_itr)
    ERROR("fanout iterator already exists...");
  fanout_itr=new list<routingEdge*>::iterator;
  *fanout_itr = fanoutList.begin();
}

// Remove an already existing fanoutList iterator 
void routingNode::remove_fanoutListItr(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  if(fanout_itr==0)
    ERROR("trying to delete NULL fanout iterator...");
  delete fanout_itr;
  fanout_itr=0;
}

// Whether this routing node is in the current route tree of the net being routed 
void routingNode::set_inrTree(bool r){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  inrTree=r;
}

// Return true if this routing node is in the route tree of the net being routed
bool routingNode::get_inrTree(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return inrTree;
}

// Get the current cost of this routing node
double routingNode::returnCost(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return baseCost;
}

// Get the name of this routing node
char* routingNode::get_name(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Get this node's wavefront predecessor
routingNode* routingNode::get_predecessor(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return predecessor;
}

// Get type of routing node
int routingNode::get_type(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return type;
}

// Get the resistance of wire routing node
double routingNode::get_R(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return R;
}

// Get the input cap of wire routing node
double routingNode::get_Cin(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return Cin;
}

// Get the output cap of wire routing node
double routingNode::get_Cout(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return Cout;
}

// Get next element from faninList
routingEdge* routingNode::getnext_faninList(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");

  if(*fanin_itr == faninList.end())
    return 0;
  else{
    routingEdge* ret_ptr = **fanin_itr;
    (*fanin_itr)++;
    return ret_ptr;
  }

}

// Get next element from fanoutList
routingEdge* routingNode::getnext_fanoutList(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");

  if(*fanout_itr == fanoutList.end())
    return 0;
  else{
    routingEdge* ret_ptr = **fanout_itr;
    (*fanout_itr)++;
    return ret_ptr;
  }

}

// Get base cost of this routing node
double routingNode::get_baseCost(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return baseCost;
}

// Get the current occupancy of this node
int routingNode::get_occupancy(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return occupancy;
}

// Get the capacity of this node
int routingNode::get_capacity(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return capacity;
}

// Get whether this node has been explored by a wavefront
bool routingNode::get_isExplored(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return isExplored;
}

// Get current path cost of this routing node
double routingNode::get_pathCost(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");
  return pathCost;
}

// Print details about this routing node.
// This function only prints architectural and netlist info.
void routingNode::print_info(){
  if(magicNum!=19111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"NODE: "<<name<<".";
  if(type==SOURCE)
    cout<<"  SOURCE NODE";
  else if(type==SINK)
    cout<<"  SINK NODE";
  else if(type==BUS_OR_WIRE)
    cout<<"  BUS_OR_WIRE";
  else if(type==INPUT_PIN)
    cout<<"  INPUT PIN";
  else
    cout<<"  OUTPUT PIN";

//   cout<<"\tPath Cost = "<<pathCost;

  int fanout_size=fanoutList.size();
  cout<<endl<<"   "<<fanout_size<<" EDGES:";
  list<routingEdge*>::iterator fanout_itr = fanoutList.begin();
  for(;fanout_itr!=fanoutList.end();fanout_itr++){
    cout<<" ";
    (*fanout_itr)->print_info();
  }

  int fanout_type;
  fanout_itr=fanoutList.begin();
  cout<<endl<<"SWITCH TYPES:";
  for(;fanout_itr!=fanoutList.end();fanout_itr++){
    cout<<" ";
    fanout_type=(*fanout_itr)->get_type();
    cout<<fanout_type;
  }

  cout<<endl<<"OCC: "<<occupancy<<" CAPACITY: "<<capacity<<".";
  cout<<endl<<"R: "<<R<<" C: "<<C;

}

// Erase the child node pointed to by ch_itr
void treeNode::eraseChild(list<treeNode*>::iterator* ch_itr){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(ch_itr==0)
    ERROR("child node has a NULL iterator...");
  childList.erase(*ch_itr);
} 

// Get the iterator that points to the node in childList that holds child_node  
list<treeNode*>::iterator treeNode::get_childItr(treeNode* child_node){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(child_node==0)
    ERROR("cannot look for NULL node in childList...");

  list<treeNode*>::iterator ch_itr=childList.begin();
  for(;ch_itr!=childList.end();ch_itr++){
    if(*ch_itr==child_node){
      return ch_itr;
    }
  }

  ERROR("could not find tree node in childList...");
}

// Set the parent node of this tree node
void treeNode::set_parent(treeNode* t){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(t==0)
    ERROR("setting parent of tree node to NULL...");
  parent=t;
}

// Set the routing node that this tree node represents
void treeNode::set_rNode(routingNode* r){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(r==0)
    ERROR("tree node cannot hold a NULL routing node...");
  rNode=r;

  /* at this point, set the inrTree member of rNode to indicate that rNode has been */
  /* associated with a node in the route tree of the net being currently routed */
//   rNode->set_inrTree(true);

}

// Set the cost of the path from the root to this tree node
void treeNode::set_pathCost(double p){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(p<0.)
    ERROR("path cost to a tree node cannot be NULL...");
  pathCost=p;
}

// Add a tree node to childList
void treeNode::addto_childList(treeNode* t){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(t==0)
    ERROR("setting child of tree node to NULL...");
  childList.push_back(t);
}

// Create a new childList iterator
void treeNode::create_childListItr(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(child_itr)
    ERROR("child iterator already exists...");
  child_itr=new list<treeNode*>::iterator;
  *child_itr = childList.begin();
}

// Get the number of children of this tree node
int treeNode::get_numChildren(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return childList.size();
} 

// Remove childList iterator
void treeNode::remove_childListItr(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(child_itr==0)
    ERROR("trying to delete NULL child iterator...");
  delete child_itr;
  child_itr=0;
}

// Get parent of this tree node
treeNode* treeNode::get_parent(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return parent;
}

// Get the routing node that this tree node represents
routingNode* treeNode::get_rNode(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return rNode;
}

// Get the cost of the path from the root to this tree node
double treeNode::get_pathCost(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  return pathCost;
}

// Get the next element in childList
treeNode* treeNode::getnext_childList(){
  if(magicNum!=31111)
    ERROR("trying to access already deleted node...");
  if(*child_itr == childList.end())
    return 0;
  else{
    treeNode* ret_ptr = **child_itr;
    (*child_itr)++;
    return ret_ptr;
  }
}

