////////////////////////////////////////////////////////////////////
// kMeans.y : The yacc grammar for parsing input files resides here.
//            Also resident here are the main function and the 
//            K-Means clustering implementation.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

%{
#include "kMeans.h"
%}

%error-verbose

%union{
  char* word;
  double double_num;
}

%token <word> NET_TOK
%token <word> PERIOD_TOK
%token <word> TO_TOK
%token <word> NODE_TOK
%token <word> CHANY_TOK
%token <word> CHANX_TOK
%token <word> SOURCE_TOK
%token <word> SINK_TOK
%token <word> IPIN_TOK
%token <word> OPIN_TOK
%token <word> PTC_TOK
%token <word> EDGES_TOK
%token <word> SWITCHTYPE_TOK
%token <word> OCC_TOK
%token <word> CAPACITY_TOK
%token <word> R_TOK
%token <word> C_TOK
%token <word> COSTINDEX_TOK
%token <double_num> DOUBLENUM_TOK;
%%
statement_list: statement
|    statement_list statement
;

statement: rr_graph_info{
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
}
;

rr_graph_info:
NODE_TOK NET_TOK PERIOD_TOK { 

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3; */

  G_currNode=find_rnode($2);
  free($1);free($2);free($3);
}
| SOURCE_TOK NET_TOK NET_TOK PTC_TOK NET_TOK { 

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  process_sourceNode(atoi($2),atoi($3),atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| SINK_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  process_sinkNode(atoi($2),atoi($3),atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| IPIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  process_ipinNode(atoi($2),atoi($3),atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| OPIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  process_opinNode(atoi($2),atoi($3),atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| CHANX_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  process_chanNode(atoi($2),atoi($3),atoi($2),atoi($3),atoi($5));
  G_currNode->isCHANX=true;
  free($1);free($2);free($3);free($4);free($5);
}
| CHANX_TOK NET_TOK NET_TOK TO_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($5),atoi($6),atoi($8));
  G_currNode->isCHANX=true;
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);
}
| CHANY_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  process_chanNode(atoi($2),atoi($3),atoi($2),atoi($3),atoi($5));
  free($1);free($2);free($3);free($4);free($5);
}
| CHANY_TOK NET_TOK NET_TOK TO_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($5),atoi($6),atoi($8));
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);
}
| NET_TOK EDGES_TOK nets {

/*   cout<<endl<<endl<<$1<<" "<<$2; */

  process_Edges(atoi($1));
  free($1);free($2);
}
| SWITCHTYPE_TOK nets {

/*   cout<<endl<<endl<<$1; */

  process_switchTypes();
  free($1);
}
| OCC_TOK NET_TOK CAPACITY_TOK NET_TOK PERIOD_TOK {

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5; */

  G_currNode->set_occupancy(atoi($2));
  G_currNode->set_capacity(atoi($4));
  free($1);free($2);free($3);free($4);free($5);
}  
| R_TOK DOUBLENUM_TOK C_TOK DOUBLENUM_TOK { 

/*   cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4; */

  G_currNode->R=$2; G_currNode->C=$4;
  free($1);free($3);
}
|  R_TOK NET_TOK C_TOK NET_TOK {   /* cout<<endl<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4; */ free($1);free($2);free($3);free($4); }
|  COSTINDEX_TOK NET_TOK {   /* cout<<endl<<endl<<$1<<" "<<$2; */ G_currNode->Cost_index=atoi($2);free($1);free($2); }
;

nets:
|    NET_TOK nets   { /* cout<<"\t"<<$1; */ stringList.push_front($1); }
;
%%

/* input file that is to be parsed */
extern FILE* yyin;

int main(int argc,char** argv){
  time_t start_time,end_time;
  time_t time_diff;
  FILE* fp;char outfile_name [MAX_STR_LEN];
  char* echo_file_name;
  long int seed;int num_logic_blocks;

  try{

    if(argc<6)
      ERROR("\nHSRA USAGE: ./astar <arch_file> <k_sample> <seed> <sub_sample> <num_logic_blocks>\n\n");

    num_logic_blocks = atoi(argv[5]);

    seed=atol(argv[3]);
    srand48(seed);

    /* parse rr_graph.echo file */
    yyin=fopen(argv[1],"r");
    while(!feof(yyin)){
      yyparse();
    }
    fclose(yyin);

/*     exit(1); */

    /* calculate the number of nodes of each type */
    rrgraph.calculateNodeTypes();

    /* set up node vectors for each type */
    rrgraph.createNodeVectors();

    /* Dump the routing graph */
/*     rrgraph.print_info(); */

/*     exit(1); */

    float sub_sample = atof(argv[4]);
    long int num_sink_nodes=(long int)(sub_sample*(float)(rrgraph.numSinkNodes));
    if(num_sink_nodes < 2)
      num_sink_nodes = 2;

    cout<<endl<<"Sub sample is "<<sub_sample<<", and num sink nodes is "<<num_sink_nodes;

    float k_sample = atof(argv[2]);
    int K = (int)(k_sample*(float)(num_logic_blocks));
    if(K < 2)
      K = 2;

    cout<<endl<<"K sample is "<<k_sample<<", and value of K is "<<K;

    sprintf(outfile_name,"c%4.3f_K%3.2f_%s.stat",sub_sample,k_sample,argv[1]);
    fp = fopen(outfile_name,"w");

    fprintf(fp,"\nSub sample is %f, and num sink nodes is %d",sub_sample,num_sink_nodes);
    fprintf(fp,"\nK sample is %f, and value of K is %d",k_sample,K);

    start_time=time(NULL);

    /* allocate and init the costTable */
    createCostTable(num_sink_nodes);

    /* use Dijkstra to populate the costTable with node --> sink distances */ 
    populateCostTable();

    end_time=time(NULL);
    time_diff=end_time-start_time;

    cout<<endl<<endl<<"Total Time Taken To Build The Cost-Table Is "<<time_diff;
    cout<<endl;
    fprintf(fp,"\n\nTotal Time Taken To Build The Cost-Table Is %d\n",time_diff);

    /* print the contents of the costTable */
/*     printCostTable(); */

    start_time=time(NULL);

    /* allocate and init clusterVec */
    createClusterVector(K,num_sink_nodes);

    /* now, select the K starting centroids */
    pickStartCentroids(K);

    /* do K-means clustering */
    performClustering(fp);

    /* Now, populate the estimate table of each cluster */
    populateEstimateTables();

    /* print cluster info */
/*     printClusterInfo(); */

    end_time=time(NULL);
    time_diff=end_time-start_time;
    cout<<endl<<endl<<"Total Time Taken To Do K-Means Clustering Is "<<time_diff;
    cout<<endl<<endl;
    fprintf(fp,"\n\nTotal Time Taken To Do K-Means Clustering Is %d\n\n",time_diff);
    fclose(fp);

/*     exit(1); */

    /* regurgitate the rrgraph file with clustering info. */
    /* also write out the A* estimate into a separate file. */
    writeOutFiles(argv[1],sub_sample,k_sample);

/*     visualClusterDump(); */

    return 0;
  }
  catch(out_of_range){
    cerr<<"vector overflow error...\n";
  }

}

void visualClusterDump(){
  ;
}

void writeOutFiles(char* rrgraph_filename,float sub_sample,float k_sample){
  char echo_filename [MAX_STR_LEN];
  char cluster_filename [MAX_STR_LEN];
  routingNode* rnode;routingNode* temp_rnode;routingNode* irnode;routingNode* jrnode;
  int inode_id,jnode_id,temp_id;

  int percent = (int)(ceil(100.0*sub_sample));

  if(strlen(rrgraph_filename) > (MAX_STR_LEN-11))
    ERROR("um, the name of the rrgraph file may be too long...");

  sprintf(echo_filename,"c%4.3f_K%3.2f_%s",sub_sample,k_sample,rrgraph_filename);
  FILE* echofp = fopen(echo_filename,"w");

  fprintf(echofp,"\n");
  for(int i=0;i<(rrgraph.rnodeVector)->size();i++){
    rnode=(rrgraph.rnodeVector)->at(i);
    if(rnode->prevClusterNum)
      fprintf(echofp,"%s\t%d\n",rnode->name,*(rnode->prevClusterNum));
    else
      fprintf(echofp,"%s\t-1\n",rnode->name);
  }
  fprintf(echofp,"\n");
  fclose(echofp);

  sprintf(cluster_filename,"c%4.3f_K%3.2f_%s.clus",sub_sample,k_sample,rrgraph_filename);
  FILE* clusterfp=fopen(cluster_filename,"w");

  fprintf(clusterfp,"\nNUM_CLUSTERS  %d",G_FinalNumClusters);
  fprintf(clusterfp,"\nNUM_SINKS  %d\n\n",(rrgraph.sinkNodeVector)->size());

  Cluster* cluster;
  for(int i=0;i<clusterVec->size();i++){
    cluster=clusterVec->at(i);
    if(cluster==0)
      continue;
    for(int j=0;j<cluster->estimate->size();j++)
      fprintf(clusterfp,"(%d,%s,%.3f)  ",cluster->number,cluster->sinkVector->at(j),cluster->estimate->at(j));
    fprintf(clusterfp,"\n\n");
  }
  fclose(clusterfp);

}

void writeOutSourceNode(routingNode* rnode,FILE* fp){
  routingEdge* redge;

  fprintf(fp,"Node: %s.  SOURCE (%d, %d)  Ptc_num: %d",rnode->name,rnode->col_low,rnode->row_low,rnode->Ptc_num);

  fprintf(fp,"\n   %d edge(s):",rnode->num_edges);
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();
  for(;redge!=0;redge=rnode->getnext_fanoutList())
    fprintf(fp," %s",redge->destNode->name);
  rnode->remove_fanoutListItr();

  fprintf(fp,"\nSwitch types:");
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();
  for(;redge!=0;redge=rnode->getnext_fanoutList())
    fprintf(fp," %d",redge->type);
  rnode->remove_fanoutListItr();

  fprintf(fp,"\nOcc: %d  Capacity: %d.",rnode->occupancy,rnode->capacity);
/*   if(rnode->R > 0.0 || rnode->C > 0.0) */
/*     ERROR("um, source node has non-zero R and/or C values..."); */
  fprintf(fp,"\nR: 0  C: 0");

  if(rnode->prevClusterNum==0)
    ERROR("source node is not assigned to any cluster...");

  fprintf(fp,"\nCluster_num: %d",*(rnode->prevClusterNum));
  fprintf(fp,"\nCost_index: %d\n\n",rnode->Cost_index);

}

void writeOutSinkNode(routingNode* rnode,FILE* fp){

  fprintf(fp,"Node: %s.  SINK (%d, %d)  Ptc_num: %d",rnode->name,rnode->col_low,rnode->row_low,rnode->Ptc_num);
  fprintf(fp,"\n   0 edge(s): ");
  fprintf(fp,"\nSwitch types: ");
  fprintf(fp,"\nOcc: %d  Capacity: %d.",rnode->occupancy,rnode->capacity);
/*   if(rnode->R > 0.0 || rnode->C > 0.0) */
/*     ERROR("um, sink node has non-zero R and/or C values..."); */
  fprintf(fp,"\nR: 0  C: 0");

  if(rnode->prevClusterNum!=0)
    ERROR("sink node is assigned to a cluster...");

  fprintf(fp,"\nCluster_num: -1");
  fprintf(fp,"\nCost_index: %d\n\n",rnode->Cost_index);

}

void writeOutInputPinNode(routingNode* rnode,FILE* fp){
  routingEdge* redge;

  fprintf(fp,"Node: %s.  IPIN (%d, %d)  Ptc_num: %d",rnode->name,rnode->col_low,rnode->row_low,rnode->Ptc_num);
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();rnode->remove_fanoutListItr();
  fprintf(fp,"\n   1 edge(s): %s",redge->destNode->name);
  fprintf(fp,"\nSwitch types: %d",redge->type);
  fprintf(fp,"\nOcc: %d  Capacity: %d.",rnode->occupancy,rnode->capacity);
/*   if(rnode->R > 0.0 || rnode->C > 0.0) */
/*     ERROR("um, input_pin node has non-zero R and/or C values..."); */
  fprintf(fp,"\nR: 0  C: 0");

  if(rnode->prevClusterNum!=0)
    ERROR("input pin is assigned to a cluster...");

  fprintf(fp,"\nCluster_num: -1");
  fprintf(fp,"\nCost_index: %d\n\n",rnode->Cost_index);

}

void writeOutOutputPinNode(routingNode* rnode,FILE* fp){
  routingEdge* redge;

  fprintf(fp,"Node: %s.  OPIN (%d, %d)  Ptc_num: %d",rnode->name,rnode->col_low,rnode->row_low,rnode->Ptc_num);

  fprintf(fp,"\n   %d edge(s):",rnode->num_edges);
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();
  for(;redge!=0;redge=rnode->getnext_fanoutList())
    fprintf(fp," %s",redge->destNode->name);
  rnode->remove_fanoutListItr();

  fprintf(fp,"\nSwitch types:");
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();
  for(;redge!=0;redge=rnode->getnext_fanoutList())
    fprintf(fp," %d",redge->type);
  rnode->remove_fanoutListItr();

  fprintf(fp,"\nOcc: %d  Capacity: %d.",rnode->occupancy,rnode->capacity);
/*   if(rnode->R > 0.0 || rnode->C > 0.0) */
/*     ERROR("um, output_pin node has non-zero R and/or C values..."); */
  fprintf(fp,"\nR: 0  C: 0");

  if(rnode->prevClusterNum==0)
    ERROR("output pin is not assigned to any cluster...");

  fprintf(fp,"\nCluster_num: %d",*(rnode->prevClusterNum));
  fprintf(fp,"\nCost_index: %d\n\n",rnode->Cost_index);

}

void writeOutWireNode(routingNode* rnode,FILE* fp){
  routingEdge* redge;

  fprintf(fp,"Node: %s.  ",rnode->name);
  if(rnode->isCHANX)
    fprintf(fp,"CHANX ");
  else
    fprintf(fp,"CHANY ");

  if(rnode->row_low==rnode->row_high && rnode->col_low==rnode->col_high)
    fprintf(fp,"(%d, %d)  Ptc_num: %d",rnode->col_low,rnode->row_low,rnode->Ptc_num);
  else
    fprintf(fp,"(%d, %d) to (%d, %d)  Ptc_num: %d",rnode->col_low,rnode->row_low,rnode->col_high,rnode->row_high,rnode->Ptc_num);

  fprintf(fp,"\n   %d edge(s):",rnode->num_edges);
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();
  for(;redge!=0;redge=rnode->getnext_fanoutList())
    fprintf(fp," %s",redge->destNode->name);
  rnode->remove_fanoutListItr();

  fprintf(fp,"\nSwitch types:");
  rnode->create_fanoutListItr();redge=rnode->getnext_fanoutList();
  for(;redge!=0;redge=rnode->getnext_fanoutList())
    fprintf(fp," %d",redge->type);
  rnode->remove_fanoutListItr();

  fprintf(fp,"\nOcc: %d  Capacity: %d.",rnode->occupancy,rnode->capacity);
/*   if(rnode->R < 0.0 || rnode->C < 0.0) */
/*     ERROR("um, wire node has -ve R and/or C values..."); */
  fprintf(fp,"\nR: 0  C: 0");

  if(rnode->prevClusterNum==0)
    ERROR("wire node is not assigned to any cluster...");

  fprintf(fp,"\nCluster_num: %d",*(rnode->prevClusterNum));
  fprintf(fp,"\nCost_index: %d\n\n",rnode->Cost_index);

}

void populateEstimateTables(){
  Cluster* cluster;routingNode* rnode;
  priority_queue<pqNode>* PQ;

  for(int i=0;i<clusterVec->size();i++){
    cluster=clusterVec->at(i);
    if(cluster==0)
      continue;

    for(int j=0;j<(rrgraph.sinkNodeVector)->size();j++)
      cluster->sinkVector->at((rrgraph.sinkNodeVector)->at(j)->sinkIndex)=(rrgraph.sinkNodeVector)->at(j);

    PQ=new priority_queue<pqNode>;
    cluster->create_memberListItr();
    for(rnode=cluster->getnext_memberList();rnode!=0;rnode=cluster->getnext_memberList())
      PQ->push(pqNode(rnode,0.,0));
    cluster->remove_memberListItr();
    Dijkstra(PQ,false,cluster,0);
    delete PQ;
  }
 
}

double calculateDistanceBetween(vector<float>* rnode_entry,Cluster* cluster){
  double distance=0.0;

  if(rnode_entry==0 || cluster==0)
    ERROR("NULL parameter in distance calculation routine...");

  if(rnode_entry->size()!=cluster->centroid->size())
    ERROR("table size mismatch...");

  for(int i=0;i<cluster->centroid->size();i++){
    if(cluster->centroid->at(i)<=-1.0 || rnode_entry->at(i) < -1.0)
      continue;
    distance+=pow((cluster->centroid->at(i) - rnode_entry->at(i)),2.0);
  }

  return distance;
}

bool terminateClustering(double num_cluster_changes,double num_clusterable_nodes){
  static int i;
  i++;
  if(i>20){
    /* alright, ENOUGH ALREADY!! */
    return true;
  }
  else{
    /* terminate if num nodes that change clusters is less than 2% */
    if((num_cluster_changes/num_clusterable_nodes)<0.01)
      return true;
    else
      return false;
  }
}

void performClustering(FILE* fp){
  routingNode* rnode;
  int closest_cluster=-1;
  double curr_distance=-1.0;double closest_distance=HUGE_NUMBER;
  int iteration_counter=0;
  int num_clusterable_nodes = rrgraph.numSourceNodes + rrgraph.numOutputPinNodes + rrgraph.numWireNodes;
  int num_cluster_changes = num_clusterable_nodes;

  cout<<endl<<endl<<"Begin Clustering...."<<endl;
  fprintf(fp,"\n\nBegin Clustering....\n");

  while(!terminateClustering(double(num_cluster_changes),double(num_clusterable_nodes))){
    iteration_counter++;

/*     if(iteration_counter>1) */
/*       break; */

    cout<<endl<<"Iteration "<<iteration_counter<<". Fracn Nodes that Changed Clusters: "<<(double)(num_cluster_changes)/(double)(num_clusterable_nodes);
    fprintf(fp,"\nIteration %d. Fracn Nodes that Changed Clusters: %f",iteration_counter,(double)(num_cluster_changes)/(double)(num_clusterable_nodes));

    if(iteration_counter>1){
      /* eliminate clusters that have no members */
      for(int y=0;y<clusterVec->size();y++){
/* 	cout<<endl<<"Looking at Cluster Num "<<y; */
	if(clusterVec->at(y)){
	  if((clusterVec->at(y)->memberList).size()==0){
	    delete clusterVec->at(y);
	    clusterVec->at(y)=0;
	    cout<<endl<<"\tEliminated Cluster at Index "<<y;
	    fprintf(fp,"\n\tEliminated Cluster at Index %d",y);
	  }
	}
      }

      break;

      for(int j=0;j<clusterVec->size();j++){
	if(clusterVec->at(j)!=0){
	  clusterVec->at(j)->recalculateCentroid();
	  (clusterVec->at(j)->memberList).clear();
	}
      }
    }

    num_cluster_changes=0;
    for(int i=0;i<rrgraph.numNodes;i++){
      if((rrgraph.rnodeVector)->at(i)->type!=INPUT_PIN && (rrgraph.rnodeVector)->at(i)->type!=SINK){
	rnode=rrgraph.rnodeVector->at(i);closest_distance=HUGE_NUMBER;closest_cluster=-1;
	for(int curr_cluster=0;curr_cluster<clusterVec->size();curr_cluster++){
	  if(clusterVec->at(curr_cluster)==0)
	    continue;
	  curr_distance = calculateDistanceBetween(costTable->at(rnode->nodeIndex),clusterVec->at(curr_cluster));
	  if(curr_distance<closest_distance){
	    closest_cluster=curr_cluster;
	    closest_distance=curr_distance;
	  }
	}

	/* handle the initial case here */
	if(rnode->prevClusterNum==0)
	  num_cluster_changes++;
	else if(*(rnode->prevClusterNum)!=closest_cluster)
	  num_cluster_changes++;
	clusterVec->at(closest_cluster)->addto_memberList(rnode);
	rnode->prevClusterNum=&(clusterVec->at(closest_cluster)->number);

      }
    }
  }

  /* alright, reassign the cluster ids to consecutive integers starting at 0 */
  int id_ctr=0;
  for(int i=0;i<clusterVec->size();i++){
    if(clusterVec->at(i)!=0){
      clusterVec->at(i)->number=id_ctr;
      id_ctr++;
    }
  }

  G_FinalNumClusters=id_ctr;

  cout<<endl<<endl<<"Clustering Complete. Number of Clustering Iterations = "<<iteration_counter;
  fprintf(fp,"\n\nClustering Complete. Number of Clustering Iterations = %d",iteration_counter);
}


void printClusterInfo(){

  cout<<endl<<endl<<"<<<<<<<<<<<<<< BEGIN PRINT CLUSTER INFO >>>>>>>>>>>>>>>";

  for(int i=0;i<clusterVec->size();i++){
    if(clusterVec->at(i))
      clusterVec->at(i)->print_info();
  }

  cout<<endl<<endl<<"<<<<<<<<<<<<<<< END PRINT CLUSTER INFO >>>>>>>>>>>>>>>>";

}

void pickStartCentroids(int k){
  routingNode *wire_rnode;

  int num_source_centroids = k;
  if(num_source_centroids>=rrgraph.numSourceNodes)
    ERROR("alright, there are more seeds than source nodes here...");
  for(int j=0;j<k;j++){
    wire_rnode=getNewWireConnectedToSource();
    for(int t=0;t<costTable->at(wire_rnode->nodeIndex)->size();t++)
      clusterVec->at(j)->addtoCentroid(t,costTable->at(wire_rnode->nodeIndex)->at(t));
    wire_rnode->isCentroidCandidate=false;
  }

}

routingNode* getNewWireConnectedToSource(){
  routingNode* source_rnode;routingNode* opin_node;routingNode* wire_node;
  int source_index=0;
  int opin_index=0;int wire_index=0;
  int num_output_pins=0;int num_output_pins_considered=0;
  int num_wires_considered=0;

  while(1){
    source_index=lrand48()%rrgraph.numSourceNodes;
    source_rnode=(rrgraph.sourceNodeVector)->at(source_index);

    /* we have a logic block source node here */
    if(source_rnode->isCentroidCandidate==false)
      continue;
    else{
      /* this source has not been considered earlier */
      num_output_pins=source_rnode->num_edges;
      num_output_pins_considered=0;
      source_rnode->isCentroidCandidate=false;

      while(1){
	opin_index=lrand48()%num_output_pins;
	opin_node=((source_rnode->fanoutVector).at(opin_index))->destNode;
	num_wires_considered=0;
	while(1){
	  /* if all wires have already been marked a centroid */
	  /* then go to the next output pin */
	  if(num_wires_considered==opin_node->num_edges)
	    break;

	  wire_index=lrand48()%(opin_node->num_edges);
	  wire_node=((opin_node->fanoutVector).at(wire_index))->destNode;

	  /* if this wire has already been marked a centroid */
	  /* try another wire */
	  if(wire_node->isCentroidCandidate==false){
	    num_wires_considered++;
	    continue;
	  }
	  else{
	    return wire_node;
	  }

	}
	num_output_pins_considered++;
	if(num_output_pins_considered==num_output_pins)
	  break;
      }

    }

  }

}

int determineNumClusters(){
  /* alright, this routine will need some smart heuristic to auto-determine */
  /* the number of clusters (i.e. K) in the K-means algorithm. For the time */
  /* being, let the user decide for you. */
}

void createClusterVector(int k,long int num_sink_nodes){
  clusterVec=new vector<Cluster*> (k);
  for(int i=0;i<k;i++)
    clusterVec->at(i)=new Cluster (rrgraph.numSinkNodes,num_sink_nodes,i);
}

void populateCostTable(){
  priority_queue<pqNode>* PQ;

  /* store costs in costTable */
  for(int i=0;i<rrgraph.numNodes;i++){
    if((rrgraph.rnodeVector)->at(i)->type==INPUT_PIN || (rrgraph.rnodeVector)->at(i)->type==SINK)
      continue;
    PQ = new priority_queue<pqNode>;
    PQ->push(pqNode((rrgraph.rnodeVector)->at(i),0.,0));
    Dijkstra(PQ,true,0,(rrgraph.rnodeVector)->at(i));
    delete PQ;
  }

}

void printCostTable(){
/*   // dump costTable */
/*   for(int j=0;j<costTable->size();j++){ */
/*     cout<<endl<<endl<<costTable->at(j)->at(0)->source_rnode->name<<":: "; */
/*     for(int k=0;k<(costTable->at(j))->size();k++) */
/*       cout<<"  ("<<costTable->at(j)->at(k)->sink_rnode->name<<","<<costTable->at(j)->at(k)->cost<<")"; */
/*   } */
}

void createCostTable(long int num_sink_nodes){
  long int num_source_nodes = rrgraph.numSourceNodes + rrgraph.numOutputPinNodes + rrgraph.numWireNodes;
  long int rand_num;routingNode* rand_sink;
  costTableEntry temp_obj;

/*   cout<<endl<<endl<<"Num source nodes is "<<num_source_nodes<<". Num sink nodes is "<<num_sink_nodes; */
/*   cout<<endl<<"Size of a costTableEntry object is "<<sizeof(temp_obj)<<". Size of a costTableEntry ptr is "<<sizeof(&temp_obj)<<"."<<endl; */

  costTable = new vector<vector<float> *> (num_source_nodes);
  for(int i=0;i<costTable->size();i++)
    costTable->at(i)=new vector<float> (num_sink_nodes);

  for(int j=0;j<costTable->size();j++){
    for(int k=0;k<(costTable->at(j))->size();k++)
      costTable->at(j)->at(k)=-1.0;
  }

/*   while(1){ ; } */

  /* alright, here we assign sinkIndex to a sub_sample of sinks */
  int sink_hash=0;
  for(int i=0;i<num_sink_nodes;i++,sink_hash++){
    while(1){
      rand_num=lrand48()%(rrgraph.numSinkNodes);
      rand_sink=(rrgraph.sinkNodeVector)->at(rand_num);
      if(rand_sink->sinkHash==-1){
	rand_sink->sinkHash=sink_hash;
	break;
      }
      else
	continue;
    }
  }

  for(int s=0;s<rrgraph.numNodes;s++){
    if((rrgraph.rnodeVector)->at(s)->type==INPUT_PIN || (rrgraph.rnodeVector)->at(s)->type==SINK)
      continue;

    for(int t=0;t<rrgraph.numSinkNodes;t++){
      if((rrgraph.sinkNodeVector)->at(t)->sinkHash==-1)
	continue;

/*       costTable->at((rrgraph.rnodeVector)->at(s)->nodeIndex)->at((rrgraph.sinkNodeVector)->at(t)->sinkHash)->source_rnode */
/* 	=(rrgraph.rnodeVector)->at(s); */
/*       costTable->at((rrgraph.rnodeVector)->at(s)->nodeIndex)->at((rrgraph.sinkNodeVector)->at(t)->sinkHash)->sink_rnode */
/* 	=(rrgraph.sinkNodeVector)->at(t); */
    }
  }

}

void Dijkstra(priority_queue<pqNode>* PQ,bool updateCostTable,Cluster* cluster,routingNode* source_rnode){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;

/*   cout<<endl<<endl<<"Looking at source node "<<source_rnode->name<<", index = "<<source_rnode->nodeIndex<<"."<<endl; */

  while(!PQ->empty()){

    rnode=(PQ->top()).rNode;
    if(rnode->isExplored==false){
      /* found unexplored routing node...mark it explored,set its pathcost and */
      /* predecessor routing node and break */
      rnode->isExplored=true;
      rnode->pathCost=(PQ->top()).pathCost;
      rnode->predecessor=(PQ->top()).predNode;

      /* fill in cost to sink rnode from source_rnode in costTable */
      if(rnode->get_type()==SINK){
/* 	cout<<"("<<rnode->name<<",index="<<rnode->sinkIndex<<",pathCost="<<rnode->pathCost<<")  "; */
	if(updateCostTable==true){
	  if(rnode->sinkHash>-1)
	    costTable->at(source_rnode->nodeIndex)->at(rnode->sinkHash)=rnode->pathCost;
	}
	else{
	  cluster->estimate->at(rnode->sinkIndex)=rnode->pathCost;
	  if(cluster->sinkVector->at(rnode->sinkIndex)!=rnode)
	    ERROR("sink and estimate mismatch here...");
	}
      }

      cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
      PQ->pop();
    }
    else{
      PQ->pop();
      continue;
    }
 
    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;
      if(nb_rnode->isExplored==false){
        path_cost=rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost();
	nb_path_cost=nb_rnode->pathCost;
	if(new_path_cost<nb_path_cost)
	  PQ->push(pqNode(nb_rnode,new_path_cost,rnode));
      }
    }
    rnode->remove_fanoutListItr();

  }
  /* end exploration */

  /* clean up all routing nodes that have been affected by the exploration */
  for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
    (*clean_itr)->isExplored=false;
    (*clean_itr)->inrTree=false;
    (*clean_itr)->predecessor=0;
    (*clean_itr)->pathCost=HUGE_NUMBER;
    (*clean_itr)->isRouted=false;
  }

}

/***
* Look for a routing node with name node_name. If one doesn't exist, create
* a new one and return.
***/
routingNode* find_rnode(char* rnode_name){
  routingNode* return_node;
  string rname = rnode_name;

  if(!rnode_name)
    ERROR("name of a routing node cannot be NULL...");
  if(return_node = rrgraph.found_routingNode(rnode_name))
    return return_node;
  else{
    return_node = new routingNode;
    return_node->set_name(rnode_name);
    rrgraph.addto_rnodeList(return_node);
    rrgraph.rnodeMap[rname] = return_node;
    return return_node;
  }
}


void process_sourceNode(int col,int row,int ptc_num){

  /*** col and row are VPR-specific geometric info. ***/
  /*** This information is redundant for A* purposes and is not used. ***/

/*   if(col<0 || row<0) */
/*     ERROR("array out of bounds here..."); */
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  G_currNode->set_type(SOURCE);
  G_currNode->Ptc_num=ptc_num;
  G_currNode->isCentroidCandidate=true;

}


void process_ipinNode(int col,int row,int ptc_num){

  /*** col and row are VPR-specific geometric info. ***/
  /*** This information is redundant for A* purposes and is not used. ***/

/*   if(col<0 || row<0) */
/*     ERROR("array out of bounds here..."); */
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  G_currNode->set_type(INPUT_PIN);
  G_currNode->Ptc_num=ptc_num;

}


void process_sinkNode(int col,int row,int ptc_num){

  /*** col and row are VPR-specific geometric info. ***/
  /*** This information is redundant for A* purposes and is not used. ***/

/*   if(col<0 || row<0) */
/*     ERROR("array out of bounds here..."); */
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  G_currNode->set_type(SINK);
  G_currNode->Ptc_num=ptc_num;
}


void process_opinNode(int col,int row,int ptc_num){

  /*** col and row are VPR-specific geometric info. ***/
  /*** This information is redundant for A* purposes and is not used. ***/

/*   if(col<0 || row<0) */
/*     ERROR("array out of bounds here..."); */
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  G_currNode->set_type(OUTPUT_PIN);
  G_currNode->Ptc_num=ptc_num;
}


void process_chanNode(int col_low,int row_low,int col_high,int row_high,int ptc_num){

  /*** col_low, col_high, row_low and row_high are VPR-specific. ***/
  /*** This information is redundant for A* purposes and is not used. ***/

/*   if(col_low<0 || col_high<0) */
/*     ERROR("array out of bounds here..."); */
/*   if(row_low<0 || row_high<0) */
/*     ERROR("array out of bounds here..."); */
  G_currNode->row_low=row_low;G_currNode->row_high=row_high;
  G_currNode->col_low=col_low;G_currNode->col_high=col_high;

  G_currNode->set_type(BUS_OR_WIRE);
  G_currNode->isCentroidCandidate=true;
  G_currNode->Ptc_num=ptc_num;
}

void process_Edges(int num_edges){
  int i;
  routingNode* fanout_node;
  routingEdge* r_edge;

  if(num_edges<0)
    ERROR("number of fanout edges cannot be -ve...");

  G_currNode->num_edges=num_edges;
  if(num_edges==0)
    return;

  // CAUTION, THIS IS A HACK THAT IS SPECIFIC TO THIS IMPLEMENTATION!!
  if(num_edges>1 && G_currNode->type==SOURCE){
    // We have the IOB source at the root of the tree.
    G_currNode->isCentroidCandidate = false;
  }

  list<char*>::iterator str_itr = stringList.begin();
  for(i=0;str_itr!=stringList.end();str_itr++,i++){
    fanout_node=find_rnode(*str_itr);
    r_edge=new routingEdge;
    r_edge->set_origNode(G_currNode);
    r_edge->set_destNode(fanout_node);

    /* increment num directed edges counter in rrgraph */
    rrgraph.numDirEdges+=1;

    G_currNode->addto_fanoutList(r_edge);

    (G_currNode->fanoutVector).push_back(r_edge);

    fanout_node->addto_faninList(r_edge);
  }

  if(i!=num_edges)
    ERROR("problems while connecting routing nodes with edges...");

}

void process_switchTypes(){
  routingEdge* r_edge;

  if(stringList.size()==0)
    return;

  list<char*>::iterator str_itr = stringList.begin();
  G_currNode->create_fanoutListItr();
  for(;str_itr!=stringList.end();str_itr++){
    r_edge = G_currNode->getnext_fanoutList();
    r_edge->set_type(atoi(*str_itr));
  }
  G_currNode->remove_fanoutListItr();

}


