////////////////////////////////////////////////////////////////////
// cluster.cc : Definition of Cluster member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////


#include "cluster.h"

void Cluster::addto_memberList(routingNode* r){
  if(r==0)
    ERROR("trying to add NULL routing node here...");
  memberList.push_back(r);
}

void Cluster::create_memberListItr(){
  if(memberListItr)
    ERROR("there is a stale iterator here...");
  memberListItr=new list<routingNode*>::iterator;
  *memberListItr=memberList.begin();
}


void Cluster::remove_memberListItr(){
  if(memberListItr==0)
    ERROR("trying to remove a NULL iterator here...");
  delete memberListItr;
  memberListItr=0;
}

routingNode* Cluster::getnext_memberList(){
  if(memberListItr==0)
    ERROR("NULL list iterator here...");

  if(*memberListItr == memberList.end())
    return 0;
  else{
    routingNode* ret_ptr = **memberListItr;
    (*memberListItr)++;
    return ret_ptr;
  }
}

void Cluster::recalculateCentroid(){
/* This function recalculates the centroid of this cluster.*/
/* Currently, the calculation is simply a mean across the  */
/* members of the cluster. */ 

  list<routingNode*>::iterator memlist_itr = memberList.begin();
  vector<float>* member_table_entry;
  double num_members = (double)(memberList.size());

  for(int i=0;i<centroid->size();i++){
    if(centroid->at(i) >= 0.0)
      centroid->at(i)=0.0;
    else if(centroid->at(i) <= -1.0){
      ;
    }
//     else{
//       cout<<endl<<centroid->at(i);
//       ERROR("dubious centroid value here...\n");
//     }
  }

  for(;memlist_itr!=memberList.end();memlist_itr++){
    member_table_entry=costTable->at((*memlist_itr)->nodeIndex);
    for(int j=0;j<member_table_entry->size();j++){
      if(member_table_entry->at(j)<=-1.0){
	if(centroid->at(j)<=-1.0)
	  continue;
	else
	  ERROR("member sink distance is -ve and centroid distance is +ve...");
      }
      else if(member_table_entry->at(j)>=1.0){
	if(centroid->at(j)<=-1.0)
	  ERROR("member sink distance is +ve and centroid distance is -ve...");
	else
	  centroid->at(j) += member_table_entry->at(j)/num_members;
      }
      else{
	cout<<endl<<member_table_entry->at(j);
	ERROR("dubious sink distance value here...\n");
      }

    }
  }   

}


void Cluster::addtoCentroid(int index,double cost){
  if(index<0)
    ERROR("array index out of bounds here...");
  centroid->at(index)=cost;
}

void Cluster::print_info(){

  cout<<endl<<endl<<"CLUSTER NUM: "<<number;
  cout<<endl<<"CENTROID:";
  for(int i=0;i<centroid->size();i++)
    cout<<"  "<<centroid->at(i);
  cout<<endl<<"ESTIMATE:";
  for(int j=0;j<estimate->size();j++)
    cout<<"  "<<estimate->at(j);

  cout<<endl<<"CLUSTER MEMBERS:";
  list<routingNode*>::iterator memlist_itr = memberList.begin();
  for(;memlist_itr!=memberList.end();memlist_itr++)
    cout<<"  "<<(*memlist_itr)->name;

}
