////////////////////////////////////////////////////////////////////
// error.h : Error-reporting functions + constant declarations.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/* Error reporting functions */

/* The next avoids multiple inclusion of header files */
#ifndef MY_OWN_ERROR_H
#define MY_OWN_ERROR_H

/* Note: since everything sources this header, we stick calls to standard
   headers here */
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stddef.h>
#include <fstream.h>
#include <iostream.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <list>
#include <vector>
#include <queue>
#include <stdexcept>
#include <functional>
#include <map>

using namespace std;

#define NDEBUG 1
#define MAX_STR_LEN 256  //set the max length of a string
#define HUGE_NUMBER 1000000.0  //the value of a HUGE floating point initializer
/* #define PARSER_DEBUG 1 */  //switch on parser debugging
/* #define ROUTER_DEBUG 1 */  //switch on router debugging
/* #define PLACER_DEBUG 1 */ //switch on placer debugging

/* define types of routing nodes */
#define SOURCE 29
#define SINK 30
#define BUS_OR_WIRE 31
#define INPUT_PIN 32
#define OUTPUT_PIN 33

/* define type of routing edges */
#define BUFFERED 41
#define UNBUFFERED 43
#define NO_TYPE -1  //default value for logic block functionality type

/* Use ERROR for internal errors */
void _Error(char *str, int line, char* file);
#define ERROR(string) _Error((string), __LINE__, __FILE__)

/* Use EXT_ERROR for user (external) errors */
void _Ext_Error(char *str);
#define EXT_ERROR(string) _Ext_Error((string))

/* Use ASSERT for invariant tests, to be turned off in production code */
inline void _Assert(int test, int line, char *file)
{
  if (!test) 
    _Error("Assertion failed", line, file);
}

/* Set NDEBUG = 1 when NOT debugging - removes ASSERT statements */
#if defined(NDEBUG)
#define ASSERT(test)
#else
#define ASSERT(test) _Assert((test), __LINE__, __FILE__)
#endif

#if defined(GPLUSPLUS)
extern "C" double pow(double x, double y);
extern "C" long random();
extern "C" void srandom(int seed);
extern "C" char *strdup(char *string);
extern "C" void getrusage(int who, struct rusage *rusage);
#endif

#define RANDOM(x) ((int)(random()%(x)))

#endif

