////////////////////////////////////////////////////////////////////
// routingGraph.cc : Definition of routingGraph member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////


#include "routingGraph.h"

void routingGraph::addto_rnodeList(routingNode* r){
  if(r==0)
    ERROR("adding NULL routing node to routingGraph...");
  rnodeList.push_back(r);
  numNodes++;
}

void routingGraph::calculateNodeTypes(){
  list<routingNode*>::iterator rnodelist_itr = rnodeList.begin();
  for(;rnodelist_itr!=rnodeList.end();rnodelist_itr++){
    if((*rnodelist_itr)->get_type()==SOURCE)
      numSourceNodes++;
    else if((*rnodelist_itr)->get_type()==SINK)
      numSinkNodes++;
    else if((*rnodelist_itr)->get_type()==INPUT_PIN)
      numInputPinNodes++;
    else if((*rnodelist_itr)->get_type()==OUTPUT_PIN)
      numOutputPinNodes++;
    else if((*rnodelist_itr)->get_type()==BUS_OR_WIRE)
      numWireNodes++;
    else
      ERROR("routing node of unknown type here...");
  }
}

void routingGraph::createNodeVectors(){

  if(numNodes==0 || numSourceNodes==0 || numSinkNodes==0 || numInputPinNodes==0 || numOutputPinNodes==0 || numWireNodes==0)
    ERROR("number of a certain routing-node type is 0...");

  rnodeVector=new vector<routingNode*> (numNodes);
  sourceNodeVector=new vector<routingNode*> (numSourceNodes);
  sinkNodeVector=new vector<routingNode*> (numSinkNodes);
  inputPinNodeVector=new vector<routingNode*> (numInputPinNodes);
  outputPinNodeVector=new vector<routingNode*> (numOutputPinNodes);
  wireNodeVector=new vector<routingNode*> (numWireNodes);

  list<routingNode*>::iterator rnodelist_itr = rnodeList.begin();
  int source_ctr=0;int sink_ctr=0;int inputpin_ctr=0;int outputpin_ctr=0;int wire_ctr=0;int node_index_ctr=0;
  for(int i=0;rnodelist_itr!=rnodeList.end();rnodelist_itr++,i++){
    rnodeVector->at(i)=*rnodelist_itr;

    if((*rnodelist_itr)->get_type()==SOURCE){
      sourceNodeVector->at(source_ctr)=*rnodelist_itr;
      source_ctr++;

      (*rnodelist_itr)->nodeIndex=node_index_ctr;
      node_index_ctr++;

    }
    else if((*rnodelist_itr)->get_type()==SINK){
      sinkNodeVector->at(sink_ctr)=*rnodelist_itr;
      (*rnodelist_itr)->sinkIndex=sink_ctr;
      sink_ctr++;
    }
    else if((*rnodelist_itr)->get_type()==INPUT_PIN){
      inputPinNodeVector->at(inputpin_ctr)=*rnodelist_itr;
      inputpin_ctr++;
    }
    else if((*rnodelist_itr)->get_type()==OUTPUT_PIN){
      outputPinNodeVector->at(outputpin_ctr)=*rnodelist_itr;
      outputpin_ctr++;

      (*rnodelist_itr)->nodeIndex=node_index_ctr;
      node_index_ctr++;

    }
    else if((*rnodelist_itr)->get_type()==BUS_OR_WIRE){
      wireNodeVector->at(wire_ctr)=*rnodelist_itr;
      wire_ctr++;

      (*rnodelist_itr)->nodeIndex=node_index_ctr;
      node_index_ctr++;

    }
    else
      ERROR("routing node of unknown type here...");
  }

}

// routingNode* routingGraph::found_routingNode(char* rname){
//   if(rname==0)
//     ERROR("looking for routing node with NULL name...");

//   list<routingNode*>::iterator rnodelist_itr = rnodeList.begin();
//   for(;rnodelist_itr!=rnodeList.end();rnodelist_itr++){
//     if(!strcmp(rname,(*rnodelist_itr)->get_name()))
//       return *rnodelist_itr;
//   }

//   return 0;
// }

routingNode* routingGraph::found_routingNode(char* rname){
  routingNode* rnode;

  if(rname==0)
    ERROR("looking for a routing node with NULL name...");

  string rn = rname;
  rnode = rnodeMap[rn];

  return rnode;

}

void routingGraph::print_info(){

  list<routingNode*>::iterator rnodelist_itr = rnodeList.begin();
  for(;rnodelist_itr!=rnodeList.end();rnodelist_itr++)
    (*rnodelist_itr)->print_info();

  cout<<endl<<endl<<"Num Routing Nodes = "<<numNodes<<". Num Routing Edges = "<<numDirEdges;
  cout<<endl<<"Num Source Nodes = "<<numSourceNodes;
  cout<<endl<<"Num Sink Nodes = "<<numSinkNodes;
  cout<<endl<<"Num InputPin Nodes = "<<numInputPinNodes;
  cout<<endl<<"Num OutputPin Nodes = "<<numOutputPinNodes;
  cout<<endl<<"Num Wire Nodes = "<<numWireNodes;
  cout<<endl;

}
