/**********************************************************************
 rrgraphLexer.l : Lex file for parsing the input architecture file.
 Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
 University of Washington.

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 The GNU General Public License can be found at 
 http://www.gnu.org/licenses/gpl.txt
***********************************************************************/

%{
#include <strings.h>
#include <iostream.h>
#include "kMeans.tab.h"
%}

%%
"("                     {}
")"                     {}
","                     {}
([0-9]+\.[0-9]+)([eE][-+]?[0-9]+)?  {yylval.double_num=atof(yytext); return DOUBLENUM_TOK; }
"."                     {yylval.word=strdup(yytext); return PERIOD_TOK; }
" to "                  {yylval.word=strdup(yytext); return TO_TOK; }
"Node:"                 {yylval.word=strdup(yytext); return NODE_TOK; }
"CHANY"                 {yylval.word=strdup(yytext); return CHANY_TOK; }
"CHANX"                 {yylval.word=strdup(yytext); return CHANX_TOK; }
"SOURCE"                {yylval.word=strdup(yytext); return SOURCE_TOK; }
"SINK"                  {yylval.word=strdup(yytext); return SINK_TOK; }
"IPIN"                  {yylval.word=strdup(yytext); return IPIN_TOK; }
"OPIN"                  {yylval.word=strdup(yytext); return OPIN_TOK; }
"Ptc_num:"              {yylval.word=strdup(yytext); return PTC_TOK; }
"edge(s):"              {yylval.word=strdup(yytext); return EDGES_TOK; }
"Switch types:"         {yylval.word=strdup(yytext); return SWITCHTYPE_TOK; }
"Occ:"                  {yylval.word=strdup(yytext); return OCC_TOK; }
"Capacity:"             {yylval.word=strdup(yytext); return CAPACITY_TOK; }
"R:"                    {yylval.word=strdup(yytext); return R_TOK; }
"C:"                    {yylval.word=strdup(yytext); return C_TOK; }
"Cost_index:"           {yylval.word=strdup(yytext); return COSTINDEX_TOK; }
"["[a-zA-Z0-9_:<>]+"]" { /* printf("found net...%s\n",yytext); */
                          yylval.word=strdup(yytext); return NET_TOK; }
[a-zA-Z0-9_:<>]+       { /* printf("found net...%s\n",yytext); */
                          yylval.word=strdup(yytext); return NET_TOK; }
[\t\n ]+                { /* printf("found whitespace...%s\n",yytext); */}
"\\"                    {}
%%

void yyerror(char* s){
  fprintf(stderr,"%s AND ERRONEOUS TOKEN IS %s\n",s,yytext);
}

