////////////////////////////////////////////////////////////////////
// cluster.h : Class that describes an interconnect cluster based on
//             a K-Means clustering implementation.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#ifndef CLUSTER_H
#define CLUSTER_H

#include "error.h"
#include "routingNode.h"

class Cluster{
 public:
  int number;  //Cluster's integer ID
  vector<double>* centroid;  //Centroid of this cluster's members
  vector<double>* estimate;  //A* estimate of this cluster's members
  vector<routingNode*>* sinkVector;  //vector of sinks. sink at index 'i' corresponds to cost in estimate->at(i).
  list<routingNode*> memberList;  //List containing cluster members 
  list<routingNode*>::iterator* memberListItr;  //Iterator for memberList

  Cluster(int sinkvec_length,int vec_length,int num){
    if(num<0 || vec_length<0)
      ERROR("bad constructor parameters...");

    memberListItr=0;
    number=num;
    centroid=new vector<double> (vec_length);
    estimate=new vector<double> (sinkvec_length);
    sinkVector=new vector<routingNode*> (sinkvec_length);
    for(int i=0;i<sinkvec_length;i++){
      estimate->at(i)=-1.0;
      sinkVector->at(i)=0;
    }
    for(int i=0;i<vec_length;i++)
      centroid->at(i)=-1.0;
  }

  Cluster(){
    number=-1;
    centroid=estimate=0;
    memberListItr=0;
    sinkVector=0;
  }
  ~Cluster(){
    delete centroid;delete estimate;
    delete sinkVector;
    if(memberListItr)
      delete memberListItr;
  }

  void addto_memberList(routingNode* r);
  void create_memberListItr();
  routingNode* getnext_memberList();
  void remove_memberListItr();
  void addtoCentroid(int index,double cost);
  void print_info();
  void recalculateCentroid();

};

class costTableEntry{
 public:
  routingNode* source_rnode;
  routingNode* sink_rnode;
  double cost;

  costTableEntry(){
    source_rnode=sink_rnode=0;
    cost=-1.0;
  }

};

/* Declare a cost table */
extern vector<vector<float> *>* costTable;

/* Declare vector of clusters */
extern vector<Cluster*>* clusterVec;

#endif
