////////////////////////////////////////////////////////////////////
// Terminal.h : Class the describes a logic-block terminal.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////


/****************************************************************
Class definition of a logic block terminal. A Terminal object 
contains information about the logic block that it belongs to 
and the net that it drives or is driven by. Further, a Terminal
object is also associated with a type, in case the user needs the
option of creating terminals of different types. Note that the
type is not an indicator of whether the terminal is an input or
output terminal of the logic block that it is associated with. 
The input / output information is separately stored in 'isSource'
and 'isSink' fields. Finally, a Terminal object also stores a
pointer to the source/sink routing node that it is currently
mapped to. 
****************************************************************/    
#ifndef TERMINAL_H
#define TERMINAL_H

#include "error.h"
#include "logicBlock.h"
#include "Net.h"
#include "routingNode.h"

class logicBlock;
class Net;
class routingNode;

class Terminal{
 public:
  char name [MAX_STR_LEN];  //name of terminal
  logicBlock* lBlock;  //logic block to which this terminal belongs
  Net* net;  //the net that this terminal drives or is driven by
  routingNode* rNode;  //the routing node that this terminal is currently mapped to
  int terminalType;  //type of terminal. User can define types in error.h
  bool isSink;  //whether this is a sink terminal of net
  bool isSource;  //whether this is the source terminal of net
  int magicNum;

  Terminal(){
    magicNum=16111;
    lBlock=0;
    net=0;
    terminalType=NO_TYPE;
    isSink=isSource=false;
    rNode=0;
  }

  ~Terminal(){
    if(magicNum!=16111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //give this terminal a name
  void set_lBlock(logicBlock* l);  //set logic block associated with this terminal
  void set_net(Net* n);  //net that this terminal drives or is driven by
  void set_terminalType(int t);  //set the type of this terminal
  void set_isSink(bool u);  //whether this is a sink terminal of net
  void set_isSource(bool u);  //whether this is the source terminal of net
  void set_rNode(routingNode* r);  //set the routing node that this terminal is mapped to

  char* get_name();  //get name of this terminal
  logicBlock* get_lBlock();  //get logic block associated with this terminal
  Net* get_net();  //get net that this terminal drives or is driven by
  int get_terminalType();  //get terminal type
  bool get_isSink();  //whether this is a sink terminal
  bool get_isSource();  //whether this is a source terminal
  routingNode* get_rNode();  //get the routing node that this terminal is mapped to

  void print_Info();  //print details about this terminal

};

#endif


