////////////////////////////////////////////////////////////////////
// mcncClb.cc: Definition of mcncClb member functions. 
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "mcncClb.h"

// Set name of clb
void mcncClb::set_name(char* n){
  if(!n || strlen(n)>MAX_STR_LEN)
    ERROR("error in naming clb...");
  if(magicNum!=12111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Add a net to inputList
void mcncClb::addto_inputList(mcncNet* p_net){
  if(!p_net)
    ERROR("cannot add NULL input net...");
  if(magicNum!=12111)
    ERROR("trying to access a deleted node...");
  inputList.push_back(p_net);
  numInputNets++;
}

// Add a net to outputList
void mcncClb::addto_outputList(mcncNet* p_net){
  if(!p_net)
    ERROR("cannot add NULL output net...");
  if(magicNum!=12111)
    ERROR("trying to access a deleted node...");
  outputList.push_back(p_net);
  numOutputNets++;
}

// Set type to clb
void mcncClb::set_isClb(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=12111)
    ERROR("trying to access a deleted node...");
  isClb=u;
}

// Set type to output block
void mcncClb::set_isOutputBlock(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=12111)
    ERROR("trying to access a deleted node...");
  isOutputBlock=u;
}

// Set type to input block
void mcncClb::set_isInputBlock(bool u){
  if(u!=false && u!=true)
    ERROR("non-bool value...");
  if(magicNum!=12111)
    ERROR("trying to access a deleted node...");
  isInputBlock=u;
}

// Get name of clb
char* mcncClb::get_name(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  return name;
}

// Check whether this is a clb
bool mcncClb::get_isClb(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  return isClb;
}

// Check whether this is an output block
bool mcncClb::get_isOutputBlock(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  return isOutputBlock;
}

// Check whether this is an input block
bool mcncClb::get_isInputBlock(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  return isInputBlock;
}  

// Get num input nets
int mcncClb::get_numInputNets(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  return inputList.size();
}

// Get num output nets
int mcncClb::get_numOutputNets(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  return outputList.size();
}

// Print info about this clb. Print name, input and output nets.
void mcncClb::print_ClbInfo(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");

  cout<<endl<<endl<<"BEGIN PRINT INFO FOR CLB "<<name;

  //print type
  if(isOutputBlock)
    cout<<endl<<"OUTPUT BLOCK";
  if(isInputBlock)
    cout<<endl<<"INPUT BLOCK";
  if(isClb)
    cout<<endl<<"LOGIC BLOCK";

  //print input list
  cout<<endl<<"INPUT NETS -->";
  list<mcncNet*>::iterator input_itr = inputList.begin();
  for(;input_itr!=inputList.end();input_itr++)
    cout<<endl<<"\t"<<(*input_itr)->get_name();

  //print output list
  cout<<endl<<"OUTPUT NETS -->";
  list<mcncNet*>::iterator output_itr = outputList.begin();
  for(;output_itr!=outputList.end();output_itr++)
    cout<<endl<<"\t"<<(*output_itr)->get_name();

}

// Create a new iterator for inputList
void mcncClb::create_inputListItr(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  if(input_itr)
    ERROR("an iterator already exists...");
  input_itr = new list<mcncNet*>::iterator;
  *input_itr = inputList.begin();
}

// Remove an already existing inputList iterator
void mcncClb::remove_inputListItr(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  if(input_itr == 0)
    ERROR("cannot delete NULL iterator...");
  delete input_itr;
  input_itr = 0;
}

// Get next element from inputList
mcncNet* mcncClb::getnext_inputList(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");

  if(*input_itr == inputList.end())
    return 0;
  else{
    mcncNet* ret_ptr = **input_itr;
    (*input_itr)++;
    return ret_ptr;
  }

}

// Create a new iterator for outputList
void mcncClb::create_outputListItr(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  if(output_itr)
    ERROR("an iterator already exists...");
  output_itr = new list<mcncNet*>::iterator;
  *output_itr = outputList.begin();
}

// Remove an already existing outputList iterator
void mcncClb::remove_outputListItr(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");
  if(output_itr == 0)
    ERROR("cannot delete NULL iterator...");
  delete output_itr;
  output_itr = 0;
}

// Get next element from outputList
mcncNet* mcncClb::getnext_outputList(){
  if(magicNum!=12111)
    ERROR("trying to access deleted node...");

  if(*output_itr == outputList.end())
    return 0;
  else{
    mcncNet* ret_ptr = **output_itr;
    (*output_itr)++;
    return ret_ptr;
  }

}

