////////////////////////////////////////////////////////////////////
// mcncNet.h: Class that describes a net in a VPR-style mcnc netlist.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition of a net as found in an mcnc netlist. Each net
has a clb source and possibly multiple clb sinks. Note that input
and output blocks are considered clbs here.
****************************************************************/ 
#ifndef MCNC_NET_H
#define MCNC_NET_H

#include "error.h"
#include "mcncClb.h"

class mcncClb;

class mcncNet{
  char name [MAX_STR_LEN];  //name of net
  list<mcncClb*> sinkList;  //list of sink clbs
  mcncClb* source;  //source clb
  int numSinks;  //num sink clbs
  int magicNum;
 public:
  mcncNet(){
    source=0;
    numSinks=0;
    magicNum=13111;
  }

  ~mcncNet(){
    if(magicNum!=13111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //set name of net
  void addto_sinkList(mcncClb* p_clb);  //add a clb to sinkList
  void set_source(mcncClb* p_clb);  //set source of net

  char* get_name();  //get name of net
  mcncClb* get_source();  //get source of net
  int get_numSinks();  //get num of sink clbs

  void print_NetInfo();  //print info about this net
};

#endif


