////////////////////////////////////////////////////////////////////
// mcncNetlist.cc: Definitions of mcncNetlist member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "mcncNetlist.h"

// Set name of mcnc netlist to n.
void mcncNetlist::set_name(char* n){
  if(!n || strlen(n)>MAX_STR_LEN)
    ERROR("error in naming netlist...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  strcpy(name,n);
}

// Add p_clb to list of clbs.
// This function assumes that p_clb has been updated
// to reflect whether it is a logic, input-only or
// output-only block. 
void mcncNetlist::addto_mcncClbList(mcncClb* p_clb){
  if(!p_clb)
    ERROR("clb cannot be NULL...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  mcncClbList.push_back(p_clb);

  if(p_clb->get_isClb())
    numClbs++;
  else if(p_clb->get_isInputBlock())
    numInputs++;
  else if(p_clb->get_isOutputBlock())
    numOutputs++;
}

// Add p_net to list of nets and increment num_nets.
void mcncNetlist::addto_mcncNetList(mcncNet* p_net){
  if(!p_net)
    ERROR("net cannot be NULL...");
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  mcncNetList.push_back(p_net);
  numNets++;
}

// Get total num clbs in this netlist.
int mcncNetlist::get_numClbs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numClbs;
}

// Get total num input-only IO blocks in this netlist.
int mcncNetlist::get_numInputs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numInputs;
}

// Get total num output-only IO blocks in this netlist.
int mcncNetlist::get_numOutputs(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numOutputs;
}

// Get total num nets in this netlist.
int mcncNetlist::get_numNets(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return numNets;
}

// Get the name of this netlist
char* mcncNetlist::get_name(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  return name;
}

// Print details of each clb in ClbList.
void mcncNetlist::print_mcncClbList(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT CLB LIST";
  cout<<endl<<"***************************************************************";

  list<mcncClb*>::iterator clb_itr = mcncClbList.begin();
  for(;clb_itr!=mcncClbList.end();clb_itr++)
    (*clb_itr)->print_ClbInfo();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT CLB LIST";
  cout<<endl<<"***************************************************************";

}

// Print details of each net in NetList.
void mcncNetlist::print_mcncNetList(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tBEGIN PRINT NET LIST";
  cout<<endl<<"***************************************************************";

  list<mcncNet*>::iterator net_itr = mcncNetList.begin();
  for(;net_itr!=mcncNetList.end();net_itr++)
    (*net_itr)->print_NetInfo();

  cout<<endl<<endl<<"***************************************************************";
  cout<<endl<<"\t\tEND PRINT NET LIST";
  cout<<endl<<"***************************************************************";

}

// Print the name of this mcnc netlist.
void mcncNetlist::print_name(){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  cout<<endl<<endl<<"NAME OF NETLIST IS "<<name;
}

// Print the details of clb pointed to by p_clb.
void mcncNetlist::print_mcncClb(mcncClb* p_clb){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!p_clb)
    ERROR("clb cannot be NULL...");
  p_clb->print_ClbInfo();
}

// Print the details of net pointed to by p_net.
void mcncNetlist::print_mcncNet(mcncNet* p_net){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!p_net)
    ERROR("net cannot be NULL...");
  p_net->print_NetInfo();
}

// Look for clb with clb_name in ClbList. If found, return the clb.
mcncClb* mcncNetlist::found_mcncClb(char* clb_name){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!clb_name)
    ERROR("clb name cannot be NULL...");
  list<mcncClb*>::iterator clb_itr = mcncClbList.begin();
  for(;clb_itr!=mcncClbList.end();clb_itr++){
    if(strcmp((*clb_itr)->get_name(),clb_name)==0)
      return *clb_itr;
  }
  return 0;
}

// Look for net with net_name in NetList. If found, return the net.
mcncNet* mcncNetlist::found_mcncNet(char* net_name){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(!net_name)
    ERROR("net name cannot be NULL...");
  list<mcncNet*>::iterator net_itr = mcncNetList.begin();
  for(;net_itr!=mcncNetList.end();net_itr++){
    if(strcmp((*net_itr)->get_name(),net_name)==0)
      return *net_itr;
  }
  return 0;
}

// Convert this mcnc netlist to a generic netlist.
// This function will have to be defined and called by a user routine.
// The objective here is to convert the user's netlist data-structs
// to the genric format that Independence will work with.
void mcncNetlist::convert2Generic(Netlist& n){
  Net* out_net;
  Net* in_net;
  mcncNet* mcnc_net;
  Terminal* output_terminal;
  Terminal* input_terminal;
  logicBlock* lb;
  int y;

  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");

  // give the generic netlist a name
  n.set_name(name);

  // for each clb, create a new logic block in generic netlist
  list<mcncClb*>::iterator clb_itr = mcncClbList.begin();
  for(;clb_itr!=mcncClbList.end();clb_itr++){
    lb = new logicBlock;

    // give the logic block a name
    lb->set_name((*clb_itr)->get_name());

    // determine whether we have an IO or logic block
    if((*clb_itr)->get_isClb()){
      lb->set_isLogic(true);
      lb->set_functionType(MCNC_CLB_TYPE);
    }
    if((*clb_itr)->get_isOutputBlock()){
      lb->set_isOutput(true);
      lb->set_functionType(MCNC_IO_TYPE);
    }
    if((*clb_itr)->get_isInputBlock()){
      lb->set_isInput(true);
      lb->set_functionType(MCNC_IO_TYPE);
    }

    // loop thru the input nets of the mcnc clb
    (*clb_itr)->create_inputListItr();
    mcnc_net=(*clb_itr)->getnext_inputList();
    for(y=0;mcnc_net!=0;mcnc_net=(*clb_itr)->getnext_inputList(),y++){
      // find input net
      in_net = n.found_Net(mcnc_net->get_name());
      if(!in_net){
        in_net = createNet(mcnc_net);
        n.addto_netList(in_net);
      }
      // create a new input terminal
      input_terminal = createTerminal(lb,in_net,MCNC_INPUT_TERMINAL_TYPE,false,y);
      lb->addto_inputList(input_terminal);
      in_net->addto_sinkList(input_terminal);
    }
    (*clb_itr)->remove_inputListItr();

    // loop thru the output nets of the mcnc clb
    (*clb_itr)->create_outputListItr();
    mcnc_net=(*clb_itr)->getnext_outputList();
    for(y=0;mcnc_net!=0;mcnc_net=(*clb_itr)->getnext_outputList(),y++){
      // find output net
      out_net = n.found_Net(mcnc_net->get_name());
      if(!out_net){
        out_net = createNet(mcnc_net);
        n.addto_netList(out_net);
      }
      // create a new output terminal
      output_terminal = createTerminal(lb,out_net,MCNC_OUTPUT_TERMINAL_TYPE,true,y);
      lb->addto_outputList(output_terminal);
      out_net->set_source(output_terminal);
    }
    (*clb_itr)->remove_outputListItr();

    // add lb to list of logicBlocks in n
    n.addto_lbList(lb);

  }

  /* now create vector of logic blocks */
  n.create_lbVector();

  /* for each net, create vector of sinks, and populate */
  n.create_netSinkVectors();

}

// this function creates a new Net object based on the mcncNet pointed
// to by n. 
Net* mcncNetlist::createNet(mcncNet* n){
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(n==0)
    ERROR("NULL mcnc net here...");
  Net* new_net = new Net;
  new_net->set_name(n->get_name());
  return new_net;
}

// create a new Terminal object. this object represents a terminal of the logicBlock
// lb which is a source / sink of tnet. term_type is a user-defined terminal type,
// is_source indicates whether this terminal is a source or sink of tnet, and 
// term_num is used to derive the name of this terminal.
Terminal* mcncNetlist::createTerminal(logicBlock* lb,Net* tnet,int term_type,
                                           bool is_source,int term_num){
  char term_name [MAX_STR_LEN];
  if(magicNum!=11111)
    ERROR("trying to access a deleted node...");
  if(lb==0)
    ERROR("NULL logic block...");
  if(tnet==0)
    ERROR("NULL net here...");
  if(term_type!=MCNC_INPUT_TERMINAL_TYPE && term_type!=MCNC_OUTPUT_TERMINAL_TYPE)
    ERROR("faulty terminal types...");

  Terminal* new_terminal = new Terminal;

  if(term_type==MCNC_INPUT_TERMINAL_TYPE)
    sprintf(term_name,"%s_INPUT%d",lb->get_name(),term_num);
  else
    sprintf(term_name,"%s_OUTPUT%d",lb->get_name(),term_num);

  new_terminal->set_name(term_name);
  new_terminal->set_lBlock(lb);
  new_terminal->set_net(tnet);
  new_terminal->set_terminalType(term_type);
  if(is_source)
    new_terminal->set_isSource(true);
  else
    new_terminal->set_isSink(true);
  return new_terminal;

}


