////////////////////////////////////////////////////////////////////
// mcncNetlist.h: Class that describes a VPR-style mcnc netlist.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

/****************************************************************
Class definition for an mcnc netlist. Note that this class 
definition is specific to mcnc netlists. For every new benchmark
set, a separate class definition will have to be created and
implemented.
****************************************************************/ 
#ifndef MCNC_NETLIST_H
#define MCNC_NETLIST_H

#include "error.h"
#include "mcncClb.h"
#include "mcncNet.h"

/* include generic netlist here so that member functions of mcncNetlist */
/* can populate the generic netlist */
#include "Netlist.h"

class mcncNetlist{
  char name [MAX_STR_LEN];  //name of mcnc netlist
  list<mcncClb*> mcncClbList;  //list of clbs in mcnc netlist
  list<mcncNet*> mcncNetList;  //list of nets in mcnc netlist
  int numClbs;  //num clbs in mcnc netlist
  int numInputs;  //num input-only IO blocks
  int numOutputs;  //num output-only IO blocks
  int numNets;  //num nets in mcnc netlist
  int magicNum;
 public:
  mcncNetlist(){
    numClbs=0;
    numNets=0;
    numInputs=numOutputs=0;
    magicNum=11111;
  }

  ~mcncNetlist(){
    if(magicNum!=11111)
      ERROR("trying to double delete a node...");
    else
      magicNum=0;
  }

  void set_name(char* n);  //set the name of mcnc netlist
  void addto_mcncClbList(mcncClb* p_clb);  //add a clb to mcncClbList
  void addto_mcncNetList(mcncNet* p_net);  //add a net to mcncNetList

  int get_numClbs();  //get num clbs in mcnc netlist
  int get_numNets();  //get num nets in mcnc netlist
  char* get_name();  //get the name of mcnc netlist
  int get_numInputs();  //total num input-only IO blocks
  int get_numOutputs();  //total num output-only IO blocks

  void print_mcncClbList();  //print the entire list of clbs
  void print_mcncNetList();  //print the entire list of nets
  void print_name();  //print the name of mcnc netlist
  void print_mcncClb(mcncClb* p_clb);  //print the details of clb pointed to by p_clb
  void print_mcncNet(mcncNet* p_net);    //print the details of net pointed to by p_net

  mcncClb* found_mcncClb(char* clb_name);  //look for clb in ClbList. if found, return clb.
  mcncNet* found_mcncNet(char* net_name);  //look for net in NetList. if found, return net.

  void convert2Generic(Netlist& n);  //convert this mcnc netlist to a generic netlist
  Net* createNet(mcncNet* n);  //create a new Net object based on the mcncNet pointed
                               //to by n.
  Terminal* createTerminal(logicBlock* lb,Net* tnet,int term_type,bool is_source,
                           int term_num);  //create a new Terminal object


};

#endif


