////////////////////////////////////////////////////////////////////
// mcncParser.y: This is the file in which the yacc grammar resides.
//               Further, all main placement and routing routines
//               are resident here.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

%{
#include "mcncParser.h"
%}

%error-verbose

%union{
  char* word;
  double double_num;
}

%token <word> INPUT_TOK 
%token <word> OUTPUT_TOK 
%token <word> CLB_TOK 
%token <word> SUBBLOCK_TOK 
%token <word> PINLIST_TOK 
%token <word> NET_TOK
%token <word> GLOBAL_TOK
%token <word> ANNOTATION_TOK

%token <word> PERIOD_TOK
%token <word> TO_TOK
%token <word> NODE_TOK
%token <word> CHANY_TOK
%token <word> CHANX_TOK
%token <word> SOURCE_TOK
%token <word> SINK_TOK
%token <word> IPIN_TOK
%token <word> OPIN_TOK
%token <word> PTC_TOK
%token <word> EDGES_TOK
%token <word> SWITCHTYPE_TOK
%token <word> OCC_TOK
%token <word> CAPACITY_TOK
%token <word> R_TOK
%token <word> C_TOK
%token <word> COSTINDEX_TOK
%token <double_num> DOUBLENUM_TOK
%token <word> CLUSTER_TOK
%token <word> NUM_CLUSTERS_TOK
%token <word> NUM_SINKS_TOK
%token <word> NETLIST_TOK
%token <word> ARCH_TOK
%token <word> ARRAY_SIZE_TOK
%token <word> LOGIC_BLOCKS_TOK
%%
statement_list: statement
|    statement_list statement
;

statement: circuit_info{
  processCircuitInfo(0);
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
}
|
rr_graph_info{
  stringListItr = stringList.begin();
  for(;stringListItr!=stringList.end();stringListItr++)
    free(*stringListItr);
  stringList.clear();
}
|
cluster_info{
  ;
}
|
placement_info{
  ;
}
;

rr_graph_info:
NODE_TOK NET_TOK PERIOD_TOK { 
  G_currNode=find_rnode($2);
/*   cout<<endl<<G_currNode->name; */
  free($1);free($2);free($3);
}
| SOURCE_TOK NET_TOK NET_TOK PTC_TOK NET_TOK { 
  process_sourceNode(atoi($2),atoi($3));
  free($1);free($2);free($3);free($4);free($5);
}
| SINK_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_sinkNode(atoi($2),atoi($3));
  free($1);free($2);free($3);free($4);free($5);
}
| IPIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_ipinNode(atoi($2),atoi($3));
  free($1);free($2);free($3);free($4);free($5);
}
| OPIN_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_opinNode(atoi($2),atoi($3));
  free($1);free($2);free($3);free($4);free($5);
}
| CHANX_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($2),atoi($3),CHANX);
  free($1);free($2);free($3);free($4);free($5);
}
| CHANX_TOK NET_TOK NET_TOK TO_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($5),atoi($6),CHANX);
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);
}
| CHANY_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($2),atoi($3),CHANY);
  free($1);free($2);free($3);free($4);free($5);
}
| CHANY_TOK NET_TOK NET_TOK TO_TOK NET_TOK NET_TOK PTC_TOK NET_TOK {
  process_chanNode(atoi($2),atoi($3),atoi($5),atoi($6),CHANY);
  free($1);free($2);free($3);free($4);free($5);free($6);free($7);free($8);
}
| NET_TOK EDGES_TOK nets {
  process_Edges(atoi($1));
  free($1);free($2);
}
| SWITCHTYPE_TOK nets {
  process_switchTypes();
  free($1);
}
| OCC_TOK NET_TOK CAPACITY_TOK NET_TOK PERIOD_TOK {
  G_currNode->set_capacity(atoi($4));
  free($1);free($2);free($3);free($4);free($5);
}  
| R_TOK DOUBLENUM_TOK C_TOK DOUBLENUM_TOK { 
  G_currNode->R=$2; G_currNode->C=$4;
  free($1);free($3);
}
| CLUSTER_TOK NET_TOK { G_currNode->clusterNumber = atoi($2);free($1);free($2); }
| R_TOK NET_TOK C_TOK NET_TOK { free($1);free($2);free($3);free($4); }
| COSTINDEX_TOK NET_TOK { free($1);free($2); }
;

cluster_info:
  NUM_CLUSTERS_TOK NET_TOK { 
  numClusters = atoi($2);free($1);free($2); 
}
| NUM_SINKS_TOK NET_TOK { 
  prepareAstarTable(numClusters,atoi($2));free($1);free($2); 
}
| NET_TOK NET_TOK DOUBLENUM_TOK {
  process_AstarTableEntry(atoi($1),$2,$3);
  free($1);free($2);
}
;

placement_info:
  NETLIST_TOK NET_TOK {
  cout<<endl<<$1<<" "<<$2;
  free($1);free($2);
}
| NETLIST_TOK NET_TOK PERIOD_TOK NET_TOK {
  cout<<endl<<$1<<" "<<$2<<" "<<$3<<$4;
  free($1);free($2);free($3);free($4);
}
| ARCH_TOK NET_TOK {
  cout<<endl<<$1<<" "<<$2;
  free($1);free($2);
}
| ARCH_TOK NET_TOK PERIOD_TOK NET_TOK {
  cout<<endl<<$1<<" "<<$2<<" "<<$3<<$4;
  free($1);free($2);free($3);free($4);
}
| ARRAY_SIZE_TOK NET_TOK NET_TOK NET_TOK LOGIC_BLOCKS_TOK{
  cout<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4<<" "<<$5;
  free($1);free($2);free($3);free($4);free($5);
}
| NET_TOK NET_TOK NET_TOK NET_TOK {
/*   cout<<endl<<$1<<" "<<$2<<" "<<$3<<" "<<$4; */
  readInVPRPlacementLocation($1,atoi($2),atoi($3),atoi($4));
  free($1);free($2);free($3);free($4);
}   
;

circuit_info: keyword nets   { ; } 
;

keyword: INPUT_TOK  { /* cout<<endl<<$1; */ curr_tok = INPUT_TOK; free($1); }
|    OUTPUT_TOK     { /* cout<<endl<<$1; */ curr_tok = OUTPUT_TOK;free($1);  }
|    CLB_TOK        { /* cout<<endl<<$1; */ curr_tok = CLB_TOK; free($1);  }
|    SUBBLOCK_TOK   { /* cout<<endl<<$1; */ curr_tok = SUBBLOCK_TOK; free($1); }
|    PINLIST_TOK    { /* cout<<endl<<$1; */ curr_tok = PINLIST_TOK; free($1); }
|    GLOBAL_TOK     { /* cout<<endl<<$1; */ curr_tok = GLOBAL_TOK; free($1); }
|    ANNOTATION_TOK     { /* cout<<endl<<$1; */ curr_tok = ANNOTATION_TOK; free($1); }
;

nets:
|    NET_TOK nets   { /* cout<<" "<<$1; */ stringList.push_front($1); }
;
%%

/* input file that is to be parsed */
extern FILE* yyin;

int main(int argc,char** argv){
  int itr;
  time_t start_time,end_time;
  FILE* fp;char outfile_name [MAX_STR_LEN];
  int num_rows=-1;int num_cols=-1;int num_iobs=-1;long int random_seed=-1;
  char* echo_file_name=0;char* netlist_name=0;char* cluster_file_name=0;char* place_file_name=0;
  bool force_seed=false;bool route_only=false;

  try{

    for(int opt_ctr=1;opt_ctr<argc;opt_ctr++){
      if(!strcmp(argv[opt_ctr],"-num_rows")){
	num_rows=atoi(argv[opt_ctr+1]);
	if(num_rows<0)
	  ERROR("-num_rows must be a positive integer...");
      }
      else if(!strcmp(argv[opt_ctr],"-num_cols")){
	num_cols=atoi(argv[opt_ctr+1]);
	if(num_cols<0)
	  ERROR("-num_cols must be a positive integer...");
      }
      else if(!strcmp(argv[opt_ctr],"-num_iobs")){
	num_iobs=atoi(argv[opt_ctr+1]);
	if(num_iobs<0)
	  ERROR("-num_iobs must be a positive integer...");
      }
      else if(!strcmp(argv[opt_ctr],"-seed")){
	random_seed=atoi(argv[opt_ctr+1]);
	if(random_seed<0)
	  ERROR("-seed must be a positive integer...");
	force_seed=true;
      }
      else if(!strcmp(argv[opt_ctr],"-echo_file")){
	echo_file_name=strdup(argv[opt_ctr+1]);
	fp=fopen(echo_file_name,"r");
	if(fp==0)
	  ERROR("could not open echo_file...");
	fclose(fp);
      }
      else if(!strcmp(argv[opt_ctr],"-netlist")){
	netlist_name=strdup(argv[opt_ctr+1]);
	fp=fopen(netlist_name,"r");
	if(fp==0)
	  ERROR("could not open netlist_file...");
	fclose(fp);
      }
      else if(!strcmp(argv[opt_ctr],"-share_fac")){
	SHARE_FAC=atof(argv[opt_ctr+1]);
	if(SHARE_FAC<0)
	  ERROR("-share_fac must be a non-negative floating-point number...");
      }
      else if(!strcmp(argv[opt_ctr],"-alpha")){
	ALPHA=atof(argv[opt_ctr+1]);
	if(ALPHA<0)
	  ERROR("-alpha must be a non-negative floating-point number...");
      }
      else if(!strcmp(argv[opt_ctr],"-inner_num")){
	INNER_NUM=atof(argv[opt_ctr+1]);
	if(INNER_NUM<0)
	  ERROR("-inner_num must be a non-negative floating-point number...");
      }
      else if(!strcmp(argv[opt_ctr],"-cluster_file")){
	cluster_file_name=strdup(argv[opt_ctr+1]);
	fp=fopen(cluster_file_name,"r");
	if(fp==0)
	  ERROR("could not open cluster_file...");
	fclose(fp);
      }
      else if(!strcmp(argv[opt_ctr],"-place_file")){
	place_file_name=strdup(argv[opt_ctr+1]);
	fp=fopen(place_file_name,"r");
	if(fp==0)
	  ERROR("could not open place_file...");
	fclose(fp);
	route_only=true;
      }
      else
	continue;
    }

    if(num_rows==-1 || num_cols==-1 || num_iobs==-1 || echo_file_name==0 || netlist_name==0 || cluster_file_name==0){
      cout<<endl<<endl<<"Insufficient command-line parameters. Need at least the following command-line params -->"<<endl;
      cout<<endl<<"./independence -num_rows <num_rows> -num_cols <num_cols> -num_iobs <num_iobs> -echo_file <echo_file_name> -netlist <netlist_name> -cluster_file <cluster_file_name>";
      cout<<endl<<endl<<"Optional params -->";
      cout<<endl<<"-place_file <place_file_name>. This needs to be a VPR-generated placement file of type name.p. This option is used when you want to route a VPR placement using Independence's router.";
      cout<<endl<<"-share_fac <fp_num>. This needs to be a positive floating-point number. Only necessary if you are gonna do placement using Independence. Default is 0.5";
      cout<<endl<<"-inner_num <fp_num>. This needs to be a positive floating-point number. Only necessary if you are gonna do placement using Independence. Default is 1.0";
      cout<<endl<<"-alpha <fp_num>. This needs to be a positive floating-point number. Only necessary if you are gonna do placement using Independence. Default is 1.0";
      cout<<endl<<endl; 
      exit(1);
    }

/*     cout<<endl<<endl<<"Reading in the netlist: "<<netlist_name<<"...."; */

    /* open input file */
    yyin=fopen(netlist_name,"r");

    /* read seed from command line */
    if(force_seed)
      randomSeed=random_seed;
    else
      randomSeed=time(NULL);
    srand48(randomSeed);

    /* give the mcnc netlist its name */
    mcnc_netlist.set_name(netlist_name);

    /* parse input file */
    while(!feof(yyin)){
      yyparse();
    }

    /* close input file */
    fclose(yyin);

/*     cout<<"done!!"; */

/*     cout<<endl<<"Converting netlist to generic netlist format...."; */

    /* convert mcnc to generic internal netlist */
    convert2GenericNetlist(mcnc_netlist);

/*     cout<<"done!!"; */

/*      print_GenericNetlist(); */

    /* create vpr challenge architecture */
    create_vprFPGA(num_rows,num_cols,num_iobs);

/*     cout<<endl<<endl<<"Reading in the echo_file: "<<echo_file_name<<"...."; */

    /* parse rr_graph.echo file */
    yyin=fopen(echo_file_name,"r");
    while(!feof(yyin)){
      yyparse();
    }
    fclose(yyin);

/*     cout<<"done!!"; */

    /* connect up IO routing nodes associated with logic blocks */
    connect_IONodes();

/*      FPGA->print_info(); */

/*     cout<<endl<<endl<<"Reading in the cluster_file: "<<cluster_file_name<<"...."; */

    yyin=fopen(cluster_file_name,"r");
    while(!feof(yyin)){
      yyparse();
    }
    fclose(yyin);

/*     cout<<"done!!"; */

/*     cout<<endl<<endl<<"Building AllSinksTables...."; */

    start_time=time(0);

    populateAllSinksTables();

    end_time=time(0);

/*     cout<<"done!!"; */

/*     cout<<" Total time to build AllSinksTables is "<<(end_time-start_time)<<endl; */

/*     exit(1); */

/*     dumpAllSinksTables(); */

/*     dumpAstarTable(); */

/*     exit(1); */

    if(route_only==false){

      /* start measuring the anneal runtime */
      start_time=time(0);

      /* set up a random initial placement of the netlist on the FPGA */
      initPlacement();

      /*      printPlacement(); */

      numBlocks=netlist.get_numlogicBlocks()+netlist.get_numInputs()+netlist.get_numOutputs();
      numNets=netlist.get_numNets();
      dLimit=(float)(FPGA->get_numRows());
/*       dLimit=dLimitMax; */

      sprintf(outfile_name,"%s_%s_%ld_%f.out",netlist.get_name(),echo_file_name,randomSeed,SHARE_FAC);
      fp=fopen(outfile_name,"w");

/*       cout<<endl<<endl; */
/*       cout<<endl<<"Netlist is "<<netlist_name<<". Routing graph built from "<<echo_file_name; */
/*       cout<<endl<<"Total num blocks (LOGIC + IO) = "<<numBlocks<<" and total num nets = "<<numNets; */
/*       cout<<endl<<"Target is a "<<FPGA->get_numCols()<<"x"<<FPGA->get_numRows()<<" device."; */
/*       cout<<endl<<"Max move range is dLimitMax = "<<dLimitMax<<". Min move range is dLimitMin = "<<dLimitMin<<"."; */

      fprintf(fp,"\n\n\nNetlist is %s. Routing graph built from %s",netlist_name,echo_file_name);
      fprintf(fp,"\nTotal num blocks (LOGIC + IO) = %d and total num nets = %d",numBlocks,numNets);
      fprintf(fp,"\nTarget is a %d x %d device.",FPGA->get_numCols(),FPGA->get_numRows());
/*       fprintf(fp,"\nMax move range is dLimitMax = %.2f. Min move range is dLimitMin = %.2f",dLimitMax,dLimitMin); */
#if defined(PLACER_DEBUG)
      printPlacement();
#endif

      /* to facilitate A*, sort sinks in increasing order of separation from source */
      netlist.sort_netSinks();

      /* Initial Pass...route each net in the netlist and calculate sharing and wire costs */
      routeNets(1,true);

      /*     exit(1); */

      /* calculate starting temperature */
      calcStartTemperature();

      /*      exit(1); */

      /* enter simulated annealing placement routine */
      annealDesign(fp);

      /* stop measuring the anneal runtime */
      end_time=time(0);

/*       cout<<endl<<endl<<"TOTAL ANNEAL RUNTIME IN SECONDS : "<<end_time-start_time<<". NUM ANNEAL ITERATIONS = "<<NUM_ANNEAL_ITERATIONS; */
      fprintf(fp,"\n\nTOTAL ANNEAL RUNTIME IN SECONDS : %d. NUM ANNEAL ITERATIONS = %d",end_time-start_time,NUM_ANNEAL_ITERATIONS);

/*       cout<<endl<<endl<<"ANNEAL RUNTIME PER ITERATION : "<<((float)(end_time-start_time)/(float)(NUM_ANNEAL_ITERATIONS)); */
      fprintf(fp,"\n\nANNEAL RUNTIME PER ITERATION : %.2f",((float)(end_time-start_time)/(float)(NUM_ANNEAL_ITERATIONS)));

      printVPRPlacement(randomSeed,echo_file_name,fp);

      fclose(fp);

      /* rip up all nets */
/*       cout<<endl<<endl<<"RIPPING UP THE ENTIRE ROUTING...."; */
      netlist.create_netListItr();
      for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){

	/*       cout<<endl<<"LOOKING AT "<<net->name; */

	if(net->sinkVector!=0)
	  net->rip_rTree();
      }
      netlist.remove_netListItr();
/*       cout<<"DONE!! WIRE USAGE = "<<wireCost<<" AND SHARING = "<<sharingCost; */

    }
    else{
      /* parse placement file */

      if(place_file_name==0)
	ERROR("need to input a placement file using the -place_file command-line option..."); 

      yyin=fopen(place_file_name,"r");

      if(yyin==0)
	ERROR("could not find placement file...");

      while(!feof(yyin)){
	yyparse();
      }
      fclose(yyin);

      sprintf(outfile_name,"%s_%s.route",netlist.get_name(),echo_file_name);
      fp=fopen(outfile_name,"w");

      /* start measuring the routing runtime */
      start_time=time(0);
      netlist.sort_netSinks();

      /* do a final routing phase consisting of 30 iterations */
      FPGA->updateHistoryCosts(1,false);
      routeNets(1,true);
      sharingCost=FPGA->countSharing(0);wireCost=calcWireCost();

      cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
      cout<<endl<<"ITERATION 1\tSHARING = "<<sharingCost;

      fprintf(fp,"\n\n<<<<<<<<<<<<<<<<<<<<<< BEGIN ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\nITERATION 1\tSHARING = %d",sharingCost);

      for(itr=2;itr<=30;itr++){
	FPGA->updateHistoryCosts(itr,false);
	routeNets(itr,true);
	cout<<endl<<"ITERATION "<<itr<<"\tSHARING = "<<sharingCost;

	fprintf(fp,"\nITERATION %d\tSHARING = %d",itr,sharingCost);

	if(sharingCost==0){
	  cout<<endl<<endl<<"SUCCESFULLY PLACED AND ROUTED THIS BLOKE!!";
	  cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
	  cout<<endl<<endl;

	  fprintf(fp,"\n\nSUCCESFULLY PLACED AND ROUTED THIS BLOKE!!\n\n<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\n\n");

	  break;
	}
      }

      if(itr==31){
	cout<<endl<<endl<<"CLOSE, BUT NO CIGAR....";
	cout<<endl<<endl<<"<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>";
	cout<<endl<<endl;

	fprintf(fp,"\n\nCLOSE, BUT NO CIGAR....\n\n<<<<<<<<<<<<<<<<<<<<<< END ROUTING PHASE >>>>>>>>>>>>>>>>>>>>>>>>\n\n");

      }

      /* end measuring the routing runtime */
      end_time=time(0);

      cout<<endl<<endl<<"TOTAL ROUTING RUNTIME IN SECONDS : "<<end_time-start_time;
      cout<<endl<<endl;
      cout<<endl;

      fprintf(fp,"\n\nTOTAL ROUTING RUNTIME IN SECONDS : %d\n\n\n",end_time-start_time);

      fclose(fp);

    }

    return 0;
  }
  catch(out_of_range){
    cerr<<"vector overflow error...\n";
  }

}

void dumpAstarTable(){

  cout<<endl<<endl<<"<<<<<<<<<<<< BEGIN DUMP A* TABLE >>>>>>>>>>>>";

  for(int i=0;i<AstarTable->size();i++){
    cout<<endl<<endl<<"CLUSTER NUM: "<<i<<endl;
    for(int j=0;j<AstarTable->at(i)->size();j++)
      cout<<AstarTable->at(i)->at(j)<<"  ";
  }

  cout<<endl<<endl<<"<<<<<<<<<<<<< END DUMP A* TABLE >>>>>>>>>>>>>";

}

void process_AstarTableEntry(int cluster_id,char* sink_name,double estimate){

  if(sink_name==0 || cluster_id<0 || cluster_id>=numClusters)
    ERROR("error while processing Astar table entry...");

  routingNode* rnode=find_rnode(sink_name);
  if(rnode==0){
    cout<<endl<<sink_name<<endl;
    ERROR("Could not find rnode in list of routing nodes...\n");
  }

  AstarTable->at(cluster_id)->at(rnode->sinkID)=estimate;
}

void prepareAstarTable(int num_clusters,int num_sinks){
  string rname;

  if(num_clusters<1 || num_sinks<1)
    ERROR("dubious Astar information here...");

  /* allocate table of Astar estimates */
  AstarTable=new vector<vector<double>*> (num_clusters);
  for(int i=0;i<AstarTable->size();i++){
    AstarTable->at(i)=new vector<double> (num_sinks);
    for(int j=0;j<AstarTable->at(i)->size();j++)
      AstarTable->at(i)->at(j)=-1.0;
  }

}

void calcStartTemperature(){
  void* block0;void* block1;
  vector<float> costArray(numBlocks);
  float prev_wireCost=(float)(wireCost);
  int index0,index1,block_type;

  for(int i=0;i<numBlocks;i++){

    /* randomly generate a pair of FPGA tiles that will be involved in the move */
    generateMove(&block0,&index0,&block1,&index1,&block_type);

    /* make the move i.e. actually move the logic/IO blocks from one location to the other */
    makeMove(&block0,&index0,&block1,&index1,&block_type,1);

#if defined(PLACER_DEBUG)
    actual_wireCost=calcWireCost();
    actual_sharingCost=FPGA->countSharing(0);
    if(wireCost!=actual_wireCost)
      ERROR("wire calculation mismatch...");
    if(sharingCost!=actual_sharingCost)
      ERROR("sharing calculation mismatch...");
#endif

    /* sort sinks for A* purposes */
    sortSinks(&block0,&index0,&block1,&index1,&block_type);

    costArray.at(i)=SHARE_FAC*(float)(sharingCost) + (float)(wireCost);

  }

  float avg_cost=0.;float cov=0.;
  /* calculate the average cost */
  for(int j=0;j<numBlocks;j++)
    avg_cost+=costArray.at(j)/(float)(numBlocks);
  /* calculate cov */
  for(int k=0;k<numBlocks;k++)
    cov+=(costArray.at(k) - avg_cost)*(costArray.at(k) - avg_cost);

  annealTemp = (float)(0.001 * pow((double)cov/(double)(numBlocks),0.5));

}

bool thereIsANeedToAnneal(){

  if(annealTemp<1.0e-5)
    return false;
  else
    return true;

}

void annealDesign(FILE* fp){
  int bf_sharingCost,bf_wireCost,af_sharingCost,af_wireCost,prev_wireCost;
  int iter=1;float deltaC;double i;int iteration;
  void* block0;void* block1;
  int index0,index1;int block_type;
  list<cachedObj*> sink_list0;list<cachedObj*> sink_list1;
  list<cachedObj*> src_list0;list<cachedObj*> src_list1;
  double num_acc=0.;double fracn_acc=1.;double alpha=0.;
  float placementCost=0.;double commandline_alpha;
  time_t startTime,stopTime,time_diff;
  double non_zero,rand_num,num_bad_acc,num_good_acc,bad_acc,good_acc;
  Net* net;

/*   cout<<endl<<endl; */
/*   cout<<endl<<"temp.\t\tw_use\t\tshare\t\taccept\t\tdlim\t    cost\tALPHA\tgood_acc\tbad_acc"; */
/*   cout<<endl<<"========================================================================================================================="; */

  fprintf(fp,"\n\n\ntemp.\t\tw_use\t\tshare\t\taccept\t\tdlim\t    cost\tALPHA\tgood_acc\tbad_acc");
  fprintf(fp,"\n=========================================================================================================================");

  while(fracn_acc > 0.03 && iter<101){
/*   for(int u=0;u<25;u++){ */
    iter++;

    FPGA->updateHistoryCosts(iter,true);

    prev_wireCost=wireCost;

    num_acc=0.;non_zero=0.0;num_good_acc=0.0;num_bad_acc=0.0;

    startTime=time(0);
    for(i=0.;i<INNER_NUM*pow(numBlocks,1.33);i+=1.0){

#if defined(PLACER_DEBUG)
      cout<<endl<<endl<<"********************** GENERATING NEW MOVE ************************";
#endif


      generateMove(&block0,&index0,&block1,&index1,&block_type);


      bf_sharingCost=sharingCost;bf_wireCost=wireCost;


      cacheRoutes(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1);

      sortSinks(&block0,&index0,&block1,&index1,&block_type);
      makeMove(&block0,&index0,&block1,&index1,&block_type,iter);


      af_sharingCost=sharingCost;af_wireCost=wireCost;

      deltaC=SHARE_FAC*(float)(af_sharingCost-bf_sharingCost)/(float)(prev_wireCost) + 
             (float)(af_wireCost-bf_wireCost)/(float)(prev_wireCost);
/*       deltaC = (float)(af_wireCost-bf_wireCost); */

      if(deltaC > 1.0e-10){
/*       if(deltaC >= 1.0){ */
	// this is a bad move
	non_zero+=1.0;

	rand_num=drand48();
/* 	cout<<rand_num<<",  "; */

        if(rand_num < exp(-0.01*deltaC/annealTemp)){
          // accept this move
/* 	  sortSinks(&block0,&index0,&block1,&index1,&block_type); */
          clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);
          num_acc+=1.0;num_bad_acc+=1.0;

#if defined(PLACER_DEBUG)
          actual_wireCost=calcWireCost();
          actual_sharingCost=FPGA->countSharing(0);
          if(wireCost!=actual_wireCost)
            ERROR("wire calculation mismatch...");
          if(sharingCost!=actual_sharingCost)
            ERROR("sharing calculation mismatch...");
#endif

        }
        else{
          // reject this move
          undoMove(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1,iter);
          clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,0);

#if defined(PLACER_DEBUG)
          actual_wireCost=calcWireCost();
          actual_sharingCost=FPGA->countSharing(0);
          if(wireCost!=actual_wireCost)
            ERROR("wire calculation mismatch...");
          if(sharingCost!=actual_sharingCost)
            ERROR("sharing calculation mismatch...");
#endif

        }

      }
      else{
	// this is a good move...keep it
/* 	sortSinks(&block0,&index0,&block1,&index1,&block_type); */
	clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);

	if(deltaC < -1.0e-10){
	  num_acc+=1.0;non_zero+=1.0;num_good_acc+=1.0;
	}
/*  	num_acc+=1.0; */

#if defined(PLACER_DEBUG)
        actual_wireCost=calcWireCost();
        actual_sharingCost=FPGA->countSharing(0);
        if(wireCost!=actual_wireCost)
          ERROR("wire calculation mismatch...");
        if(sharingCost!=actual_sharingCost)
          ERROR("sharing calculation mismatch...");
#endif

      }

    }

    stopTime=time(0);
    time_diff = stopTime - startTime;

/*     fracn_acc=num_acc/i; */
    fracn_acc=num_acc/non_zero;bad_acc=num_bad_acc/non_zero;good_acc=num_good_acc/non_zero;
    placementCost=SHARE_FAC*(float)(sharingCost) + (float)(wireCost);
    printf("\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t%10.2f\t%4.2f\t%5.4f\t\t%5.4f\t\t%d",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,ALPHA,good_acc,bad_acc,time_diff);
    fprintf(fp,"\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t%10.2f\t%4.2f\t%5.4f\t\t%5.4f\t\t%d",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,ALPHA,good_acc,bad_acc,time_diff);

    if(fracn_acc>0.96)
      alpha=0.5;
    else if(fracn_acc<=0.96 && fracn_acc>0.8)
      alpha=0.9;
    else if(fracn_acc<=0.8 && fracn_acc>0.15)
      alpha=0.95;
    else
      alpha=0.8;

/*     alpha=0.9; */

    annealTemp=alpha*annealTemp;

/*     dLimit = dLimit*(1-0.44+fracn_acc); */
/*     if(dLimit > dLimitMax) */
/*       dLimit = dLimitMax; */
/*     if(dLimit < dLimitMin) */
/*       dLimit=dLimitMin; */

    dLimit = dLimit*(1-0.44+fracn_acc);
    if(dLimit > (float)(FPGA->get_numRows()))
      dLimit = (float)(FPGA->get_numRows());
    if(dLimit < 1)
      dLimit=1;

    if(fracn_acc>0.10){
      netlist.sort_netSinks();
      routeNets(iter,true);
    }
/*     cout<<endl<<"WIRE_COST = "<<wireCost<<", SHARE_COST = "<<sharingCost; */

  }

  /* quench now */
  for(int n=0;n<3;n++){

    /* increment iteration */
    iter++;

    /* init normalizing factor */
    prev_wireCost=wireCost;
    /* init num moves accepted to 0 */
    num_acc=0.;

    startTime = time(0);
    for(i=0.;i<INNER_NUM*pow(numBlocks,1.33);i+=1.0){

#if defined(PLACER_DEBUG)
      cout<<endl<<endl<<"********************** GENERATING NEW MOVE ************************";
#endif

      /* randomly generate a pair of FPGA tiles that will be involved in the move */
      generateMove(&block0,&index0,&block1,&index1,&block_type);

      /* cache sharing and wire costs before making the move */
      bf_sharingCost=sharingCost;bf_wireCost=wireCost;

      /* cache route trees & terminal mapping of the nets that connect to the moved blocks */
      cacheRoutes(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1);

      sortSinks(&block0,&index0,&block1,&index1,&block_type);
      /* make the move i.e. actually move the logic/IO blocks from one location to the other */
      makeMove(&block0,&index0,&block1,&index1,&block_type,iter);

      /* cache sharing and wire costs after making the move */
      af_sharingCost=sharingCost;af_wireCost=wireCost;

      deltaC=SHARE_FAC*(float)(af_sharingCost-bf_sharingCost)/(float)(prev_wireCost) + 
	(float)(af_wireCost-bf_wireCost)/(float)(prev_wireCost);
/*       deltaC = (float)(af_wireCost-bf_wireCost); */

      if(deltaC > 1.0e-10){
	/* this is a bad move....reject */
	undoMove(&block0,&index0,&block1,&index1,&block_type,&src_list0,&src_list1,&sink_list0,&sink_list1,iter);
        clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,0);

#if defined(PLACER_DEBUG)
        actual_wireCost=calcWireCost();
        actual_sharingCost=FPGA->countSharing(0);
        if(wireCost!=actual_wireCost)
          ERROR("wire calculation mismatch...");
        if(sharingCost!=actual_sharingCost)
          ERROR("sharing calculation mismatch...");
#endif

      }
      else{
	/* this is a good move...keep it */
/* 	sortSinks(&block0,&index0,&block1,&index1,&block_type); */
	clearCachedRoutes(&src_list0,&src_list1,&sink_list0,&sink_list1,1);

	if(deltaC < -1.0e-10)
	  num_acc+=1.0;
/*  	num_acc+=1.0; */

#if defined(PLACER_DEBUG)
        actual_wireCost=calcWireCost();
        actual_sharingCost=FPGA->countSharing(0);
        if(wireCost!=actual_wireCost)
          ERROR("wire calculation mismatch...");
        if(sharingCost!=actual_sharingCost)
          ERROR("sharing calculation mismatch...");
#endif

      }

    }

    fracn_acc=num_acc/i;
    placementCost=SHARE_FAC*(float)(sharingCost) + (float)(wireCost);

    stopTime=time(0);
    time_diff = stopTime - startTime;
    printf("\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t\t%10.2f\t\t%d",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,time_diff);
    fprintf(fp,"\n%9.7f\t %ld\t\t  %ld\t\t%5.4f\t\t%4.1f\t\t%10.2f\t\t%d",annealTemp,wireCost,sharingCost,fracn_acc,dLimit,placementCost,time_diff);

  }

  NUM_ANNEAL_ITERATIONS = iter;

}


void undoMove(void** block0,int* index0,void** block1,int* index1,int* block_type,list<cachedObj*>* src_list0,
              list<cachedObj*>* src_list1,list<cachedObj*>* sink_list0,
              list<cachedObj*>* sink_list1,int iter){
  logicBlock* lb0;logicBlock* lb1;
  Terminal* term;
  int connected=0;

  if(*block_type==0){
    // We have an IOB
    lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
    lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
  }
  else{
    // We have a CLB
    lb0=((vprCLB*)(*block0))->get_currBlock();
    lb1=((vprCLB*)(*block1))->get_currBlock();
  }

  /* alright, rip up the source nets and sink segments that connect to lb0 and/or lb1 */
  ripRoutes(lb0,lb1);

  /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
  if(*block_type==1){
    ((vprCLB*)(*block0))->get_sourceNode()->clear_terminalList();
    ((vprCLB*)(*block0))->get_sinkNode()->clear_terminalList();
    ((vprCLB*)(*block1))->get_sourceNode()->clear_terminalList();
    ((vprCLB*)(*block1))->get_sinkNode()->clear_terminalList();
  }
  else{
    ((vprIOB*)(*block0))->getfrom_sourceArray(*index0)->clear_terminalList();
    ((vprIOB*)(*block0))->getfrom_sinkArray(*index0)->clear_terminalList();
    ((vprIOB*)(*block1))->getfrom_sourceArray(*index1)->clear_terminalList();
    ((vprIOB*)(*block1))->getfrom_sinkArray(*index1)->clear_terminalList();
  }

  /* translate/swap lb0 and/or lb1 */
  if(*block_type==0){
    placeBlock(lb1,*block0,0,*index0);
    if(lb0)
      placeBlock(lb0,*block1,0,*index1);
    else
      ((vprIOB*)(*block1))->addto_currBlockArray(0,*index1);
  }
  else{
    placeBlock(lb1,*block0,1,*index0);
    if(lb0)
      placeBlock(lb0,*block1,1,*index1);
    else
      ((vprCLB*)(*block1))->set_currBlock(0);
  }

  if(*block_type==0){
    // We have an IOB
    lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
    lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
  }
  else{
    // We have a CLB
    lb0=((vprCLB*)(*block0))->get_currBlock();
    lb1=((vprCLB*)(*block1))->get_currBlock();
  }

  /* now reconnect the previously cached nets and net segments to the terminals of lb0 */
  /* and/or lb1. Route trees that were cached in cacheRoutes() will be used here. */

  if(lb0){
    /* loop thru the output terminals of lb0. for each output terminal, reconnect the */
    /* cached net back to lb0. */
    list<cachedObj*>::iterator list_itr=src_list0->begin();
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* repopulate the net's route tree with previously cached route tree. */
	(term->get_net()->get_rTree())->set_root(((*list_itr)->rTree).get_root());
	/* increment occupancies */
	(term->get_net()->get_rTree())->incrementOccupancies(0);
	term->get_net()->set_isCounted(true);
	/* increment the source list iterator */
	list_itr++;
 
      }
    }
    lb0->remove_outputListItr();
  }

  if(lb1){
    /* loop thru the output terminals of lb1. for each output terminal, reconnect the */
    /* cached net back to lb1. */
    list<cachedObj*>::iterator list_itr=src_list1->begin();
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* repopulate the net's route tree with previously cached route tree. */
	(term->get_net()->get_rTree())->set_root(((*list_itr)->rTree).get_root());
	/* increment occupancies. */
	(term->get_net()->get_rTree())->incrementOccupancies(0);
	term->get_net()->set_isCounted(true);
	/* increment the source list iterator */
	list_itr++;
 
      }
    }
    lb1->remove_outputListItr();
  }

  if(lb0){
    /* now, loop thru the input terminals of lb0. for each input terminal, reconnect */
    /* the previously cached net segment back to the sink terminal of the logic block */
    list<cachedObj*>::iterator list_itr=sink_list0->begin();
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* now reconnect the previously cached net segment to the sink terminal. */
	/* while doing so, increment occupancies. */
	connected=term->get_net()->reconnect(((*list_itr)->rTree).get_root());

        /* if unable to connect, store the terminal and branch. the connection will be made later. */
        if(connected==-1){
          routeSink(iter,term->get_net(),term->get_rNode(),1);
          (*list_itr)->clearTree=1;
        }

	/* increment the source list iterator */
	list_itr++;

      }
    }
    lb0->remove_inputListItr();
  }

  if(lb1){
    /* now, loop thru the input terminals of lb1. for each input terminal, reconnect */
    /* the previously cached net segment back to the sink terminal of the logic block */
    list<cachedObj*>::iterator list_itr=sink_list1->begin();
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	if((*list_itr)->rNode!=term->get_rNode())
	  ERROR("terminal mismatch while undoing move...");
	/* now reconnect the previously cached net segment to the sink terminal. */
	/* while doing so, increment occupancies. */
	connected=term->get_net()->reconnect(((*list_itr)->rTree).get_root());

        /* if unable to connect, store the terminal and branch. the connection will be made later. */
        if(connected==-1){
          routeSink(iter,term->get_net(),term->get_rNode(),1);
          (*list_itr)->clearTree=1;
        }

	/* increment the source list iterator */
	list_itr++;

      }
    }
    lb1->remove_inputListItr();
  }

  /* reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

}

void clearCachedRoutes(list<cachedObj*>* src_list0,list<cachedObj*>* src_list1,
                list<cachedObj*>* sink_list0,list<cachedObj*>* sink_list1,int clear_tree){
  cachedObj* ptr;

  list<cachedObj*>::iterator list_itr=src_list0->begin();
  for(;list_itr!=src_list0->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  src_list0->clear();

  list_itr=src_list1->begin();
  for(;list_itr!=src_list1->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  src_list1->clear();

  list_itr=sink_list0->begin();
  for(;list_itr!=sink_list0->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  sink_list0->clear();

  list_itr=sink_list1->begin();
  for(;list_itr!=sink_list1->end();list_itr++){
    if(clear_tree==1 || (clear_tree==0 && (*list_itr)->clearTree==1))
      ((*list_itr)->rTree).clear();
    ptr=*list_itr;
    delete ptr;
  }
  sink_list1->clear();

}

void cacheRoutes(void** block0,int* index0,void** block1,int* index1,int* block_type,list<cachedObj*>* src_list0,
                 list<cachedObj*>* src_list1,list<cachedObj*>* sink_list0,
                 list<cachedObj*>* sink_list1){
  logicBlock* lb0;logicBlock* lb1;
  Terminal* term;
  cachedObj* cache_obj;

  if(*block_type==0){
    // We have an IOB
    lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
    lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
  }
  else{
    // We have a CLB
    lb0=((vprCLB*)(*block0))->get_currBlock();
    lb1=((vprCLB*)(*block1))->get_currBlock();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN CACHE ROUTES...";
#endif

  if(lb0){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb0->get_name();
#endif

    /* loop thru the output terminals of lb0. for each output terminal, store the routing */
    /* node that the terminal maps to and the ENTIRE route tree of the net that it drives */
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){ 
      /* create new cache object */
      cache_obj=new cachedObj;
      /* store routing node that the terminal maps to */
      cache_obj->rNode=term->get_rNode();
      /* copy ENTIRE route tree */
      (cache_obj->rTree).populateWith(term->get_net()->get_rTree()->get_root());
      /* add to list of cached objects */
      src_list0->push_back(cache_obj);

#if defined(PLACER_DEBUG)
      (cache_obj->rTree).print_info();
#endif

      term->get_net()->set_isCounted(true);
    }
    lb0->remove_outputListItr();
  }

  if(lb1){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb1->get_name();
#endif

    /* loop thru the output terminals of lb1. for each output terminal, store the routing */
    /* node that the terminal maps to and the ENTIRE route tree of the net that it drives */
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      /* create new cache object */
      cache_obj=new cachedObj;
      /* store routing node that the terminal maps to */
      cache_obj->rNode=term->get_rNode();
      /* copy ENTIRE route tree */
      (cache_obj->rTree).populateWith((term->get_net()->get_rTree())->get_root());
      /* add to list of cached objects */
      src_list1->push_back(cache_obj);

#if defined(PLACER_DEBUG)
      (cache_obj->rTree).print_info();
#endif

      term->get_net()->set_isCounted(true);
    }
    lb1->remove_outputListItr();
  }

  if(lb0){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb0->get_name();
#endif

    /* now, loop thru the input terminals of lb0. for each input terminal, store the */
    /* routing node that it maps to and the segment of the net's route tree that sinks */
    /* at this input terminal */
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	/* create new cache object */
	cache_obj=new cachedObj;
	/* store routing node that the terminal maps to */
	cache_obj->rNode=term->get_rNode();
	/* get the sub-tree that sinks at this terminal and store a copy */
	(cache_obj->rTree).populateWith(
         term->get_net()->get_rTree()->get_subTree(term->get_rNode()),term->get_rNode());

#if defined(PLACER_DEBUG)
	(cache_obj->rTree).print_info();
#endif

	/* add to list of cached objects */
	sink_list0->push_back(cache_obj);

      }
    }
    lb0->remove_inputListItr();
  }

  if(lb1){

#if defined(PLACER_DEBUG)
    cout<<endl<<"LOOKING AT "<<lb1->get_name();
#endif

    /* now, loop thru the input terminals of lb1. for each input terminal, store the */
    /* routing node that it maps to and the segment of the net's route tree that sinks */
    /* at this input terminal */
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false){
        /* only consider those nets that have not already been ripped up */

	/* create new cache object */
	cache_obj=new cachedObj;
	/* store routing node that the terminal maps to */
	cache_obj->rNode=term->get_rNode();
	/* get the sub-tree that sinks at this terminal and store a copy */
	(cache_obj->rTree).populateWith(
         term->get_net()->get_rTree()->get_subTree(term->get_rNode()),term->get_rNode());

#if defined(PLACER_DEBUG)
	(cache_obj->rTree).print_info();
#endif

	/* add to list of cached objects */
	sink_list1->push_back(cache_obj);

      }
    }
    lb1->remove_inputListItr();
  }

  /* reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<"EXITING CACHE ROUTES...";
#endif

}

void sortSinks(void** block0,int* index0,void** block1,int* index1,int* block_type){
  logicBlock* lb0;logicBlock* lb1;
  Terminal* term;

  if(*block_type==0){
    // We have an IOB
    lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
    lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
  }
  else{
    // We have a CLB
    lb0=((vprCLB*)(*block0))->get_currBlock();
    lb1=((vprCLB*)(*block1))->get_currBlock();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN SORT SINKS...";
#endif

  /* now, sort the nets that connect to lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->sort_sinkVector();
    lb0->remove_outputListItr();

    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList())
      term->get_net()->sort_sinkVector();
    lb0->remove_inputListItr();
  }

  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->sort_sinkVector();
    lb1->remove_outputListItr();

    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList())
      term->get_net()->sort_sinkVector();
    lb1->remove_inputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<"EXITING SORT SINKS...";
#endif

}


void makeMove(void** block0,int* index0,void** block1,int* index1,int* block_type,int iter){
  Terminal* term;
  logicBlock* lb0;logicBlock* lb1;

  if(*block_type==0){
    // We have an IOB
    lb0=((vprIOB*)(*block0))->getfrom_currBlockArray(*index0);
    lb1=((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
  }
  else{
    // We have a CLB
    lb0=((vprCLB*)(*block0))->get_currBlock();
    lb1=((vprCLB*)(*block1))->get_currBlock();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN MAKE MOVE...";
#endif

  /* first, rip up the entire net that is sourced at a logic block. also, rip up the */
  /* branch of each net that has a sink at the logic block. the occupancy of the nodes */  
  /* in the ripped-up net segments are accordingly decremented. */
  ripRoutes(lb0,lb1);

  /* clear the terminal lists of the SOURCE and SINK routing nodes in block0 and block1 */
  if(*block_type==1){
    ((vprCLB*)(*block0))->get_sourceNode()->clear_terminalList();
    ((vprCLB*)(*block0))->get_sinkNode()->clear_terminalList();
    ((vprCLB*)(*block1))->get_sourceNode()->clear_terminalList();
    ((vprCLB*)(*block1))->get_sinkNode()->clear_terminalList();
  }
  else{
    ((vprIOB*)(*block0))->getfrom_sourceArray(*index0)->clear_terminalList();
    ((vprIOB*)(*block0))->getfrom_sinkArray(*index0)->clear_terminalList();
    ((vprIOB*)(*block1))->getfrom_sourceArray(*index1)->clear_terminalList();
    ((vprIOB*)(*block1))->getfrom_sinkArray(*index1)->clear_terminalList();
  }

  /* translate/swap lb0 and/or lb1 */
  if(*block_type==0){
    placeBlock(lb0,*block1,0,*index1);
    if(lb1)
      placeBlock(lb1,*block0,0,*index0);
    else
      ((vprIOB*)(*block0))->addto_currBlockArray(0,*index0);
  }
  else{
    placeBlock(lb0,*block1,1,*index1);
    if(lb1)
      placeBlock(lb1,*block0,1,*index0);
    else
      ((vprCLB*)(*block0))->set_currBlock(0);
  }

  /* now, re-route the nets that connect to lb0 and lb1 */
  if(lb0){
    /* route the output nets of lb0 */
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      routeNet(iter,term->get_net(),1,false);
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb0->remove_outputListItr();
  }

  /* now, re-route the nets that connect to lb0 and lb1 */
  if(lb1){
    /* route the output nets of lb0 */
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      routeNet(iter,term->get_net(),1,false);
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb1->remove_outputListItr();
  }

  if(lb0){
    /* now, route the input terminals of lb0 */ 
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        routeSink(iter,term->get_net(),term->get_rNode(),1);
    }
    lb0->remove_inputListItr();
  }

  if(lb1){
    /* now, route the input terminals of lb0 */ 
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        routeSink(iter,term->get_net(),term->get_rNode(),1);
    }
    lb1->remove_inputListItr();
  }

  /* now that the nets have been ripped up, reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<"EXITING MAKE MOVE...";
#endif

}


/***
* This function rips up the entire source net, and the branch of each net that sinks at
* lb0 and/or lb1. Occupancies are appropriately decremented.
***/  
void ripRoutes(logicBlock* lb0,logicBlock* lb1){
  Terminal* term;

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"IN RIP ROUTES...";
#endif

  if(lb0){
    /* loop thru the output nets of lb0 and delete route trees */
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      term->get_net()->rip_rTree();
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb0->remove_outputListItr();
  }

  if(lb1){
    /* loop thru the output nets of lb1 and delete route trees */
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList()){
      /* only consider those nets that have not already been ripped up */
      term->get_net()->rip_rTree();
      /* this net has now been counted */
      term->get_net()->set_isCounted(true);
    }
    lb1->remove_outputListItr();
  }

  if(lb0){
    /* now, loop thru the input nets of lb0 and rip up the branch that connects */
    /* lb0 to the route tree of each input net */
    lb0->create_inputListItr();
    for(term=lb0->getnext_inputList();term!=0;term=lb0->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        term->get_net()->rip_rTreeBranch(term->get_rNode());
    }
    lb0->remove_inputListItr();
  }

  if(lb1){
    /* now, loop thru the input nets of lb1 and rip up the branch that connects */
    /* lb1 to the route tree of each input net */
    lb1->create_inputListItr();
    for(term=lb1->getnext_inputList();term!=0;term=lb1->getnext_inputList()){
      if(term->get_net()->get_isCounted()==false)
        term->get_net()->rip_rTreeBranch(term->get_rNode());
    }
    lb1->remove_inputListItr();
  }

  /* now that the nets have been ripped up, reset the output nets of lb0 and lb1 */
  if(lb0){
    lb0->create_outputListItr();
    for(term=lb0->getnext_outputList();term!=0;term=lb0->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb0->remove_outputListItr();
  }
  if(lb1){
    lb1->create_outputListItr();
    for(term=lb1->getnext_outputList();term!=0;term=lb1->getnext_outputList())
      term->get_net()->set_isCounted(false);
    lb1->remove_outputListItr();
  }

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"EXITING RIP ROUTES...";
#endif

}


/***
* This function checks to see if lb is a movable logic block. The
* reason it is here is because there are 'global' nets that are 
* sourced at input blocks that have no sinks (e.g. clock, reset etc)
***/
bool movable_block(logicBlock* lb){
  Net* net;

  /* if we have an input block that sources a net with no sinks, */
  /* then the block is not movable. */
  if(lb->get_isInput()){
    lb->create_outputListItr();
    net=lb->getnext_outputList()->get_net();
    lb->remove_outputListItr();
    if(net->sinkVector==0)
      return false;
  }
  return true;

}


/***
* This function generates an anneal move. When the function is done executing
* block0 and block1 point to the 2 FPGA locations that will be involved in the
* move. block0 always points to a location that holds a logic or IO block, while 
* block1 may or may not be currently occupied. If the blocks involved are IOBs,
* then index0 and index1 hold the indices of the specific IOs involved. block_type
* indicates whether the involved blocks are CLBs or IOBs. 
***/
void generateMove(void** block0,int* index0,void** block1,int* index1,int* block_type){
  int num_rows=FPGA->get_numRows();
  int num_cols=FPGA->get_numCols();
  int num_els=netlist.get_numlogicBlocks()+netlist.get_numInputs()+netlist.get_numOutputs();
  int col0,row0,col1,row1;
  double selector;
  logicBlock* lb1;

  int dlim = (int)(floor((double)dLimit));

  if(dlim>FPGA->get_numRows() || dlim>FPGA->get_numCols())
    dlim=FPGA->get_numRows();
  else if(dlim<1)
    dlim=1;

  logicBlock* lb0=netlist.get_logicBlock(lrand48()%num_els);
  while(movable_block(lb0)==false)
  /* randomly select a logic or IO block */
    lb0=netlist.get_logicBlock(lrand48()%num_els);

  /* record its current coordinates in the vpr FPGA */
  col0=lb0->col;row0=lb0->row;*index0=lb0->io_index;

#if defined(PLACER_DEBUG)
  cout<<endl<<endl<<"LOGIC BLOCK "<<lb0->get_name()<<" AT COL="<<col0<<" AND ROW="<<row0;
#endif

  /* assign the tile at these coordinates to block0 */
  *block0=FPGA->get_blockAt(row0,col0);

  if(lb0->get_isLogic()!=true){
    *block_type=0;

    /* lb0 is an IO block...randomly select another IO location */
    while(1){
      selector=drand48();
      /* randomly select an IO location at the periphery of the FPGA */
      if(selector<=0.25){
        col1=0;
        row1=1 + lrand48()%num_rows;
      }
      else if(selector>0.25 && selector<=0.5){
        col1=num_cols+1;
        row1=1 + lrand48()%num_rows;
      }
      else if(selector>0.5 && selector<=0.75){
        row1=0;
        col1=1 + lrand48()%num_cols;
      }
      else{
        row1=num_rows+1;
        col1=1 + lrand48()%num_cols;
      }
      /* check to see whether randomly generated IO tile is the same as the other */ 
      if((row0==row1 && col0==col1) || (abs(row0-row1)>dlim || abs(col0-col1)>dlim))
        continue;
      else{
	*block1=FPGA->get_blockAt(row1,col1);
	*index1=lrand48()%( ((vprIOB*)(*block1))->get_numIOs() );
	lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(*index1);
	if(lb1){
	  if(movable_block(lb1)==false)
	    continue;
	}
        break;
      }
    }

#if defined(PLACER_DEBUG)
    if(lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(*index1) ){
      cout<<endl<<endl<<"LOGIC BLOCK "<<lb1->get_name()<<" AT COL="<<lb1->col<<" AND ROW="<<
        lb1->row<<" AND INDEX_NUM="<<*index1;
    }
#endif

  }

  else{
    *block_type=1;

    /* lb0 is a logic block...randomly select another logic block location */
    while(1){
      /* randomly select a logic block location in the FPGA */
      col1=1 + lrand48()%num_cols;
      row1=1 + lrand48()%num_rows;
      *index1=-1;
      /* check to see whether randomly generated location is the same as the other */ 
      if((row0==row1 && col0==col1) || abs(row0-row1)>dlim || abs(col0-col1)>dlim)
        continue;
      else
        break;

    }
    *block1=FPGA->get_blockAt(row1,col1);

#if defined(PLACER_DEBUG)
    if(lb1=((vprCLB*)(*block1))->get_currBlock()){
      cout<<endl<<endl<<"LOGIC BLOCK "<<lb1->get_name()<<" AT COL="<<lb1->col<<" AND ROW="<<
        lb1->row;
    }
#endif

  }

}


bool withinRange(void* block0,void* block1,int index0,int index1,int block_type){
  routingNode* src0;routingNode* sink1;

  if(block_type==0){
    /* this is an IOB */
    vprIOB* iob0 = (vprIOB *)block0;vprIOB* iob1 = (vprIOB *)block1;
    src0 = iob0->sourceArray->at(index0);sink1 = iob1->sinkArray->at(index1);
    if((src0->allSinksTable).at(sink1->sinkID) < 0.0)
      return false;
    else{
      if((src0->allSinksTable).at(sink1->sinkID) > dLimit)
	return false;
      else
	return true;
    }
  }
  else{
    /* this is a logic block */
    vprCLB* clb0 = (vprCLB *)block0;vprCLB* clb1 = (vprCLB *)block1;
    src0 = clb0->sourceNode;sink1 = clb1->sinkNode;
    if((src0->allSinksTable).at(sink1->sinkID) < 0.0)
      return false;
    else{
      if((src0->allSinksTable).at(sink1->sinkID) > dLimit)
	return false;
      else
	return true;
    }
  }

}


/* void generateMove(void** block0,int* index0,void** block1,int* index1,int* block_type){ */
/*   int num_rows=FPGA->get_numRows(); */
/*   int num_cols=FPGA->get_numCols(); */
/*   int num_els=netlist.get_numlogicBlocks()+netlist.get_numInputs()+netlist.get_numOutputs(); */
/*   int col0,row0,col1,row1; */
/*   double selector; */
/*   logicBlock* lb1; */

/*   logicBlock* lb0=netlist.get_logicBlock(lrand48()%num_els); */
/*   while(movable_block(lb0)==false) */
/*     lb0=netlist.get_logicBlock(lrand48()%num_els); */

/*   col0=lb0->col;row0=lb0->row;*index0=lb0->io_index; */

/* #if defined(PLACER_DEBUG) */
/*   cout<<endl<<endl<<"LOGIC BLOCK "<<lb0->get_name()<<" AT COL="<<col0<<" AND ROW="<<row0; */
/* #endif */

/*   *block0=FPGA->get_blockAt(row0,col0); */

/*   if(lb0->get_isLogic()!=true){ */
/*     *block_type=0; */

/*     while(1){ */
/*       selector=drand48(); */

/*       if(selector<=0.25){ */
/*         col1=0; */
/*         row1=1 + lrand48()%num_rows; */
/*       } */
/*       else if(selector>0.25 && selector<=0.5){ */
/*         col1=num_cols+1; */
/*         row1=1 + lrand48()%num_rows; */
/*       } */
/*       else if(selector>0.5 && selector<=0.75){ */
/*         row1=0; */
/*         col1=1 + lrand48()%num_cols; */
/*       } */
/*       else{ */
/*         row1=num_rows+1; */
/*         col1=1 + lrand48()%num_cols; */
/*       } */

/*       *block1=FPGA->get_blockAt(row1,col1); */
/*       *index1=lrand48()%( ((vprIOB*)(*block1))->get_numIOs() ); */

/*       if((row0==row1 && col0==col1) || !withinRange(*block0,*block1,*index0,*index1,0)) */
/*         continue; */
/*       else{ */
/* 	lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(*index1); */
/* 	if(lb1){ */
/* 	  if(movable_block(lb1)==false) */
/* 	    continue; */
/* 	} */
/*         break; */
/*       } */
/*     } */

/* #if defined(PLACER_DEBUG) */
/*     if(lb1 = ((vprIOB*)(*block1))->getfrom_currBlockArray(*index1) ){ */
/*       cout<<endl<<endl<<"LOGIC BLOCK "<<lb1->get_name()<<" AT COL="<<lb1->col<<" AND ROW="<< */
/*         lb1->row<<" AND INDEX_NUM="<<*index1; */
/*     } */
/* #endif */

/*   } */

/*   else{ */
/*     *block_type=1; */

/*     while(1){ */
/*       col1=1 + lrand48()%num_cols; */
/*       row1=1 + lrand48()%num_rows; */
/*       *index1=-1; */

/*       *block1=FPGA->get_blockAt(row1,col1); */

/*       if((row0==row1 && col0==col1) || !withinRange(*block0,*block1,*index0,*index1,1)) */
/*         continue; */
/*       else */
/*         break; */

/*     } */

/* #if defined(PLACER_DEBUG) */
/*     if(lb1=((vprCLB*)(*block1))->get_currBlock()){ */
/*       cout<<endl<<endl<<"LOGIC BLOCK "<<lb1->get_name()<<" AT COL="<<lb1->col<<" AND ROW="<< */
/*         lb1->row; */
/*     } */
/* #endif */

/*   } */

/* } */


/* return the smaller of two floating point numbers */
double flomin(double x,double y){
  double diff=y-x;
  if(diff<-1.0e-15)
    return y;
  else
    return x;
}


void printNetsAndRecalcCosts(){
  int wire_cost=0;int sharing_cost=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){

    if(net->sinkVector==0)
      continue;

/*     net->print_Info(); */
/*     net->print_rTree(); */
    wire_cost+=net->get_rNodeCount();
/*     cout<<endl<<"WIRE COST = "<<wire_cost; */
  }
  netlist.remove_netListItr();

  sharing_cost=FPGA->countSharing(0);

  cout<<endl<<endl<<"^^^^^^^^^^^ FINAL WIRE COST = "<<wire_cost<<" ^^^^^^^^^^^^";
  cout<<endl<<"^^^^^^^^^^^ FINAL SHARING COST = "<<sharing_cost<<" ^^^^^^^^^^^^";
  cout<<endl<<endl;

}


/***
* This function routes each net in the netlist. The parameter 'iter'
* will be used to determine the cost (Pathfinder) of routing nodes. 
***/
void routeNets(int iter,bool pathfinder_pshare){
  int net_ctr=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){

    if(net->sinkVector==0)
      continue;

    if(iter==1){
      routeNet(iter,net,0,pathfinder_pshare);
/*       net->print_Info(); */
/*       net->print_rTree(); */
      wireCost+=net->get_rNodeCount();
/*       cout<<endl<<"WIRE COST = "<<wireCost; */
    }
    else{
      net->rip_rTree();
      routeNet(iter,net,1,pathfinder_pshare);
    }

#if defined(ROUTER_DEBUG)
    net->print_Info();
    net->print_rTree();
#endif

  }
  netlist.remove_netListItr();

  sharingCost=FPGA->countSharing(0);

}

int calcWireCost(){
  int wire_cost=0;

  netlist.create_netListItr();
  for(Net* net=netlist.getnext_netList();net!=0;net=netlist.getnext_netList()){
    if(net->sinkVector==0)
      continue;
    wire_cost+=net->get_rNodeCount();
  }
  netlist.remove_netListItr();

  return wire_cost;

}

void populateAllSinksTables(){
  routingNode* rnode;
  priority_queue<pqNode>* PQ;

  FPGA->create_rnodeListItr();
  for(rnode=FPGA->getnext_rnodeList();rnode!=0;rnode=FPGA->getnext_rnodeList()){
    if(rnode->type==SOURCE){
      (rnode->allSinksTable).resize(FPGA->numSinkNodes);
      for(int i=0;i<(rnode->allSinksTable).size();i++)
	(rnode->allSinksTable).at(i)=-1.0;
      PQ=new priority_queue<pqNode>;
      PQ->push(pqNode(rnode,0.0,0.0,0));
      Dijkstra(PQ,rnode);
      delete PQ;
    }
  }
  FPGA->remove_rnodeListItr();
}

void dumpAllSinksTables(){
  routingNode* rnode;

  cout<<endl<<endl<<"<<<<<<<<< Begin Print Sink Distances >>>>>>>>>"<<endl; 

  FPGA->create_rnodeListItr();
  for(rnode=FPGA->getnext_rnodeList();rnode!=0;rnode=FPGA->getnext_rnodeList()){
    if(rnode->type==SOURCE){
      cout<<endl<<rnode->name<<":  ";
      for(int i=0;i<(rnode->allSinksTable.size());i++)
	cout<<(rnode->allSinksTable).at(i)<<"  ";
      cout<<endl;
    }
  }
  FPGA->remove_rnodeListItr();

  cout<<endl<<"<<<<<<<<< End Print Sink Distances >>>>>>>>>"<<endl; 

}

void Dijkstra(priority_queue<pqNode>* PQ,routingNode* source_rnode){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;

/*   cout<<endl<<endl<<"Looking at source node "<<source_rnode->name<<", index = "<<source_rnode->nodeIndex<<"."<<endl; */

  while(!PQ->empty()){

    rnode=(PQ->top()).rNode;
    if(rnode->isExplored==false){
      /* found unexplored routing node...mark it explored,set its pathcost and */
      /* predecessor routing node and break */
      rnode->isExplored=true;
      rnode->pathCost=(PQ->top()).pathCost;
      rnode->predecessor=(PQ->top()).predNode;

      /* fill in cost to sink rnode from source_rnode in costTable */
      if(rnode->get_type()==SINK){
/* 	cout<<"("<<rnode->name<<",index="<<rnode->sinkIndex<<",pathCost="<<rnode->pathCost<<")  "; */
	(source_rnode->allSinksTable).at(rnode->sinkID)=rnode->pathCost;

	if(rnode->pathCost > dLimitMax)
	  dLimitMax = rnode->pathCost;
	if(rnode->pathCost < dLimitMin)
	  dLimitMin = rnode->pathCost;

      }

      cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
      PQ->pop();
    }
    else{
      PQ->pop();
      continue;
    }
 
    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;
      if(nb_rnode->isExplored==false){
        path_cost=rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost(1,true);
	nb_path_cost=nb_rnode->pathCost;
	if(new_path_cost<nb_path_cost)
	  PQ->push(pqNode(nb_rnode,new_path_cost,0.,rnode));
      }
    }
    rnode->remove_fanoutListItr();

  }
  /* end exploration */

  /* clean up all routing nodes that have been affected by the exploration */
  for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
    (*clean_itr)->isExplored=false;
    (*clean_itr)->inrTree=false;
    (*clean_itr)->predecessor=0;
    (*clean_itr)->pathCost=HUGE_NUMBER;
    (*clean_itr)->isRouted=false;
  }

}


/***
* This function finds a route for the net pointed to by 'net'. It assumes that all
* routingNode and Net objects have been cleaned up and initialized for routing.
* This cleanup MUST be separately handled by another routine. Specifically, the 
* previous routing tree should have been ripped up, and routing node members (isExplored,
* occupancy,presentSharing,pathCost,predecessor and inrTree) should have been 
* accordingly changed. 
***/
void routeNet(int iter,Net* net,int change_costs,bool pathfinder_pshare){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  Terminal* terminal;
  priority_queue<pqNode>* PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;
  list<routingNode*>::iterator clean_itr;

  /* new root tree node. initialize it to source routing node */
  net->init_rTreeRoot();
  PQ=new priority_queue<pqNode>;

  /* get source routing node and add to priority queue at path cost 0. */
  /* include the source in the routing tree, and increment occupancy. */
  terminal=net->get_source();
  rnode=terminal->get_rNode();
  PQ->push(pqNode(rnode,0.,0,0));
  rnode->set_inrTree(true);
  rnode->set_occupancy(rnode->get_occupancy()+1);
  if(change_costs==1){
    if(rnode->get_occupancy()>rnode->get_capacity())
      ERROR("number of signals occupying source node exceeds capacity...");
  }

#if defined(ROUTER_DEBUG)
  cout<<endl<<endl<<"<<<<<<<<<============ BEGIN ROUTER DEBUG FOR "<<net->get_name();
  cout<<"============>>>>>>>>>";
  cout<<endl<<"SOURCE LOGIC BLOCK OF THIS NET IS "<<terminal->get_lBlock()->get_name();
  cout<<" AND ITS OUTPUT IS MAPPED TO "<<rnode->get_name();
#endif

  /* start exploration */
  routingNode* sink_rnode;
  int sink_count=0;
  for(;;){
    sink_rnode=net->sinkVector->at(sink_count)->rNode;
    
    /* loop until an unexplored routing node is obtained */
    while(1){
      if(PQ->empty())
        ERROR("priority queue is empty before net routing completed...");
      rnode=(PQ->top()).rNode;
      if(rnode->isExplored==false){
        /* found unexplored routing node...mark it explored,set its pathcost and */
        /* predecessor routing node and break */
        rnode->isExplored=true;
	rnode->pathCost=(PQ->top()).pathCost;
	rnode->predecessor=(PQ->top()).predNode;
        cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
        PQ->pop();
        break;
      }
      PQ->pop();
    }

#if defined(ROUTER_DEBUG)
    cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
    cout<<" AND PREDECESSOR NODE ";
    if(rnode->get_predecessor())
      cout<<rnode->get_predecessor()->get_name();
#endif

    if(rnode->type==SINK){
      /* found a sink terminal of net */
      if(rnode==sink_rnode && rnode->isRouted==false){
#if defined(ROUTER_DEBUG)
        cout<<endl<<endl<<"++++++++++++ FOUND A SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif
        tree_node=backTrace(net,rnode,iter,change_costs);

	if(sink_count<(net->sinkVector->size()-1)){
	  /* will need to delete the old PQ first, and create a new one */
	  delete PQ;PQ=new priority_queue<pqNode>;

/* 	  net->print_rTree(); */

	  /* now reinit the PQ to the current routing tree ala VPR's directed search */ 
	  updatePQ(PQ,(net->rTree).root,net->sinkVector->at(sink_count+1)->rNode);
	}

        rnode->isRouted=true;

        /* clean up all routing nodes */
        for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
	  if((*clean_itr)->inrTree==false){
	    (*clean_itr)->isExplored=false;
	    (*clean_itr)->predecessor=0;
	    (*clean_itr)->pathCost=HUGE_NUMBER;
	    (*clean_itr)->isRouted=false;
	  }
	}

	sink_count++;

	if(sink_count==net->sinkVector->size()){
	  delete PQ;
	  break;
	}
	else
	  continue;

      }
    }

    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;

      /* A* algorithm */
      heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);

#if defined(ROUTER_DEBUG)
      cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif

      if(nb_rnode->isExplored==false){
        path_cost=rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost(iter,pathfinder_pshare);
	PQ->push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode));
#if defined(ROUTER_DEBUG)
	cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	  new_path_cost<<" AND PREDECESSOR "<<rnode->get_name();
#endif

      }

    }
    rnode->remove_fanoutListItr();

  }
  /* end exploration */

  /* clean up all routing nodes that have been affected by the exploration */
  for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
    (*clean_itr)->isExplored=false;
    (*clean_itr)->inrTree=false;
    (*clean_itr)->predecessor=0;
    (*clean_itr)->pathCost=HUGE_NUMBER;
    (*clean_itr)->isRouted=false;
  }

}

/* Macro used below to ensure that fractions are rounded up, but floating   *
 * point values very close to an integer are rounded to that integer.       */

#define ROUND_UP(x) (ceil (x - 0.001))

/* VPR-style cost-to-target calculation. Adopted directly from VPR's code-base */
/* double cost_to_target(routingNode* rnode,routingNode* sink_rnode){ */
/*   double num_segs_same_dir, num_segs_ortho_dir, rr_type; */
/*   int no_need_to_pass_by_clb; */
/*   double inv_length, ortho_inv_length, ylow, yhigh, xlow, xhigh, target_x, target_y; */

/*   target_x = (double)sink_rnode->col_low;   */
/*   target_y = (double)sink_rnode->row_low; */
/*   inv_length = ortho_inv_length = 1.0/4.0;  //uh oh, this is a fucking hack!! */
/*   rr_type = rnode->rr_type; */

/*   ylow = (double)rnode->row_low; */
/*   yhigh = (double)rnode->row_high; */
/*   xlow = (double)rnode->col_low; */
/*   xhigh = (double)rnode->col_high; */

/*   if (rr_type == CHANX) { */
/*     if (ylow > target_y) {         /* Coming from a row above target? */
/*       num_segs_ortho_dir = ROUND_UP ((ylow - target_y + 1.) *  */
/*                                           ortho_inv_length); */
/*       no_need_to_pass_by_clb = 1; */
/*     } */
/*     else if (ylow < target_y - 1) {  /* Below the CLB bottom? */
/*       num_segs_ortho_dir = ROUND_UP ((target_y - ylow) *  */
/* 					  ortho_inv_length); */
/*       no_need_to_pass_by_clb = 1; */
/*     } */
/*     else {              /* In a row that passes by target CLB */
/*       num_segs_ortho_dir = 0; */
/*       no_need_to_pass_by_clb = 0; */
/*     } */

/*     /* Now count horizontal (same dir. as inode) segs. */
/*     if (xlow > target_x + no_need_to_pass_by_clb) { */
/*       num_segs_same_dir = ROUND_UP ((xlow - no_need_to_pass_by_clb -  */
/* 				     target_x) * inv_length); */
/*     } */
/*     else if (xhigh < target_x - no_need_to_pass_by_clb) { */
/*       num_segs_same_dir = ROUND_UP ((target_x - no_need_to_pass_by_clb -  */
/* 				     xhigh) * inv_length); */
/*     } */
/*     else { */
/*       num_segs_same_dir = 0; */
/*     } */
/*   } */

/*   else {  /* inode is a CHANY */
/*     if (xlow > target_x) {         /* Coming from a column right of target? */
/*       num_segs_ortho_dir = ROUND_UP ((xlow - target_x + 1.) *  */
/* 					  ortho_inv_length); */
/*       no_need_to_pass_by_clb = 1; */
/*     } */
/*     else if (xlow < target_x - 1) {  /* Left of and not adjacent to the CLB? */
/*       num_segs_ortho_dir = ROUND_UP ((target_x - xlow) *  */
/* 					  ortho_inv_length); */
/*       no_need_to_pass_by_clb = 1; */
/*     } */
/*     else {              /* In a column that passes by target CLB */
/*       num_segs_ortho_dir = 0; */
/*       no_need_to_pass_by_clb = 0; */
/*     } */

/*     /* Now count vertical (same dir. as inode) segs. */

/*     if (ylow > target_y + no_need_to_pass_by_clb) { */
/*       num_segs_same_dir = ROUND_UP ((ylow - no_need_to_pass_by_clb -  */
/* 				     target_y) * inv_length); */
/*     } */
/*     else if (yhigh < target_y - no_need_to_pass_by_clb) { */
/*       num_segs_same_dir = ROUND_UP ((target_y - no_need_to_pass_by_clb -  */
/* 				     yhigh) * inv_length); */
/*     } */
/*     else { */
/*       num_segs_same_dir = 0; */
/*     } */
/*   } */

/*   return (num_segs_same_dir + num_segs_ortho_dir); */

/* } */


double cost_to_target(routingNode* rnode,routingNode* sink_rnode){

  if(rnode==0 || sink_rnode==0)
    ERROR("\nNULL node in cost to target calculation function...\n");

  int rnode_cluster = rnode->clusterNumber;

  /* if we have an input pin or sink node, return 0 */
  if(rnode_cluster<0.0)
    return 0.0;

  double estimate = AstarTable->at(rnode_cluster)->at(sink_rnode->sinkID);
  if(estimate < 0.0)
    ERROR("currently, the estimate to a sink cannot be -ve...");

  return estimate;

}


/***
* This function finds a route to sink_rnode from net's route tree. It is assumed here
* that there is an existing route tree that connects all sinks EXCEPT sink_rnode. Routing 
* node members (isExplored,occupancy,presentSharing,pathCost,predecessor and inrTree) 
* should be appropriately initialized.  
***/
void routeSink(int iter,Net* net,routingNode* sink_rnode,int change_costs){
  routingNode* rnode;routingNode* nb_rnode;
  routingEdge* nb_edge;
  double path_cost,nb_path_cost,new_path_cost,heur_estimate;
  Terminal* terminal;
  priority_queue<pqNode> PQ;
  treeNode* tree_node;
  list<routingNode*> cleanupList;

  /* init the PQ to the existing route tree of net */
  initPQ(PQ,(net->get_rTree())->get_root(),sink_rnode);

  /* start exploration */
  while(1){
    
    /* loop until an unexplored routing node is obtained */
    while(1){
      if(PQ.empty())
	ERROR("priority queue is empty before net routing completed...");

      rnode=PQ.top().rNode;
      if(rnode->isExplored==false){
	/* found unexplored routing node...mark it explored,set its pathcost and */
	/* predecessor routing node and break */
	rnode->isExplored=true;
	rnode->pathCost=PQ.top().pathCost;
	rnode->predecessor=PQ.top().predNode;
	cleanupList.push_back(rnode);  //this list holds routing nodes that need a clean-up
	PQ.pop();
	break;
      }

      PQ.pop();

    }

#if defined(ROUTER_DEBUG)
    cout<<endl<<"POPPED "<<rnode->get_name()<<" AT PATH COST "<<rnode->get_pathCost();
    cout<<" AND PREDECESSOR NODE ";
    if(rnode->get_predecessor())
      cout<<rnode->get_predecessor()->get_name();
#endif

    if(rnode->type==SINK){
      /* found the sink terminal that we are looking for */
      if(rnode==sink_rnode){
#if defined(ROUTER_DEBUG)
	cout<<endl<<endl<<"++++++++++++ FOUND SINK; WILL NOW ENTER BACKTRACE ++++++++++++";
#endif

	backTrace(net,rnode,iter,change_costs);

	/* clean up all routing nodes that have been affected by the exploration */
	list<routingNode*>::iterator clean_itr;
	for(clean_itr=cleanupList.begin();clean_itr!=cleanupList.end();clean_itr++){
	  (*clean_itr)->isExplored=false;
	  (*clean_itr)->inrTree=false;
	  (*clean_itr)->predecessor=0;
	  (*clean_itr)->pathCost=HUGE_NUMBER;
	  (*clean_itr)->isRouted=false;
	}

	/* now clean up all the routing nodes that are in the new route tree */
	(net->get_rTree())->clean_rNodes(net->get_rTree()->get_root());

	/* exploration ends...return */
	return;

      }
    }

    /* loop thru the fanout neighbor nodes of rnode */ 
    rnode->create_fanoutListItr();
    for(nb_edge=rnode->getnext_fanoutList();nb_edge!=0;nb_edge=rnode->getnext_fanoutList()){
      nb_rnode=nb_edge->destNode;

      /* A* algorithm */
      heur_estimate=ALPHA*cost_to_target(nb_rnode,sink_rnode);

#if defined(ROUTER_DEBUG)
      cout<<endl<<"LOOKING AT NEIGHBOR NODE "<<nb_rnode->get_name();
#endif

      if(nb_rnode->isExplored==false){
	path_cost=rnode->pathCost;
	nb_path_cost=nb_rnode->pathCost;
	new_path_cost=path_cost+nb_rnode->returnCost(iter,false);
	PQ.push(pqNode(nb_rnode,new_path_cost,heur_estimate,rnode));
#if defined(ROUTER_DEBUG)
	cout<<endl<<"\tADDED NODE "<<nb_rnode->get_name()<<" AT PATH COST "<<
	  new_path_cost<<" AND PREDECESSOR "<<rnode->get_name();
#endif
      }

    }
    rnode->remove_fanoutListItr();

  }

}

void initPQ(priority_queue<pqNode>& PQ,treeNode* tree_node,routingNode* new_sink_rnode){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;

  t->get_rNode()->set_inrTree(true);

  heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
  if(t->get_rNode()->get_type()!=SINK && t->get_rNode()->get_type()!=INPUT_PIN){
    if(t->get_parent())
      PQ.push(pqNode(t->get_rNode(),0.,heur_estimate,t->get_parent()->get_rNode()));
    else
      PQ.push(pqNode(t->get_rNode(),0.,heur_estimate,0));
    t->get_rNode()->set_pathCost(0.);
  }

  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    initPQ(PQ,child,new_sink_rnode);
  t->remove_childListItr();

}

void updatePQ(priority_queue<pqNode>* PQ,treeNode* tree_node,routingNode* new_sink_rnode){
  treeNode* t=tree_node;
  treeNode* child;
  double heur_estimate;

  if(t->get_rNode()==0)
    ERROR("something fishy with the backTrace operation here...");
  if(t->get_parent()){
    if(t->get_parent()->get_rNode()!=t->get_rNode()->get_predecessor())
      ERROR("parent mismatch here...");
  }

  if(t->get_rNode()->get_type()!=SINK && t->get_rNode()->get_type()!=INPUT_PIN){
    /* A* algorithm */

    heur_estimate=ALPHA*cost_to_target(t->get_rNode(),new_sink_rnode);
    PQ->push(pqNode(t->get_rNode(),0.,heur_estimate,t->get_rNode()->get_predecessor()));
    t->get_rNode()->set_pathCost(0.);
    t->get_rNode()->set_inrTree(true);
    t->get_rNode()->set_isExplored(false);
  }

  t->create_childListItr();
  for(child=t->getnext_childList();child!=0;child=t->getnext_childList())
    updatePQ(PQ,child,new_sink_rnode);
  t->remove_childListItr();

}

treeNode* backTrace(Net* net,routingNode* sink_rnode,int iter,int change_costs){
  treeNode* curr_tree_node;
  treeNode* prev_tree_node=0;
  routingNode* curr_node;

  /* start backtracing from the newly found sink to a node in the routing tree */
  for(curr_node=sink_rnode;curr_node->get_inrTree()==false;
      curr_node=curr_node->get_predecessor()){

    /* create a new tree node for the current routing node on the backtraced path */
    curr_tree_node=new treeNode(curr_node,curr_node->get_pathCost());

#if defined(ROUTER_DEBUG)
  cout<<endl<<"\tADDED "<<curr_node->get_name()<<" TO ROUTING TREE";
#endif

/*     cout<<endl<<"\tADDED "<<curr_node->get_name()<<" TO ROUTING TREE"; */

    if(prev_tree_node!=0){
      /* hook up parent and child tree nodes */
      curr_tree_node->addto_childList(prev_tree_node);
      prev_tree_node->set_parent(curr_tree_node);
    }

    curr_node->set_occupancy(curr_node->get_occupancy()+1);
    prev_tree_node=curr_tree_node;

    /* change the sharing and wire costs to reflect the addition of this node */
    if(change_costs==1){
      if(curr_node->get_occupancy()>curr_node->get_capacity())
        sharingCost+=1;
      if(curr_node->get_type()==BUS_OR_WIRE)
        wireCost+=1;
    }

  }

  /* curr_node now points to a routing node that is present in the route tree. */
/*   cout<<endl<<endl<<"NEED TO FIND "<<curr_node->get_name()<<" IN ROUTE TREE"; */

  /* find the tree node corresponding to curr_node */
  curr_tree_node=net->find_treeNode(curr_node);

  /* connect the new tree segment to the existing route tree */
  curr_tree_node->addto_childList(prev_tree_node);
  prev_tree_node->set_parent(curr_tree_node);

#if defined(ROUTER_DEBUG)
  cout<<endl<<endl<<"LEAVING BACKTRACE...";
#endif

/*   cout<<endl<<endl<<"FOUND TREE NODE "<<curr_tree_node->get_rNode()->get_name(); */

  /* return the tree node that joins the new tree segment to the existing route tree */
  return prev_tree_node; 

}

void readInVPRPlacementLocation(char* block_name,int x,int y,int io_index){
  vprIOB* io_block;vprCLB* clb;

  if(block_name==0 || x<0 || y<0 || io_index<0 || x>(FPGA->numRows+1) || y>(FPGA->numCols+1) || io_index>=FPGA->numIOs_perIOB)
    ERROR("spurious logic block parameters here..."); 

  netlist.create_lbListItr();
  for(logicBlock* lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){
    if(!strcmp(block_name,lb->get_name())){

/*       cout<<endl<<"Found "<<lb->get_name(); */

      /* found the logic block in the netlist */
      if(lb->get_functionType()==MCNC_IO_TYPE){
	io_block=(vprIOB*)(FPGA->get_blockAt(x,y));
	placeBlock(lb,(void *)(io_block),0,io_index);	    
      }
      else{
	clb=(vprCLB*)(FPGA->get_blockAt(x,y));
	placeBlock(lb,(void *)(clb),1,-1);
      }
    }
  }
  netlist.remove_lbListItr();

}


/***
* This function places the logic block lb on FPGA block b. The parameter
* t indicates whether this is a vprCLB or vprIOB. In case we have a vprIOB,
* then io_index corresponds to the specific IO at which lb is placed. 
* Note that we assume here that there is no other logic block already mapped
* to the FPGA block b...the calling function MUST check occupancy.
***/
void placeBlock(logicBlock* lb,void* b,int t,int io_index){
  vprCLB* clb;vprIOB* iob;
  routingNode* rnode;
  Terminal* term;

  if(lb==0 || b==0)
    ERROR("NULL parameters passed to function...");
  if(t!=0 && t!=1){
    cout<<endl<<endl<<"BLOCK TYPE = "<<t;
    ERROR("invalid FPGA block type...");
  }

  if(t==1){
    /* We have a CLB */
    clb=(vprCLB*)(b);

/*      if(clb->get_currBlock()) */
/*        ERROR("hey, this CLB is already taken...");  */

    clb->set_currBlock(lb);
    lb->col=clb->get_col();lb->row=clb->get_row();

    /* first map the input nets of lb to the sink node of clb */
    rnode=clb->get_sinkNode();
    lb->create_inputListItr();
    for(term=lb->getnext_inputList();term!=0;term=lb->getnext_inputList()){
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
    }
    lb->remove_inputListItr();

    /* now map the output nets of lb to the source node of clb */
    rnode=clb->get_sourceNode();
    lb->create_outputListItr();
    for(term=lb->getnext_outputList();term!=0;term=lb->getnext_outputList()){
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
    }
    lb->remove_outputListItr();

  }
  else{
    /* We have an IOB */
    iob=(vprIOB*)(b);

/*      if(iob->getfrom_currBlockArray(io_index)) */
/*        ERROR("hey, this IOB is already taken..."); */

    iob->addto_currBlockArray(lb,io_index);
    lb->col=iob->get_col();lb->row=iob->get_row();lb->io_index=io_index;

    if(lb->get_isOutput()){
      /* We have an output pad */
      rnode=iob->getfrom_sinkArray(io_index);
      lb->create_inputListItr();
      term=lb->getnext_inputList();
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
      lb->remove_inputListItr();
    }
    else{
      /* We have an input pad */
      rnode=iob->getfrom_sourceArray(io_index);
      lb->create_outputListItr();
      term=lb->getnext_outputList();
      rnode->addto_terminalList(term);
      term->set_rNode(rnode);
      lb->remove_outputListItr();
    }

  }

}


/***
* This function will create a random initial placement of the netlist on the
* FPGA architecture pointed to by vprFPGA
***/
void initPlacement(){
  int num_rows=FPGA->get_numRows()+2;
  int num_cols=FPGA->get_numCols()+2;
  int c,r;
  int i;
  vprIOB* io_block;vprCLB* clb;

  netlist.create_lbListItr();
  for(logicBlock* lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){
    while(1){
      c=lrand48()%num_cols;r=lrand48()%num_rows;

      /* there are no logic or IO blocks at FPGA corners */
      if((c==0 && r==0) || (c==0 && r==num_rows-1) || (c==num_cols-1 && r==0) ||
         (c==num_cols-1 && r==num_rows-1))
        continue;

      /* this is an IOB */
      if(c==0 || r==0 || c==num_cols-1 || r==num_rows-1){
        if(lb->get_isLogic()==false){
          /* the current logic block is in fact an IO pad */
          io_block=(vprIOB*)(FPGA->get_blockAt(r,c));
	  for(i=0;i<io_block->get_numIOs();i++){
	    if(io_block->getfrom_currBlockArray(i)==0){
	      /* found an empty IO location */
		placeBlock(lb,(void *)(io_block),0,i);
		break;
	    }
	  }

	  if(i==io_block->get_numIOs()){
	    /* no unoccupied IOs at this location...try another coordinate */
	    continue;
	  }
	  else{
	    /* the IO pad has been assigned...break out of while loop */
	    break;
	  }

        }
        else{
          /* the current logic block is not an IO pad...try another coordinate */
          continue;
        }
      }

      /* this is a normal CLB */
      else{
        if(lb->get_isLogic()==true){
          /* the current logic block is a CLB */
          clb=(vprCLB*)(FPGA->get_blockAt(r,c));
          if(clb->get_currBlock()){
            /* a logic block has already been mapped to this location...try another coord */
            continue;
          }
          else{
            /* found a location that this logic block can be placed on */
            placeBlock(lb,(void *)(clb),1,-1);
            break;
          }
        }
        else{
          /* the current logic block is an IO pad...try another coordinate */
          continue;
        }
      }

    }
  }
  netlist.remove_lbListItr();

}


void printVPRPlacement(long int random_seed,char* rr_echo,FILE* out_fp){
  char filename [MAX_STR_LEN];
  FILE* fp;

  sprintf(filename,"%s_%s_%ld_%f.p",netlist.get_name(),rr_echo,random_seed,SHARE_FAC);
  fp=fopen(filename,"w");

  fprintf(fp,"\n");
  fprintf(fp,"Netlist file: %s  Architecture file: 8cluster.arch",netlist.get_name());
  fprintf(fp,"\nArray size: %d x %d logic blocks\n",FPGA->get_numCols(),FPGA->get_numRows());

  fprintf(out_fp,"\n");
  fprintf(out_fp,"Netlist file: %s  Architecture file: 8cluster.arch",netlist.get_name());
  fprintf(out_fp,"\nArray size: %d x %d logic blocks\n",FPGA->get_numCols(),FPGA->get_numRows());

/*    cout<<endl; */
/*    cout<<"Netlist file: "<<netlist.get_name();cout<<"  Architecture file: 8cluster.arch"; */
/*    cout<<endl<<"Array size: "<<FPGA->get_numCols()<<" x "<<FPGA->get_numRows()<<" logic blocks"; */
/*    cout<<endl; */

  // this is done because VPR expects the io_indices of blocks
  // mapped to a given IOB to start at 0 and count up.
  FPGA->lineup_IOBs();

  netlist.create_lbListItr();
  for(logicBlock* lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){
/*      cout<<endl<<lb->get_name()<<"\t\t"<<lb->col<<"\t"<<lb->row; */
    fprintf(fp,"\n%s\t\t%d\t%d",lb->get_name(),lb->col,lb->row);
    fprintf(out_fp,"\n%s\t\t%d\t%d",lb->get_name(),lb->col,lb->row);
    if(lb->get_functionType()==MCNC_IO_TYPE){
/*        cout<<"\t"<<lb->io_index; */
      fprintf(fp,"\t%d",lb->io_index);
      fprintf(out_fp,"\t%d",lb->io_index);
    }
    else{
/*        cout<<"\t0"; */
      fprintf(fp,"\t0");
      fprintf(out_fp,"\t0");
    }
  }
  netlist.remove_lbListItr();

/*    cout<<endl; */
  fprintf(fp,"\n\n");
  fprintf(out_fp,"\n\n");
  fclose(fp);

}


/***
* This function prints the current placement of the netlist on the FPGA
***/
void printPlacement(){
  void* block;
  routingNode* rnode;
  Terminal* term;

  cout<<endl<<endl;
  cout<<endl<<"<<<<<<<<<<<<<<<<<<<<< BEGIN PRINT PLACEMENT >>>>>>>>>>>>>>>>>>>>>";

  netlist.create_lbListItr();
  for(logicBlock* lb=netlist.getnext_lbList();lb!=0;lb=netlist.getnext_lbList()){
    cout<<endl<<lb->get_name()<<" AT POSITION ["<<lb->col<<","<<lb->row<<"]"<<" IO_INDEX "<<
    lb->io_index;

    block=FPGA->get_blockAt(lb->row,lb->col);
    /* CORRESPONDENCE CHECK */

    if(lb->get_isLogic()){
      if(lb!=((vprCLB*)(block))->get_currBlock())
	ERROR("correspondence violation in placement phase...");
    }
    else{
      if(lb!=((vprIOB*)(block))->getfrom_currBlockArray(lb->io_index))
	ERROR("correspondence violation in placement phase...");
    }

    /* print the nets that connect to the inputs of the logic block mapped to tile */

    if(lb->get_isLogic())
      rnode=((vprCLB*)(block))->get_sinkNode();
    else
      rnode=((vprIOB*)(block))->getfrom_sinkArray(lb->io_index);

    cout<<endl<<"SINK NODE IS "<<rnode->get_name()<<". NETS MAPPED TO SINK NODE -->"; 
    rnode->create_terminalListItr();
    for(term=rnode->getnext_terminalList();term!=0;term=rnode->getnext_terminalList()){
      /* CORRESPONDENCE CHECK */
      if(term->get_rNode()!=rnode)
        ERROR("correspondence violation in placement phase...");
      cout<<endl<<"\t"<<term->get_net()->get_name();
    }
    rnode->remove_terminalListItr();

    /* print the nets that connect to the outputs of the logic block mapped to tile */

    if(lb->get_isLogic())
      rnode=((vprCLB*)(block))->get_sourceNode();
    else
      rnode=((vprIOB*)(block))->getfrom_sourceArray(lb->io_index);

    cout<<endl<<"SOURCE NODE IS "<<rnode->get_name()<<". NETS MAPPED TO SOURCE NODE -->"; 
    rnode->create_terminalListItr();
    for(term=rnode->getnext_terminalList();term!=0;term=rnode->getnext_terminalList()){
      /* CORRESPONDENCE CHECK */
      if(term->get_rNode()!=rnode)
        ERROR("correspondence violation in placement phase...");
      cout<<endl<<"\t"<<term->get_net()->get_name();
    }
    rnode->remove_terminalListItr();

    cout<<endl;
  }
  netlist.remove_lbListItr();

  cout<<endl<<"<<<<<<<<<<<<<<<<<<<<< END PRINT PLACEMENT >>>>>>>>>>>>>>>>>>>>>";

}

/***
* Error reporting function for the yacc parser.
***/
/*  void yyerror(char* s){ */
/*    fprintf(stderr,"%s AND ERRONEOUS TOKEN IS %s\n",s,yytext); */
/*  } */

/***
* curr_tok, stringList and curr_el are declared in mcncParser.y.  
* If curr_tok is INPUT_TOK, OUTPUT_TOK or CLB_TOK, create a new object
* of mcncClb type. If curr_tok is GLOBAL_TOK or SUBBLOCK_TOK, do
* nothing for the time being. Finally, if curr_tok is PINLIST_TOK, use
* stringList to assign the various fields of curr_el. In case curr_tok is a
* CLB_TOK, use annotation_string to determine latch and 4-LUT usage.
***/
void processCircuitInfo(char* annotation_string){

  if(curr_tok==INPUT_TOK)
    process_mcncInputBlock(annotation_string);
  else if(curr_tok==OUTPUT_TOK)
    process_mcncOutputBlock(annotation_string);
  else if(curr_tok==CLB_TOK)
    process_mcncClb(annotation_string);
  else if(curr_tok==PINLIST_TOK)
    process_mcncPinlist(annotation_string);

  /* For the time being, do nothing about global signals, and subblock info */
  else if(curr_tok==GLOBAL_TOK){
    ;
  }
  else if(curr_tok==SUBBLOCK_TOK){
    ;
  }

}

/***
* Create a new input block and set its name using stringList.
***/
void process_mcncInputBlock(char* annotation_string){

  /* an input block does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating an input block...");
  }

  /* create a new mcncIO block */
  mcncClb* input_block = new mcncClb;

  /* give it a name */
  input_block->set_name(*(stringList.begin()));

  /* make it an input block */
  input_block->set_isInputBlock(true);

  /* set curr_el to point to this block */
  curr_el = input_block;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to mcnc_netlist */
  mcnc_netlist.addto_mcncClbList(input_block);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING INPUT BLOCK WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name<<endl;
#endif

}

/***
* Create a new output block and set its name using stringList. 
***/
void process_mcncOutputBlock(char* annotation_string){

  /* an output block does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating an output block...");
  }

  /* create a new mcncIO block */
  mcncClb* output_block = new mcncClb;

  /* give it a name */
  output_block->set_name(*(stringList.begin()));

  /* make it an output block */
  output_block->set_isOutputBlock(true);

  /* set curr_el to point to this block */
  curr_el = output_block;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to mcnc_netlist */
  mcnc_netlist.addto_mcncClbList(output_block);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING OUTPUT BLOCK WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name<<endl;
#endif

}

/***
* Create a new clb and set its name using stringList. Also, use
* annotation_string to determine latch and/or 4-LUT usage. 
***/
void process_mcncClb(char* annotation_string){

  /* a CLB does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating an input block...");
  }

  /* create a new clb */
  mcncClb* clb = new mcncClb;
  clb->set_isClb(true);

  /* give it a name */
  clb->set_name(*(stringList.begin()));

  /* set curr_el to point to this block */
  curr_el = clb;

  /* set pinlist_tok to curr_tok */
  pinlist_tok = curr_tok;

  /* add to mcnc_netlist */
  mcnc_netlist.addto_mcncClbList(clb);

#if defined(PARSER_DEBUG)
  cout<<endl<<"PROCESSING CLB WITH NAME ";
  char* name = *(stringList.begin());
  cout<<name;
#endif

}

/***
* Process pinlist for the block that is pointed to by curr_el. For now,
* curr_el can point to blocks of type mcncClb. If we have an
* IO block, pinlist should contain exactly 1 net. In case the block
* is a clb, pinlist should contain exactly 27 strings, some or all
* which can be marked 'open'. The first 18 strings correspond to the 
* inputs of the clb, 19 through 26 are the outputs of the clb, and the 27th 
* corresponds to a clock pin. The clock signal is not considered here.
*
* Each net in the pinlist is first searched in mcnc_netlist by name. 
* If found, the returned net is made an input / output net in case 
* curr_el is of type mcncIO, or added to fanin-list if curr_el is of type
* mcncClb.    
***/
void process_mcncPinlist(char* annotation_string){
  mcncClb* iob=0;
  mcncClb* clb=0;
  mcncNet* net;

  /* a pinlist does not have any annotation in an mcnc netlist */ 
  if(annotation_string){
    cout<<endl<<annotation_string<<endl;
    ERROR("found annotation while creating a pinlist...");
  }

  /* determine the type of block for which this pinlist is being processed */
  if(pinlist_tok==INPUT_TOK || pinlist_tok==OUTPUT_TOK)
    iob = curr_el;
  else if(pinlist_tok==CLB_TOK)
    clb = curr_el;
  else
    ERROR("incorrect token type here...");

  if(!iob && !clb)
    ERROR("error in locating block whose pinlist is being processed...");

  /* if we have an mcncIO block, set its input / output net */
  if(iob){
    /* look for the net */
    net = find_mcncNet(*(stringList.begin()));
    /* set its input / output net */
    /* update net data */
    if(iob->get_isOutputBlock()){
      iob->addto_inputList(net);
      net->addto_sinkList(iob);
    }
    else if(iob->get_isInputBlock()){
      iob->addto_outputList(net);
      net->set_source(iob);
    }
    else
      ERROR("io block must be either input_only or output_only...");
  }

  /* we have a clb */
  else if(clb){
    list<char*>::iterator s_itr = stringList.begin();
    for(int i=1;s_itr!=stringList.end();i++,s_itr++){
      /* if we find an 'open', do nothing */
      if(strcmp(*s_itr,"open")==0)
        continue;
      /* look for the net */
      net = find_mcncNet(*s_itr);
      /* first 18 pins correspond to clb inputs */
      /* add clb to net's sinkList */
      if(i<=18){
        clb->addto_inputList(net);
        net->addto_sinkList(clb);
      }
      /* 19 - 26 pins are clb outputs */
      /* make clb the source of each net */
      if(i>=19 && i<=26){
        clb->addto_outputList(net);
        net->set_source(clb);
      }
    }
  }

#if defined(PARSER_DEBUG)
  cout<<"PROCESSING PINLIST:";
  list<char*>::iterator str_itr = stringList.begin();
  for(;str_itr!=stringList.end();str_itr++)
    cout<<endl<<"\t"<<*str_itr;
#endif

}

/***
* Look for a net with name net_name in mcnc_netlist. If found, return it
* else create a new net, give it a name, add it to mcnc_netlist and then
* return it.
***/
mcncNet* find_mcncNet(char* net_name){
  mcncNet* return_net;
  if(!net_name)
    ERROR("name of a net cannot be NULL...");
  if(return_net = mcnc_netlist.found_mcncNet(net_name))
    return return_net;
  else{
    return_net = new mcncNet;
    return_net->set_name(net_name);
    mcnc_netlist.addto_mcncNetList(return_net);
    return return_net;
  }
}




















/***
* Look for a routing node with name node_name. If one doesn't exist, create
* a new one and return.
***/
routingNode* find_rnode(char* rnode_name){
  routingNode* return_node=0;
  string rname = rnode_name;

  if(!rnode_name)
    ERROR("name of a routing node cannot be NULL...");

/*   cout<<endl<<"String object name is "<< rname; */

  if(return_node = rNodeMap[rname]){
    ;
  }
  else{
    return_node = new routingNode;
    return_node->set_name(rnode_name);
    FPGA->addto_rnodeList(return_node);
    rNodeMap[rname]=return_node;
/*     cout<<". String read-back from map is "<<(rNodeMap[rname])->name; */
  }

  return return_node;
  
}

void process_sourceNode(int col,int row){
  int i;

  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(SOURCE);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  if(col==0 || col==(FPGA->numCols+1) || row==0 || row==(FPGA->numRows+1)){
    // We have an IOB here
    vprIOB* io_block=(vprIOB *)(FPGA->get_blockAt(row,col));
    for(i=0;i<io_block->numIOs;i++){
      if(io_block->getfrom_sourceArray(i)==0){
	io_block->addto_sourceArray(G_currNode,i);
	break;
      }
    }
    if(i==io_block->numIOs){
      cout<<endl<<i<<"\t";
      ERROR("oops, too many source nodes assigned to IO location...");
    }
  }
  else{
    // We have a CLB here
    vprCLB* clb=(vprCLB *)(FPGA->get_blockAt(row,col));
    clb->set_sourceNode(G_currNode);
  }

}


void process_ipinNode(int col,int row){
  int i;

  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(INPUT_PIN);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  if(col==0 || col==(FPGA->numCols+1) || row==0 || row==(FPGA->numRows+1)){
    // We have an IOB here
    vprIOB* io_block=(vprIOB *)(FPGA->get_blockAt(row,col));
    for(i=0;i<io_block->numIOs;i++){
      if(io_block->getfrom_inputPinArray(i)==0){
	io_block->addto_inputPinArray(G_currNode,i);
	break;
      }
    }
    if(i==io_block->numIOs)
      ERROR("oops, too many source nodes assigned to IO location...");
  }
  else{
    // We have a CLB here
    vprCLB* clb=(vprCLB *)(FPGA->get_blockAt(row,col));
    clb->addto_inputPinList(G_currNode);
  }

}


void process_sinkNode(int col,int row){
  static int sink_id; 

  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(SINK);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;
  G_currNode->sinkID=sink_id;
  FPGA->numSinkNodes+=1;
  ++sink_id;

  // Sinks are associated in a later step

}


void process_opinNode(int col,int row){
  if(col<0 || col>(FPGA->numCols+1) || row<0 || row>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(OUTPUT_PIN);
  G_currNode->row_low=G_currNode->row_high=row;
  G_currNode->col_low=G_currNode->col_high=col;

  // Output pins are associated in a later step

}


void process_chanNode(int col_low,int row_low,int col_high,int row_high,int chan_info){

  if(col_low<0 || col_high<0 || col_low>(FPGA->numCols+1) || col_high>(FPGA->numCols+1))
    ERROR("array out of bounds here...");
  if(row_low<0 || row_high<0 || row_low>(FPGA->numRows+1) || row_high>(FPGA->numRows+1))
    ERROR("array out of bounds here...");
  G_currNode->set_type(BUS_OR_WIRE);

  G_currNode->row_low=row_low;G_currNode->row_high=row_high;
  G_currNode->col_low=col_low;G_currNode->col_high=col_high;
  G_currNode->rr_type = chan_info;

}


void connect_IONodes(){
  vprIOB* io_block; vprCLB* clb;
  routingNode* rnode; routingEdge* redge;

  for(int i=0;i<=(FPGA->get_numCols()+1);i++){
    for(int j=0;j<=(FPGA->get_numRows()+1);j++){
      if((i==0 && j==0) || (i==0 && j==(FPGA->get_numRows()+1)) ||
	 (j==0 && i==(FPGA->get_numCols()+1)) || (i==(FPGA->get_numCols()+1) && j==(FPGA->get_numRows()+1)))
	continue;

      if(i==0 || i==(FPGA->get_numCols()+1) || j==0 || j==(FPGA->get_numRows()+1)){
	//We have an IOB here...associate output routing nodes
	//and sink routing nodes with IOB
	io_block=(vprIOB*)(FPGA->get_blockAt(j,i));
	for(int k=0;k<io_block->get_numIOs();k++){
	  rnode = io_block->getfrom_sourceArray(k);
	  rnode->create_fanoutListItr();
	  redge = rnode->getnext_fanoutList();
	  rnode->remove_fanoutListItr();
	  if(redge->get_origNode()!=rnode)
	    ERROR("routing edge originating node mismatch...");
	  if(redge->get_destNode()->get_type()!=OUTPUT_PIN)
	    ERROR("routing edge destination node mismatch...");
	  io_block->addto_outputPinArray(redge->get_destNode(),k);

	  rnode = io_block->getfrom_inputPinArray(k);
	  rnode->create_fanoutListItr();
	  redge = rnode->getnext_fanoutList();
	  rnode->remove_fanoutListItr();
	  if(redge->get_origNode()!=rnode)
	    ERROR("routing edge originating node mismatch...");
	  if(redge->get_destNode()->get_type()!=SINK)
	    ERROR("routing edge destination node mismatch...");
	  io_block->addto_sinkArray(redge->get_destNode(),k);
	}
      }
      else{
	//We have a CLB here...associate output routing nodes and 
	//sink node with IOB
	clb = (vprCLB*)(FPGA->get_blockAt(j,i));
	rnode = clb->get_sourceNode();
	rnode->create_fanoutListItr();
	redge=rnode->getnext_fanoutList();
	for(;redge!=0;redge=rnode->getnext_fanoutList()){
	  if(redge->get_origNode()!=rnode)
	    ERROR("routing edge originating node mismatch...");
	  if(redge->get_destNode()->get_type()!=OUTPUT_PIN)
	    ERROR("routing edge destination node mismatch...");
	  clb->addto_outputPinList(redge->get_destNode());
	}
	rnode->remove_fanoutListItr();

	clb->create_inputPinListItr();
	rnode=clb->getnext_inputPinList();
	for(;rnode!=0;rnode=clb->getnext_inputPinList()){
	  rnode->create_fanoutListItr();
	  redge=rnode->getnext_fanoutList();
	  rnode->remove_fanoutListItr();
	  if(redge->get_origNode()!=rnode)
	    ERROR("routing edge originating node mismatch...");
	  if(redge->get_destNode()->get_type()!=SINK)
	    ERROR("routing edge destination node mismatch...");
	  if(redge->get_destNode()->get_capacity()==1)
	    clb->clock_sinkNode=redge->get_destNode();
	  else
	    clb->set_sinkNode(redge->get_destNode());
	}
	clb->remove_inputPinListItr();
      }

    }
  }

}


void process_Edges(int num_edges){
  int i;
  routingNode* fanout_node;
  routingEdge* r_edge;

  if(num_edges<0)
    ERROR("number of fanout edges cannot be -ve...");
  if(num_edges==0)
    return;

  list<char*>::iterator str_itr = stringList.begin();
  for(i=0;str_itr!=stringList.end();str_itr++,i++){
    fanout_node=find_rnode(*str_itr);
    r_edge=new routingEdge;
    r_edge->set_origNode(G_currNode);
    r_edge->set_destNode(fanout_node);
    G_currNode->addto_fanoutList(r_edge);
    fanout_node->addto_faninList(r_edge);
  }

  if(i!=num_edges)
    ERROR("problems while connecting routing nodes with edges...");

}

void process_switchTypes(){
  routingEdge* r_edge;

  if(stringList.size()==0)
    return;

  list<char*>::iterator str_itr = stringList.begin();
  G_currNode->create_fanoutListItr();

  for(;str_itr!=stringList.end();str_itr++){
    r_edge = G_currNode->getnext_fanoutList();

    if(!strcmp(*str_itr,"0")){
      r_edge->set_type(0);
      r_edge->buffered=false;
      r_edge->set_R(196.728);
      r_edge->set_Cin(2.0574e-14);
      r_edge->set_Cout(2.0574e-14);
      r_edge->set_intrinsicDelay(0.0);
    }
    else if(!strcmp(*str_itr,"1")){
      r_edge->set_type(1);
      r_edge->buffered=true;
      r_edge->set_R(393.47);
      r_edge->set_Cin(7.512e-15);
      r_edge->set_Cout(2.0574e-14);
      r_edge->set_intrinsicDelay(5.24e-10);
    }
    else if(!strcmp(*str_itr,"2")){
      r_edge->set_type(2);
      r_edge->buffered=true;
      r_edge->set_R(786.9);
      r_edge->set_Cin(7.512e-15);
      r_edge->set_Cout(1.0762e-14);
      r_edge->set_intrinsicDelay(4.56e-10);
    }
    else if(!strcmp(*str_itr,"3")){
      r_edge->set_type(3);
      r_edge->buffered=true;
      r_edge->set_R(0.0);
      r_edge->set_Cin(7.512e-15);
      r_edge->set_Cout(0.0);
      r_edge->set_intrinsicDelay(1.5e-09);
    }
    else if(!strcmp(*str_itr,"4")){
      r_edge->set_type(4);
      r_edge->buffered=true;
      r_edge->set_R(0.0);
      r_edge->set_Cin(0.0);
      r_edge->set_Cout(0.0);
      r_edge->set_intrinsicDelay(0.0);
    }

  }

  G_currNode->remove_fanoutListItr();

}












/***
* Print the entire mcnc netlist data structure
***/
void print_mcncNetlist(){

  cout<<endl<<endl;
  cout<<endl<<"NAME OF MCNC NETLIST IS "<<mcnc_netlist.get_name();
  cout<<endl<<"NUMBER OF CLBS = "<<mcnc_netlist.get_numClbs();
  cout<<endl<<"NUMBER OF INPUT BLOCKS = "<<mcnc_netlist.get_numInputs();
  cout<<endl<<"NUMBER OF OUTPUT BLOCKS = "<<mcnc_netlist.get_numOutputs();
  cout<<endl<<"NUMBER OF NETS = "<<mcnc_netlist.get_numNets();
  mcnc_netlist.print_mcncClbList();
  mcnc_netlist.print_mcncNetList();

}

/***
* This function converts the mcnc netlist to a generic netlist. This is the
* step that converts benchmark netlists of a particular format to a represe
* ntation that the Independence algorithm routines will recognize and
* operate upon.
***/
void convert2GenericNetlist(mcncNetlist& n){
  n.convert2Generic(netlist);
}
  
/***
* Print the entire generic netlist data structure
***/
void print_GenericNetlist(){

  cout<<endl<<endl;
  cout<<endl<<"NAME OF NETLIST IS "<<netlist.get_name();
  cout<<endl<<"NUMBER OF LOGIC BLOCKS = "<<netlist.get_numlogicBlocks();
  cout<<endl<<"NUMBER OF INPUT BLOCKS = "<<netlist.get_numInputs();
  cout<<endl<<"NUMBER OF OUTPUT BLOCKS = "<<netlist.get_numOutputs();
  cout<<endl<<"NUMBER OF NETS = "<<netlist.get_numNets();
  netlist.print_lbList();
  netlist.print_netList();

}

/***
* Create a vpr challenge FPGA object.
***/
void create_vprFPGA(int num_clb_rows,int num_clb_cols,int num_ios){

  FPGA = new vprFPGA(num_clb_rows,num_clb_cols,num_ios);
/*   cout<<endl<<endl; */
/*   FPGA->print_info(); */

}

