////////////////////////////////////////////////////////////////////
// routingEdge.cc: Definition of routingEdge member functions.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "routingEdge.h"

// Set type of routing edge i.e. buffered or unbuffered
void routingEdge::set_type(int t){  
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
//    if(t!=41 && t!=43)
//      ERROR("mysterious edge type here...");
  type=t;
}

// Set the routing node that originates this edge
void routingEdge::set_origNode(routingNode* n){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  if(n==0)
    ERROR("NULL originating routing node...");
  origNode=n;
}

// Set the routing node that sinks this edge
void routingEdge::set_destNode(routingNode* n){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  if(n==0)
    ERROR("NULL sink routing node...");
  destNode=n;
}

// Set the resistance of this edge
void routingEdge::set_R(double r){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  if(r<0.)
    ERROR("resistance cannot be -ve...");
  R=r;
}

// Set the input capacitance of this edge
void routingEdge::set_Cin(double c){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  if(c<0.)
    ERROR("capacitance cannot be -ve...");
  Cin=c;
}

// Set the output capacitance of this edge
void routingEdge::set_Cout(double c){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  if(c<0.)
    ERROR("capacitance cannot be -ve...");
  Cout=c;
}

// Set the intrinsice delay of this edge
void routingEdge::set_intrinsicDelay(double d){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  if(d<0.)
    ERROR("intrinsic delay cannot be -ve...");
  intrinsicDelay=d;
}

// Get type of this routing edge
int routingEdge::get_type(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return type;
}

// Get the routing node that originates this edge
routingNode* routingEdge::get_origNode(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return origNode;
}

// Get the routing node that sinks this edge
routingNode* routingEdge::get_destNode(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return destNode;
}

// Get the resistance of this edge
double routingEdge::get_R(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return R;
}

// Get the input capacitance of this edge
double routingEdge::get_Cin(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return Cin;
}

// Get the output capacitance of this edge
double routingEdge::get_Cout(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return Cout;
}

// Get the intrinsic delay of this edge
double routingEdge::get_intrinsicDelay(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");
  return intrinsicDelay;
}

// Print this routing edge's info
void routingEdge::print_info(){
  if(magicNum!=20111)
    ERROR("trying to access a deleted node...");

  cout<<destNode->get_name();

//    cout<<endl<<"\t\tPRINT ROUTING EDGE";
//    cout<<endl<<"\t\tORIG IS "<<origNode->get_name()<<" AND DEST IS "<<destNode->get_name();
//    cout<<endl<<"\t\tR = "<<R<<", Cin = "<<Cin<<", Cout = "<<Cout<<", DELAY = "
//        <<intrinsicDelay;
}

