////////////////////////////////////////////////////////////////////
// vprArch.cc : Definition of the member functions of classes that
//              describe the VPR architecture.
// Copyright (C) 2005 Akshay Sharma, Scott Hauck, and the
// University of Washington.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// The GNU General Public License can be found at 
// http://www.gnu.org/licenses/gpl.txt
/////////////////////////////////////////////////////////////////////

#include "vprArch.h"

vprBaseBlock::vprBaseBlock(int r, int c){
  if(r<0 || c<0)
    ERROR("row/col coords and channel width cannot be -ve...");
  row=r;col=c;
  magicNum=21111;
}

vprBaseBlock::~vprBaseBlock(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;
}

int vprBaseBlock::get_row(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return row;
}

int vprBaseBlock::get_col(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  return col;
}

void vprBaseBlock::print_info(){
  if(magicNum!=21111)
    ERROR("trying to double delete node here...");
  cout<<endl<<endl<<"FPGA LOCATION AT x = "<<col<<" AND y = "<<row;
}

/******************************************************************/

vprCLB::vprCLB(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum=21112;
  sourceNode=sinkNode=clock_sinkNode=0;
  ipin_itr=opin_itr=0;
  currBlock=0;
}

vprCLB::~vprCLB(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  else
    derived_magicNum=0;
}

void vprCLB::set_sourceNode(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("source routing node cannot be NULL...");
  sourceNode=r;
}

void vprCLB::set_sinkNode(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("sink routing node cannot be NULL...");
  sinkNode=r;
}

void vprCLB::set_currBlock(logicBlock* l){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  currBlock=l;
}

void vprCLB::addto_inputPinList(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("input routing node cannot be NULL...");
  inputPinList.push_back(r);
}

void vprCLB::addto_outputPinList(routingNode* r){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("output routing node cannot be NULL...");
  outputPinList.push_back(r);
}

void vprCLB::create_inputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(ipin_itr)
    ERROR("an iterator already exists...");
  ipin_itr = new list<routingNode*>::iterator;
  *ipin_itr = inputPinList.begin();
}

void vprCLB::create_outputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(opin_itr)
    ERROR("an iterator already exists...");
  opin_itr = new list<routingNode*>::iterator;
  *opin_itr = inputPinList.begin();
}

void vprCLB::remove_inputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!ipin_itr)
    ERROR("there is no iterator to remove...");
  delete ipin_itr;
  ipin_itr = 0;
}

void vprCLB::remove_outputPinListItr(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!opin_itr)
    ERROR("there is no iterator to remove...");
  delete opin_itr;
  opin_itr = 0;
}

routingNode* vprCLB::getnext_inputPinList(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!ipin_itr)
    ERROR("NULL iterator here...");

  if(*ipin_itr == inputPinList.end())
    return 0;
  else{
    routingNode* ret_ptr = **ipin_itr;
    (*ipin_itr)++;
    return ret_ptr;
  }

}

routingNode* vprCLB::getnext_outputPinList(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  if(!opin_itr)
    ERROR("NULL iterator here...");

  if(*opin_itr == outputPinList.end())
    return 0;
  else{
    routingNode* ret_ptr = **opin_itr;
    (*opin_itr)++;
    return ret_ptr;
  }

}

logicBlock* vprCLB::get_currBlock(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return currBlock;
}

routingNode* vprCLB::get_sourceNode(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return sourceNode;
}

routingNode* vprCLB::get_sinkNode(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");
  return sinkNode;
}

void vprCLB::print_info(){
  if(derived_magicNum!=21112)
    ERROR("trying to double delete node here...");

  vprBaseBlock::print_info();

  if(currBlock)
    cout<<endl<<"LOGIC BLOCK : "<<currBlock->get_name();
  else
    cout<<endl<<"LOGIC BLOCK : ----";

  cout<<endl<<"SOURCE NODE : "<<sourceNode->get_name()<<", SINK NODE : "<<sinkNode->get_name();

  list<routingNode*>::iterator input_itr = inputPinList.begin();
  cout<<endl<<"INPUT ROUTING NODES  :";
  for(;input_itr!=inputPinList.end();input_itr++)
    cout<<"  "<<(*input_itr)->get_name();

  list<routingNode*>::iterator output_itr = outputPinList.begin();
  cout<<endl<<"OUTPUT ROUTING NODES  :";
  for(;output_itr!=outputPinList.end();output_itr++)
    cout<<"  "<<(*output_itr)->get_name();

}

/******************************************************************/

vprIOB::vprIOB(int r,int c) : vprBaseBlock(r,c){
  derived_magicNum=21113;
  sourceArray=sinkArray=inputPinArray=outputPinArray=0;
  currBlockArray=0;
  numIOs=0;
}

vprIOB::vprIOB(int r,int c,int n) : vprBaseBlock(r,c) {
  derived_magicNum=21113;
  numIOs=n;
  sourceArray=new vector<routingNode*>(n);
  sinkArray=new vector<routingNode*>(n);
  inputPinArray=new vector<routingNode*>(n);
  outputPinArray=new vector<routingNode*>(n);
  currBlockArray=new vector<logicBlock*>(n);

  for(int i=0;i<n;i++){
    sourceArray->at(i)=0;
    sinkArray->at(i)=0;
    inputPinArray->at(i)=0;
    outputPinArray->at(i)=0;
    currBlockArray->at(i)=0;
  }

}

vprIOB::~vprIOB(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  else
    derived_magicNum=0;
}

void vprIOB::addto_sourceArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  sourceArray->at(i)=r;
}

void vprIOB::addto_sinkArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  sinkArray->at(i)=r;
}

void vprIOB::addto_inputPinArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  inputPinArray->at(i)=r;
}

void vprIOB::addto_outputPinArray(routingNode* r,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(!r || i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  outputPinArray->at(i)=r;
}

void vprIOB::addto_currBlockArray(logicBlock* l,int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  currBlockArray->at(i)=l;
}

int vprIOB::get_numIOs(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  return numIOs;
}

routingNode* vprIOB::getfrom_sourceArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return sourceArray->at(i);
}

routingNode* vprIOB::getfrom_sinkArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return sinkArray->at(i);
}

routingNode* vprIOB::getfrom_inputPinArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return inputPinArray->at(i);
}

routingNode* vprIOB::getfrom_outputPinArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return outputPinArray->at(i);
}

logicBlock* vprIOB::getfrom_currBlockArray(int i){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");
  if(i<0 || i>=numIOs)
    ERROR("dubious parameters here...");
  return currBlockArray->at(i);
}

void vprIOB::print_info(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");

  vprBaseBlock::print_info();

  for(int i=0;i<numIOs;i++){
    cout<<endl<<"SOURCE: "<<sourceArray->at(i)<<", OUTPUT: "<<outputPinArray->at(i)<<
      ", SINK: "<<sinkArray->at(i)<<", INPUT: "<<inputPinArray->at(i);
    if(currBlockArray->at(i))
      cout<<", IO BLOCK: "<<currBlockArray->at(i)->get_name();
    else
      cout<<", IO BLOCK: ----";
  }

}

void vprIOB::lineup_IOs(){
  if(derived_magicNum!=21113)
    ERROR("trying to double delete node here...");

  logicBlock* lb;routingNode* rnode;Terminal* term;int index=0;
  for(int i=0;i<numIOs;i++){
    if(lb=currBlockArray->at(i)){
      currBlockArray->at(index)=lb;
      lb->io_index=index;
      if(lb->get_isOutput()){
	/* We have an output pad */
	rnode=sinkArray->at(index);
	lb->create_inputListItr();
	term=lb->getnext_inputList();
	rnode->addto_terminalList(term);
	term->set_rNode(rnode);
	lb->remove_inputListItr();
      }
      else{
	/* We have an input pad */
	rnode=sourceArray->at(index);
	lb->create_outputListItr();
	term=lb->getnext_outputList();
	rnode->addto_terminalList(term);
	term->set_rNode(rnode);
	lb->remove_outputListItr();
      }
      index++;
    }
  }

}

/******************************************************************/

void vprFPGA::lineup_IOBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  for(int i=1;i<=numCols;i++){
    ( (vprIOB*)(blockArray->at(i)->at(0)) )->lineup_IOs();
    ( (vprIOB*)(blockArray->at(i)->at(numRows+1)) )->lineup_IOs();
  }

  for(int j=1;j<=numRows;j++){
    ( (vprIOB*)(blockArray->at(0)->at(j)) )->lineup_IOs();
    ( (vprIOB*)(blockArray->at(numCols+1)->at(j)) )->lineup_IOs();
  }

}

long int vprFPGA::countSharing(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  int sharing=0;long int totalSharing=0;
  list<routingNode*>::iterator rnode_itr = rnodeList.begin();
  for(;rnode_itr!=rnodeList.end();rnode_itr++){
    sharing=(*rnode_itr)->get_occupancy() - (*rnode_itr)->get_capacity();
    if(sharing<0)
      sharing=0;
    totalSharing+=sharing;
    if(opt==1)
      cout<<endl<<(*rnode_itr)->get_name()<<"  "<<sharing;
    else if(opt==2){
      if(sharing>0)
	cout<<endl<<(*rnode_itr)->get_name()<<"  "<<(*rnode_itr)->get_occupancy();
    }
  }

  return totalSharing;

}

long int vprFPGA::countCongestedNodeOccupancy(int opt){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
}

void vprFPGA::updateHistoryCosts(int iter,bool during_placement){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  list<routingNode*>::iterator rnodelist_itr = rnodeList.begin();
  int occ=0;int cap=0;double hist=0.;double new_hist=0.;
  for(;rnodelist_itr!=rnodeList.end();rnodelist_itr++){
    occ=(*rnodelist_itr)->get_occupancy();
    cap=(*rnodelist_itr)->get_capacity();
    hist=(*rnodelist_itr)->get_historyCost(iter);

    if(iter==1)
      new_hist=1.0;
    else{
      if(during_placement){
        if(occ>cap)
          new_hist=flomax(1.0,0.9*hist + 0.5);
        else
          new_hist=flomax(1.0,0.9*hist);
      }
      else{
        if(occ>cap)
          new_hist=hist + flomax(0.0,(double)(occ - cap));
        else
          new_hist=hist;
      }
    }

    (*rnodelist_itr)->historyCost=new_hist;

  }

}

void vprFPGA::print_info(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");

  list<routingNode*>::iterator rn_itr=rnodeList.begin();
  for(;rn_itr!=rnodeList.end();rn_itr++)
    (*rn_itr)->print_info();

}

vprFPGA::~vprFPGA(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  else
    magicNum=0;
}

vprFPGA::vprFPGA(int num_rows,int num_cols,int num_ios){
  int i,j;

  numRows=num_rows; numCols=num_cols; numIOs_perIOB=num_ios;
  numCLBs=numRows*numCols; numIOBs=2*numIOs_perIOB*(numRows+numCols);
  magicNum=23111;

  numSinkNodes=0;

  /* allocate the array of blocks */
  blockArray=new vector<vector<void*>*>(numCols+2);
  for(i=0;i<=numCols+1;i++)
    blockArray->at(i)=new vector<void*>(numRows+2);

  /* initialize the array of blocks */
  for(i=0;i<=numCols+1;i++){
    for(j=0;j<=numRows+1;j++)
      blockArray->at(i)->at(j)=0;
  }

  /* populate with CLBs */
  for(i=1;i<=numCols;i++){
    for(j=1;j<=numRows;j++)
      blockArray->at(i)->at(j)=(void *) new vprCLB(j,i);
  }

  /* populate with IOBs */
  for(i=1;i<=numCols;i++){
    blockArray->at(i)->at(0)=(void *) new vprIOB(0,i,numIOs_perIOB);
    blockArray->at(i)->at(numRows+1)=(void *) new vprIOB(numRows+1,i,numIOs_perIOB);
  }
  for(i=1;i<=numRows;i++){
    blockArray->at(0)->at(i)=(void *) new vprIOB(i,0,numIOs_perIOB);
    blockArray->at(numCols+1)->at(i)=(void *) new vprIOB(i,numCols+1,numIOs_perIOB);
  }

}

void* vprFPGA::get_blockAt(int r,int c){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(r<0 || c<0 || r>(numRows+1) || c>(numCols+1))
    ERROR("array overflow here...");

  return blockArray->at(c)->at(r);

}

int vprFPGA::get_numRows(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numRows;
}

int vprFPGA::get_numCols(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numCols;
}

int vprFPGA::get_numCLBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numCLBs;
}

int vprFPGA::get_numIOBs(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  return numIOBs;
}

void vprFPGA::addto_rnodeList(routingNode* r){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!r)
    ERROR("trying to add NULL rnode...");
  rnodeList.push_back(r);
}

void vprFPGA::create_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(rnode_itr)
    ERROR("an iterator already exists...");
  rnode_itr=new list<routingNode*>::iterator;
  *rnode_itr=rnodeList.begin();
}

void vprFPGA::remove_rnodeListItr(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("there is no iterator to remove...");
  delete rnode_itr;
  rnode_itr=0;
}

routingNode* vprFPGA::getnext_rnodeList(){
  if(magicNum!=23111)
    ERROR("trying to double delete node here...");
  if(!rnode_itr)
    ERROR("trying to access NULL iterator here...");

  if(*rnode_itr == rnodeList.end())
    return 0;
  else{
    routingNode* ret_ptr = **rnode_itr;
    (*rnode_itr)++;
    return ret_ptr;
  }

}

routingNode* vprFPGA::found_routingNode(char* rname){
  if(magicNum!=23111)
    ERROR("trying to access a deleted node...");
  if(!rname)
    ERROR("node name cannot be NULL...");
  list<routingNode*>::iterator node_itr = rnodeList.begin();
  for(;node_itr!=rnodeList.end();node_itr++){
    if(strcmp((*node_itr)->get_name(),rname)==0)
      return *node_itr;
  }
  return 0;
}

